// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "GlobalField.hh"
#include <Display.hh>
#include <Settings.hh>
#include <Grid.hh>
#include <Vector.hh>

extern Settings *settings;

//-----------------------------------------------------------------------------
GlobalField::GlobalField()
{
  nvd          = 0;
  dof_values   = 0;
  no_dof       = 0;
  cell_values  = 0;
  constant     = 0;
  
  representation = NONE;
  grid = 0;
}
//-----------------------------------------------------------------------------
GlobalField::GlobalField(Grid *grid)
{
  this->grid = grid;

  no_dof = 0;
  nvd = 0;

  representation = NONE;
  dof_values     = 0;
  cell_values    = 0;
}
//-----------------------------------------------------------------------------
GlobalField::GlobalField(Grid *grid, real constant)
{
  this->grid = grid;

  no_dof = 1;
  this->nvd = 1;

  representation = CONSTANT;
  dof_values     = 0;
  cell_values    = 0;
  this->constant = constant;
}
//-----------------------------------------------------------------------------
GlobalField::GlobalField(Grid *grid, Vector *x, int nvd = 1)
{  
  this->grid = grid;

  no_dof = x->Size()/nvd;
  this->nvd = nvd;

  representation = NODAL;
  dof_values     = x;
  cell_values    = 0;
  constant       = 0.0;
}
//-----------------------------------------------------------------------------
GlobalField::~GlobalField()
{

  
}
//-----------------------------------------------------------------------------
int GlobalField::GetNoDof()
{
  return no_dof;
}
//-----------------------------------------------------------------------------
int GlobalField::GetVectorDim()
{
  return nvd;
}
//-----------------------------------------------------------------------------
void GlobalField::AttachGrid(Grid* grid)
{
  this->grid = grid;
}
//-----------------------------------------------------------------------------
int GlobalField::GetDim(int cell)
{
  // FIXME: no shapefunctions does not have to be same as no nodes 
  return ( grid->GetCell(cell)->GetSize() );
}
//-----------------------------------------------------------------------------
void GlobalField::GetLocalDof(int cell, real *local_dof, int component = 0)
{
  // FIXME: no shapefunctions does not have to be same as no nodes 
  int local_dim = grid->GetCell(cell)->GetSize();

  switch( representation ){ 
  case NODAL: 
    for (int i=0;i<local_dim;i++){
      local_dof[i] = dof_values->Get(grid->GetCell(cell)->GetNode(i)*nvd + component); 
    }
    break;
  case CELL:
    for (int i=0;i<local_dim;i++) 
      local_dof[i] = cell_values->Get(cell*nvd + component); 
    break;
  case CONSTANT:
    for (int i=0;i<local_dim;i++) 
      local_dof[i] = constant;
    break;
  case NONE:
    display->InternalError("GlobalField::GetLocalNodes()","GlobalField is not initialized");
  default:
    display->InternalError("GlobalField::GetLocalNodes()","Wrong GlobalField representation");
  }  
}
//-----------------------------------------------------------------------------
void GlobalField::InitByNodes(Vector *nodal_values, int nvd)
{
  // FIXME: dof and nodes does not have to be identical
  InitByDof(nodal_values,nvd);
}
//-----------------------------------------------------------------------------
void GlobalField::InitByDof(Vector *dof_values, int nvd)
{
  this->dof_values = dof_values;
  this->nvd        = nvd;
  representation              = NODAL;
  no_dof           = dof_values->Size()/nvd;
}
//-----------------------------------------------------------------------------
void GlobalField::InitByCells(Vector *cell_values, int nvd)
{
  this->cell_values = cell_values;
  this->nvd = nvd;
  representation = CELL;
}
//-----------------------------------------------------------------------------    
void GlobalField::InitByConstant(real c)
{
  constant = c;
  representation = CONSTANT;
}
//-----------------------------------------------------------------------------
real GlobalField::GetConstantValue()
{
  switch( representation ){ 
  case CONSTANT:
    return constant;
    break;
  default:
    display->InternalError("GlobalField::GetLocalNodes()","Not a constant GlobalField");
  }
}
//-----------------------------------------------------------------------------
real GlobalField::GetNodalValue(int node)
{
  return GetNodalValue(node,0);
}
//-----------------------------------------------------------------------------
real GlobalField::GetNodalValue(int node, int component)
{
  return GetDofValue(node,component);
}
//-----------------------------------------------------------------------------
real GlobalField::GetDofValue(int dof)
{
  return GetDofValue(dof,0);
}
//-----------------------------------------------------------------------------
real GlobalField::GetDofValue(int dof, int component)
{
  switch( representation ){ 
  case NODAL: 
    return dof_values->Get(dof*nvd+component);
    break;
  case CONSTANT:
    return constant;
    break;
  default:
    display->InternalError("GlobalField::GetLocalNodes()","GlobalField is not initialized");
  }
}
//-----------------------------------------------------------------------------
real GlobalField::GetCellValue(int cell)
{
  return GetCellValue(cell,0);
}
//-----------------------------------------------------------------------------
real GlobalField::GetCellValue(int cell, int component)
{
  switch( representation ){ 
  case CELL:
    return cell_values->Get(cell*nvd+component);
    break;
  case CONSTANT:
    return constant;
    break;
  default:
    display->InternalError("GlobalField::GetLocalNodes()","GlobalField is not initialized");
  }
}
//-----------------------------------------------------------------------------
