// Original author: Johan Hoffman. Added to Kraftwerk 2002-02-05.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "LocalField.hh"
#include "GlobalField.hh"
#include "FunctionSpace.hh"
#include "FiniteElement.hh"
#include "ShapeFunction.hh"
#include <Settings.hh>

extern Settings *settings;

//-----------------------------------------------------------------------------
LocalField::LocalField()
{
  // Local field is not initialised until explicitly updated
  dim               = 0;
  dof               = 0;
  gradient          = 0;
  functionspace     = 0;
  gradient_computed = false;
  component         = 0;
  globalfield       = 0;
}
//-----------------------------------------------------------------------------
LocalField::LocalField(FiniteElement *element)
{
  globalfield = 0;
  
  functionspace = element->GetFunctionSpace();

  dim = functionspace->GetDim();



  //display->Message(0,"new 1: %d",dof);
  
  dof = new real[dim];


  //display->Message(0,"new 1: %d",dof);
  
  
  gradient   = new real[dim];

  gradient_computed = false;
}
//-----------------------------------------------------------------------------
LocalField::~LocalField()
{
  if ( dof )
	 delete [] dof;
  dof = 0;
  if ( gradient )
	 delete [] gradient;
  gradient = 0;
}
//----------------------------------------------------------------------------- 
void LocalField::AttachFiniteElement(FiniteElement *element)
{
  functionspace = element->GetFunctionSpace();

  dim = functionspace->GetDim();

  //display->Message(0,"new 2: %d",dof);
  dof = new real[dim];
  //display->Message(0,"new 2: %d",dof);
  
  gradient = new real[functionspace->GetSpaceDim()];
  
  gradient_computed = false;
}
//----------------------------------------------------------------------------- 
void LocalField::AttachGlobalField(GlobalField *globalfield, int component)
{
  this->globalfield = globalfield;
  this->component = component;
}
//----------------------------------------------------------------------------- 
void LocalField::operator = ( LocalField &lf )
{
  if ( functionspace != lf.GetFunctionSpace() ){
    functionspace = lf.GetFunctionSpace();
    dim = functionspace->GetDim();

    //display->Message(0,"new 3: %d",dof);
    dof = new real[dim];
    //display->Message(0,"new 3: %d",dof);
    
    gradient   = new real[functionspace->GetSpaceDim()];
  }
  
  for (int i=0;i<dim;i++) dof[i] = lf.GetDofValue(i);
    
  if ( lf.GradientComputed() ){
    for (int i=0;i<functionspace->GetSpaceDim();i++) gradient[i] = lf.d(i);
    gradient_computed = true;
  } else{
    gradient_computed = false;
  }
} 
//-----------------------------------------------------------------------------
void LocalField::Add(LocalField &lf)
{
  if ( functionspace != lf.GetFunctionSpace() )
    display->InternalError("LocalField::Add()","Cannot add LocalFields from different functionspaces");

  for (int i=0;i<dim;i++) dof[i] += lf.GetDofValue(i);
}
//-----------------------------------------------------------------------------
void LocalField::Mult(real a)
{
  for (int i=0;i<dim;i++) dof[i] *= a;
}
//-----------------------------------------------------------------------------
const ShapeFunction* LocalField::GetShapeFunction(int i) const 
{
  return functionspace->GetShapeFunction(i);
}
//-----------------------------------------------------------------------------
real LocalField::GetCoord(int node, int dim) const
{
  return functionspace->GetCoord(node,dim);
}
//-----------------------------------------------------------------------------
real LocalField::GetDofValue(int i) const
{
  return dof[i];
}
//-----------------------------------------------------------------------------
real LocalField::GetMeanValue() const
{
  real mean = 0.0;
  for (int i=0;i<dim;i++) mean += dof[i];

  return mean/dim;
}
//-----------------------------------------------------------------------------
int LocalField::GetDim() const
{
  functionspace->GetDim();
}
//-----------------------------------------------------------------------------
int LocalField::GetCellNo() const
{
  return functionspace->GetCellNo();
}
//-----------------------------------------------------------------------------
void LocalField::Update(int cell)
{
  if ( !globalfield )
	 display->Error("Unable to update localfield, no global field attached.");
  
  Resize(globalfield->GetDim(cell));
  
  globalfield->GetLocalDof(cell,dof,component);
  
  gradient_computed = false;
}
//-----------------------------------------------------------------------------
void LocalField::Refill(int c_no, int component, GlobalField *field)
{
  field->GetLocalDof(c_no,dof,component);

  gradient_computed = false;
}
//-----------------------------------------------------------------------------
void LocalField::Refill(int c_no, real *nvals)
{
  for (int i=0; i < functionspace->GetDim(); i++) dof[i] = nvals[i];

  gradient_computed = false;
}
//-----------------------------------------------------------------------------
FunctionSpace* LocalField::GetFunctionSpace()
{
  return functionspace;
}
//-----------------------------------------------------------------------------
FiniteElement* LocalField::GetFiniteElement() const
{
  return functionspace->GetFiniteElement();
}
//-----------------------------------------------------------------------------
void LocalField::ComputeGradient()
{
  EvalGradient(gradient);
}
//-----------------------------------------------------------------------------
bool LocalField::GradientComputed()
{
  return gradient_computed;
}
//-----------------------------------------------------------------------------
void LocalField::EvalGradient(real *grad)
{
  for (int j=0;j<functionspace->GetSpaceDim();j++){
    grad[j] = 0.0;
    for (int i=0;i<functionspace->GetDim();i++) 
      grad[j] += dof[i]*functionspace->GetGradient(i,j);
  }
  gradient_computed = true;
}
//-----------------------------------------------------------------------------
real LocalField::EvalGradient(int component)
{
  if ( !gradient_computed ) ComputeGradient();
  return gradient[component];
}
//-----------------------------------------------------------------------------
real LocalField::dx()
{
  if ( !gradient_computed ) ComputeGradient();
  return gradient[0];
}
//-----------------------------------------------------------------------------
real LocalField::dy()
{
  if ( !gradient_computed ) ComputeGradient();
  return gradient[1];
}
//-----------------------------------------------------------------------------
real LocalField::dz()
{
  if ( !gradient_computed ) ComputeGradient();
  return gradient[2];
}
//-----------------------------------------------------------------------------
real LocalField::d(int i)
{
  if ( !gradient_computed ) ComputeGradient();
  return gradient[i];
}
//-----------------------------------------------------------------------------
real LocalField::operator* (const LocalField &v) const
{
  real sum_i = 0.0;
  real sum = 0.0;
  
  for (int i=0;i<dim;i++){
	 sum_i = 0.0;
	 for (int j=0;j<v.dim;j++)
		sum_i += v.dof[j] * ((*GetShapeFunction(i))*(*v.GetShapeFunction(j)));
	 sum += dof[i] * sum_i;
  }

  for (int i=0;i<dim;i++)
	 printf("   %f",dof[i]);
  printf("\n");
  
  return sum;
}
//-----------------------------------------------------------------------------
real LocalField::operator* (const ShapeFunction &v) const
{
  real sum = 0.0;

  for (int i=0;i<dim;i++)
	 sum += dof[i] * ( (*GetShapeFunction(i))*v );

  return sum;
}
//-----------------------------------------------------------------------------
real LocalField::operator* (real a) const
{
  real sum = 0.0;

  for (int i=0;i<dim;i++)
	 sum += (*GetShapeFunction(i)) * dof[i];
  
  return a*sum;
}
//-----------------------------------------------------------------------------
real operator* (real a, LocalField &v)
{
  // This makes sure that a * LocalField commutes

  return ( v*a );
}
//-----------------------------------------------------------------------------
real operator* (const ShapeFunction &v, const LocalField &w)
{
  // This makes sure that ShapeFunction * LocalField commutes

  return ( w*v );
}
//-----------------------------------------------------------------------------
void LocalField::Resize(int dim)
{
  if ( this->dim != dim ){
	 if ( dof )
		delete [] dof;
	 dof = new real[dim];
	 this->dim  = dim;
  }
}
//-----------------------------------------------------------------------------
