// Original author: Anders Logg. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include <iostream.h>
#include <stdio.h>
#include <strings.h>

#include <Display.hh>
#include <utils.h>
#include <kw_constants.h>

#include "Grid.hh"
#include "Tetrahedron.hh"
#include "Triangle.hh"

//-----------------------------------------------------------------------------
Grid::Grid()
{
  nnodes   = 0;
  ncells   = 0;
  nodes    = 0;
  cells    = 0;
  celltype = CELL_NONE;
  mem      = sizeof(Grid);
  h        = 0.0;
}
//-----------------------------------------------------------------------------
Grid::~Grid()
{
  if ( nodes )
	 delete [] nodes;
  if ( cells )
	 delete [] cells;
}
//-----------------------------------------------------------------------------
void Grid::Init()
{
  ComputeNeighborInfo();
  ComputeSmallestDiameter();
}
//-----------------------------------------------------------------------------
void Grid::Clear()
{
  ClearNeighborInfo();
}
//-----------------------------------------------------------------------------
int Grid::GetNoNodes()
{
  return ( nnodes );
}
//-----------------------------------------------------------------------------
int Grid::GetNoCells()
{
  return ( ncells );
}
//-----------------------------------------------------------------------------
int Grid::GetMaxCellSize()
{
  return ( maxcellsize );
}
//-----------------------------------------------------------------------------
bool Grid::AllCellsEqual()
{
  // No support for hybrid grids yet. All cells are the same.

  return true;
}
//-----------------------------------------------------------------------------
real Grid::GetSmallestDiameter()
{
  if ( h == 0.0 )
	 display->Error("Diameter of smallest cell has not been computed.");

  return h;
}
//-----------------------------------------------------------------------------
Node* Grid::GetNode(int node)
{
   if ( (node < 0) || (node >= nnodes) )
  	 display->InternalError("Grid::GetNode","Illegal node number: %d",node);

	return ( nodes+node );
}
//-----------------------------------------------------------------------------
Cell* Grid::GetCell(int cell)
{
  if ( (cell < 0) || (cell >= ncells) ){
    display->InternalError("Grid::GetCell","Illegal cell number.");
  }

  return ( cells[cell] );
}
//-----------------------------------------------------------------------------
void Grid::Display()
{
  char tmp[KW_WORDLENGTH];

  switch ( celltype ) {
  case CELL_TRIANGLE:
	 sprintf(tmp,"triangles");
	 break;
  case CELL_TETRAHEDRON:
	 sprintf(tmp,"tetrahedrons");
	 break;
  default:
	 sprintf(tmp,"none");
  }
  
  display->Message(0,"Grid: %d nodes, %d cells (%s), %d bytes",
						 nnodes,ncells,tmp,mem);

}
//-----------------------------------------------------------------------------
void Grid::DisplayAll()
{
  Display();
  
  Node n;
  Point p;
  Cell *c;
  
  display->Message(0,"");

  for (int i=0;i<nnodes;i++){
	 n = nodes[i];
	 p = n.GetCoord();
	 printf("n = %d x = (%f,%f,%f) n-n = [ ",i,p.x,p.y,p.z);
	 for (int j=0;j<n.GetNoNodeNeighbors();j++)
		printf("%d ",n.GetNodeNeighbor(j));
	 printf("] n-c = [ ");
	 for (int j=0;j<n.GetNoCellNeighbors();j++)
		printf("%d ",n.GetCellNeighbor(j));
	 printf("]\n");
  }
  printf("\n");
  
  for (int i=0;i<ncells;i++){
	 c = cells[i];
	 printf("c = %d n = ( ",i);
	 for (int j=0;j<c->GetSize();j++)
		printf("%d,",c->GetNode(j));
	 printf(") c-c = [ ");
	 for (int j=0;j<c->GetNoCellNeighbors();j++)
		printf("%d ",c->GetCellNeighbor(j));
	 printf("]\n");
  }
  printf("\n");

}
//-----------------------------------------------------------------------------
void Grid::Read(const char *file)
{
  display->Message(10,"Reading grid file: \"%s\"",file);

  // Try to open the file for reading
  FILE *fp = fopen(file,"r");
  if ( !fp )
	 display->Error("Unable to read grid file: %s.",file);
  
  // Determine file type by suffix
  
  if ( suffix(file,".inp") ) {
	 display->Message(10,"Seems to be an .inp-file (from suffix).");
    ReadINP(fp);
  }
  else{
	 fclose(fp);
	 display->Error("Unknown file format for grid.");
  }

  // Close the file
  fclose(fp);

}
//-----------------------------------------------------------------------------
void Grid::Write(const char *file)
{
  display->Message(10,"Writing grid file: \"%s\"",file);

  // Try to open the file for writing
  FILE *fp = fopen(file,"w");
  if ( !fp )
	 display->Error("Unable to write grid file: %s.",file);
  
  if ( suffix(file,".inp") ) {
	 display->Message(10,"Seems to be an .inp-file (from suffix).");
    WriteINP(fp);
  }
  else{
	 fclose(fp);
	 display->Error("Unknown file format for grid.");
  }
  
  // Close the file
  fclose(fp);

}
//-----------------------------------------------------------------------------
void Grid::ReadINP(FILE *fp)
{
  int d1,d2,d3;
  int material;
  float x,y,z;
  int n1, n2, n3, n4;
  int new_nnodes, new_ncells;
  char tmp[KW_WORDLENGTH];
  CellType new_celltype;

  Node n;
  Tetrahedron *tet;
  Triangle *tri;
  
  // Read first line
  fscanf(fp,"%d %d %d %d %d\n",&new_nnodes,&new_ncells,&d1,&d2,&d3);

  // Check data
  if ( new_nnodes <= 0 )
	 display->Error("Incorrect grid data on row 1.");
  if ( new_ncells <= 0 )
	 display->Error("Incorrect grid data on row 1.");
	 
  // Allocate memory for the nodes (also sets nnodes)
  AllocNodes(new_nnodes);
  
  // Read node positions
  for (int i=0;i<nnodes;i++){
	 fscanf(fp,"%d %f %f %f\n",&d1,&x,&y,&z);
	 n.p.x = x;
	 n.p.y = y;
	 n.p.z = z;
	 nodes[i] = n;
  }

  // Read first cell to determine grid type and allocate data
  fscanf(fp,"%d %d %s %d %d %d",&d1,&material,tmp,&n1,&n2,&n3);
  if ( strcasecmp(tmp,"tri") == 0 ){
	 new_celltype = CELL_TRIANGLE;
	 fscanf(fp,"\n");
  }
  else if ( strcasecmp(tmp,"tet") == 0 ){
	 new_celltype = CELL_TETRAHEDRON;
	 fscanf(fp,"%d\n",&n4);
  }
  else{
	 fclose(fp);
	 display->Error("Unknown cell type: %s",tmp);
  }
  
  // Allocate memory for the cells (also sets ncells)
  AllocCells(new_ncells,new_celltype);

  // Save first cell and read the rest of the cells
  switch ( celltype ){
  case CELL_TRIANGLE:
	 tri = (Triangle *) cells[0];
	 tri->Set(n1-1,n2-1,n3-1,material);
	 for (int i=1;i<ncells;i++){
		fscanf(fp,"%d %d %s %d %d %d\n",&d1,&material,tmp,&n1,&n2,&n3);
		if ( strcasecmp(tmp,"tri") != 0 )
		  display->Error("All cells must be of the same type. Found non-matching cell: %s",tmp);
		tri = (Triangle *) cells[i];
		tri->Set(n1-1,n2-1,n3-1,material);
	 }
	 break;
  case CELL_TETRAHEDRON:
	 tet = (Tetrahedron *) cells[0];
	 tet->Set(n1-1,n2-1,n3-1,n4-1,material);
	 for (int i=1;i<ncells;i++){
		fscanf(fp,"%d %d %s %d %d %d %d\n",&d1,&d2,tmp,&n1,&n2,&n3,&n4);
		if ( strcasecmp(tmp,"tet") != 0 )
		  display->Error("All cells must be of the same type. Found non-matching cell: %s",tmp);
		tet = (Tetrahedron *) cells[i];
		tet->Set(n1-1,n2-1,n3-1,n4-1,material);
	 }
	 break;
  default:
	 display->InternalError("Grid::ReadINP()","Unknown cell type: %d",celltype);
  }

  // Compute maximum cell size
  ComputeMaximumCellSize();
}
//-----------------------------------------------------------------------------
void Grid::WriteINP(FILE *fp)
{
  Point p;
  Triangle *tri;
  Tetrahedron *tet;
  
  // Write first line
  fprintf(fp,"%d %d 0 0 0\n",nnodes,ncells);
  
  // Write node position
  for (int i=0;i<nnodes;i++){
	 p = nodes[i].p;
	 fprintf(fp,"%d %f %f %f\n",i+1,p.x,p.y,p.z);
  }

  // Write cells
  switch ( celltype ){
  case CELL_TRIANGLE:
	 for (int i=0;i<ncells;i++){
		tri = (Triangle *) cells[i];
		fprintf(fp,"%d %d tri %d %d %d\n",i+1,
				  tri->material,
				  tri->nodes[0]+1,tri->nodes[1]+1,tri->nodes[2]+1);
	 }
	 break;
  case CELL_TETRAHEDRON:
	 for (int i=0;i<ncells;i++){
		tet = (Tetrahedron *) cells[i];
		fprintf(fp,"%d %d tet %d %d %d %d\n",i+1,
				  tri->material,
				  tet->nodes[0],tet->nodes[1]+1,tet->nodes[2]+1,tet->nodes[3]+1);
	 }
	 break;
  default:
	 display->InternalError("Grid::WriteINP()","Unknown cell type: %d",
									celltype);
  }
  
}
//-----------------------------------------------------------------------------
void Grid::AllocNodes(int newsize)
{
  // Delete old nodes if any
  if ( nodes )
	 delete nodes;

  // Keep track of allocated size
  mem -= nnodes*sizeof(Node);

  // Allocate new nodes
  nodes = new Node[newsize];

  // Check that we got the memory
  if ( !nodes )
	 display->Error("Out of memory. Unable to allocate new memory for Grid.");

  // Save the number of nodes
  nnodes = newsize;

  // Keep track of allocated size
  mem += nnodes*sizeof(Node);
}
//-----------------------------------------------------------------------------
void Grid::AllocCells(int newsize, CellType newtype)
{
  // Delete old cells
  if ( cells ){

	 delete cells;
	 
	 switch ( celltype ){
	 case CELL_TRIANGLE:
		mem -= ncells*sizeof(Triangle);
		break;
	 case CELL_TETRAHEDRON:
		mem -= ncells*sizeof(Tetrahedron);
		break;
	 default:
		display->InternalError("Grid::AllocCells()","Unknown cell type: %d",celltype);
	 }
	 
  }

  // Save data
  ncells   = newsize;
  celltype = newtype;

  // Allocate memory for cell pointers
  cells = new (Cell *)[ncells];
  mem += ncells*sizeof(Cell *);
  
  // Allocate memory for the cells themselves
  switch ( celltype ){
  case CELL_TRIANGLE:
	 for (int i=0;i<ncells;i++)
		cells[i] = new Triangle;
	 mem  += ncells*sizeof(Triangle);
	 break;
  case CELL_TETRAHEDRON:
	 for (int i=0;i<ncells;i++)
		cells[i] = new Tetrahedron;
	 mem  += ncells*sizeof(Triangle);
	 break;
  default:
	 display->InternalError("Grid::AllocCells()","Unknown cell type: %d",celltype);
  }

  // Check that we got the memory
  if ( !cells )
	 display->Error("Out of memory. Unable to allocate new memory for Grid.");

}
//-----------------------------------------------------------------------------
void Grid::ClearNeighborInfo()
{
  // Clear nodes
  for (int i=0;i<nnodes;i++)
	 nodes[i].Clear();

  // Clear cells
  for (int i=0;i<ncells;i++)
	 cells[i]->Clear();
  
  // FIXME: Add memory tracking

}
//-----------------------------------------------------------------------------
void Grid::ComputeMaximumCellSize()
{
  int maxsize = 0;
  int n;
  
  for (int i=0;i<ncells;i++)
	 if ( (n=cells[i]->GetSize()) > maxsize )
		maxsize = n;
}
//-----------------------------------------------------------------------------
void Grid::ComputeSmallestDiameter()
{
  real hh;
  
  h = 2.0 * cells[0]->ComputeCircumRadius(this);
  
  for (int i=1;i<ncells;i++)
	 if ( (hh = 2.0 * cells[i]->ComputeCircumRadius(this)) < h )
		h = hh;
}
//-----------------------------------------------------------------------------
void Grid::ComputeNeighborInfo()
{
  // Computes the following information:
  //
  // 1. All neighbor cells of a node: n-c
  // 2. All neighbor cells of a cell: c-c (including the cell itself)
  // 3. All neighbor nodes of a node: n-n (including the node itself)
  //
  // To save memory, the algorithm is not optimal w.r.t. speed.
  // The trick is to compute the connections in the correct order.
    
  int *tmp;
  int tmpsize, newtmpsize;

  // Reset all previous connections
  ClearNeighborInfo();

  // Prepare temporary storage
  tmp = new int[nnodes];
  for (int i=0;i<nnodes;i++)
	 tmp[i] = 0;

  display->Message(10,"Grid: Computing n-c connections.");
  
  // Start by computing the n-c connections. This is straightforward, since
  // we won't get any duplicate entries.
  for (int i=0;i<ncells;i++)
	 cells[i]->CountCell(nodes);
  for (int i=0;i<nnodes;i++)
	 nodes[i].AllocateForNeighborCells();
  for (int i=0;i<ncells;i++)
	 cells[i]->AddCell(nodes,tmp,i);

  // Delete temporary storage
  delete tmp;

  display->Message(10,"Grid: Computing c-c connections.");
  
  // Now use this information to compute the c-c connections
  for (int i=0;i<ncells;i++)
	 cells[i]->ComputeCellNeighbors(nodes,i);

  // Prepare temporary storage
  tmpsize = nodes[0].GetMaxNodeNeighbors(cells);
  tmp = new int[tmpsize];

  display->Message(10,"Grid: Computing n-n connections.");
  
  // Now again use the n-c informaton to compute the n-n connections
  for (int i=0;i<nnodes;i++){
	 newtmpsize = nodes[i].GetMaxNodeNeighbors(cells);
	 if ( newtmpsize > tmpsize ){
		tmpsize = newtmpsize;
		delete tmp;
		tmp = new int[tmpsize];
	 }
	 nodes[i].ComputeNodeNeighbors(cells,i,tmp);
  }
  
  // Clear temporary storage
  delete tmp;
}
//-----------------------------------------------------------------------------
