// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "OpenDX.hh"
#include <Settings.hh>
#include <Vector.hh>

#include <stdlib.h>

extern Settings *settings;

//-----------------------------------------------------------------------------
OpenDX::OpenDX(const char *cFileName, int iNumberOfData, ...)
{
  // Save the file name
  sprintf(this->cFileName,cFileName);

  // Save the number of data
  this->iNumberOfData = iNumberOfData;

  // Allocate memory for the dimension array
  iDataDimensions = new int[iNumberOfData];
  
  // Allocate memory for the data labels
  cDataLabels = new (char *)[iNumberOfData];
  for (int i=0;i<iNumberOfData;i++){
	 cDataLabels[i] = new char[KW_LINELENGTH];
	 sprintf(cDataLabels[i],"data component %d",i);
  }

  // Get dimensions of the data
  va_list aptr;
  va_start(aptr,iNumberOfData);
  for (int i=0;i<iNumberOfData;i++)
	 iDataDimensions[i] = va_arg(aptr,int);
  va_end(aptr);

  // Compute total data size
  datasize = 0;
  for (int i=0;i<iNumberOfData;i++)
	 datasize += iDataDimensions[i];

  // Clear file only first time
  bFileCleared = false;
  
  // Write the header only once
  bWrittenHeader = false;
  
  // Write the grid only once
  // FIXME: should be able to handle changing grids
  bWrittenGrid = false;

  // Reset the frame counter
  iFrame = 1;

  // Reset the object counter
  iObject = 1;

  // Allocate the time values
  iTimeValuesSize = 512;
  rTimeValues = new real[iTimeValuesSize];
}
//-----------------------------------------------------------------------------
OpenDX::~OpenDX()
{
  // Clear the dimension array
  delete iDataDimensions;

  // Clear the labels
  for (int i=0;i<iNumberOfData;i++)
	 delete cDataLabels[i];
  delete cDataLabels;

  // Clear the time values
  delete rTimeValues;
}
//-----------------------------------------------------------------------------
void OpenDX::SetLabel(int iDataComponent, const char *cLabel)
{
  // Check that we got a proper index
  if ( (iDataComponent < 0) || (iDataComponent > (iNumberOfData-1)) ){
	 cout << "Non-existent data component." << endl;
	 return;
  }

  // Set the label
  sprintf(cDataLabels[iDataComponent],cLabel);
}
//-----------------------------------------------------------------------------
void OpenDX::Reset()
{
  bFileCleared = false;
  bWrittenHeader = false;
  bWrittenGrid = false;
  iFrame = 1;
  iObject = 1;
}
//-----------------------------------------------------------------------------
void OpenDX::AddFrame(Grid *grid, Vector *u, real t)
{  
  // Clear file the first time
  if ( !bFileCleared ){
	 fp = fopen(cFileName,"w");
	 if ( !fp ){
		cout << "Unable to write to file " << cFileName << "." << endl;
		assert(fp);
	 }
	 fclose(fp);
	 bFileCleared = true;
  }

  // Reopen the file and append new data
  fp = fopen(cFileName,"r+");
  if ( !fp ){
	 cout << "Unable to write to file " << cFileName << "." << endl;
	 assert(fp);
  }
  
  // Step to the end of the file
  fseek(fp,0L,SEEK_END);
  
  // Write header the first time
  if ( !bWrittenHeader ){
	 WriteHeader();
	 bWrittenHeader = true;
  }
  
  // Write grid the first time
  if ( !bWrittenGrid ){
	 WriteGrid(grid);
	 bWrittenGrid = true;
  }

  // Remove the previous time series
  if ( iFrame > 1 )
	 RemoveSeries();

  // Save time value
  SaveTime(t);
  
  // Write field
  WriteField(grid,u);

  // Write the time series
  WriteSeries();
  
  // Close the file
  fclose(fp);
}
// //-----------------------------------------------------------------------------
// void OpenDX::WriteInp (Grid *grid, real t)
// {
//   MV_ColMat<real> data;
//   GetData(grid,&data,t);
    
//   char *string = new char[KW_LINELENGTH];
//   sprintf(string,"%s_t_%.10.inp",cFileName,t);
  
//   WriteInpFromData(string, grid, &data, iNumberOfData, 0);
// }
// //-----------------------------------------------------------------------------
// void OpenDX::WriteInpFromData(char *file, Grid *grid, MV_ColMat<real>* data, int nsys, int firstcomp)
// {
//   fp = fopen(file,"w");

//   fprintf(fp, "%i %i %i 0 0\n",grid->GetNoNodes(),grid->GetNoCells(), nsys);

//   Cell *c;
//   Point p;
  
//   for(int i=0;i<grid->GetNoNodes();i++){
// 	 p = grid->GetNode(i)->GetCoord();
//     fprintf(fp, "%i %f %f %f\n", i+1,p.x,p.y,p.z);
//   }
//   for(int i=0;i<grid->GetNoCells();i++){
// 	 c = grid->GetCell(i);
//     fprintf(fp, "%i %i tet %i %i %i %i\n", i+1,
// 				c->GetMaterial(),
// 				c->GetNode(0)+1,
// 				c->GetNode(1)+1,
// 				c->GetNode(2)+1,
// 				c->GetNode(3)+1);
//   }
  
//   fprintf(fp,"%i ", nsys);
//   for (int i=0;i<nsys;i++) fprintf(fp,"1 ");
//   fprintf(fp,"\n");
//   for (int i=0;i<nsys;i++) fprintf(fp,"u %i, Unit\n",i+1);
//   for(int i=0;i<grid->GetNoNodes();i++){
//     fprintf(fp, "%i ", i+1); 
//     for (int j=0;j<nsys;j++){
//       fprintf(fp, "%f ", (*data)(i,firstcomp+j));
//     }
//     fprintf(fp, "\n"); 
//   }
//   fclose(fp);
  
// }
// //-----------------------------------------------------------------------------
// void OpenDX::GetData(Grid *grid, MV_ColMat<real> *data, real t)
// {
//   // Note that this function is a little shaky. I depends on things being
//   // done in the correct order, and that the file should be originally
//   // written in a special way.
//   // Ideally it should be able to handle any reasonable dx-file.

//   int obj1, obj2;
//   float t1, t2;
  
//   // Open the file
//   fp = fopen(cFileName,"r");
//   if ( !fp ){
// 	 cout << "Unable to read file " << cFileName << "." << endl;
// 	 assert(fp);
//   }

//   // Find the object numbers of the data to be interpolated
//   FindObject(t,&obj1,&obj2,&t1,&t2);

//   // Read the grid
//   ReadGrid(grid);

//   // Read data
//   for (int i=0;i<iNumberOfData;i++){
// 	 ReadData(data,i,obj1,obj2,t1,t2,t);
// 	 obj1 += 2;
// 	 obj2 += 2;
//   }
  
//   // Close the file
//   fclose(fp);
  
// }
// //-----------------------------------------------------------------------------
// void OpenDX::GetDataComponent(Grid *grid, MV_ColMat<real> *data,
// 										real t, int component)
// {
//   cout << "Does not work yet." << endl;
//   exit(1);
// }
//-----------------------------------------------------------------------------
void OpenDX::WriteHeader()
{
  char cDescription[KW_LINELENGTH];
  int  i;
  
  // Get the problem description
  settings->Get("problem description",cDescription);

  // Write the header, avoid trashing the file if description contains newline
  fprintf(fp,"# This file is generated by Kraftwerk, version %s.\n",KW_VERSION);
  fprintf(fp,"# The format is intended for use with OpenDX (Data Explorer).\n");
  fprintf(fp,"#\n");
  fprintf(fp,"# Problem description:\n");
  fprintf(fp,"#\n");
  fprintf(fp,"#   ");
  for (i=0;(i<KW_LINELENGTH) && cDescription[i] && (cDescription[i] != '\n');i++)
	 fprintf(fp,"%c",cDescription[i]);
  fprintf(fp,"\n#\n");
  fprintf(fp,"\n");
  
}
//-----------------------------------------------------------------------------
void OpenDX::WriteGrid(Grid *grid)
{
  int noNodes = grid->GetNoNodes();
  int noElems = grid->GetNoCells();
  int i;
  float x,y,z;
  int n1, n2, n3, n4;
  
  // Write nodes

  Point p;
  Cell *c;
  
  fprintf(fp,"# A list of all node positions\n");
  fprintf(fp,"object 1 class array type float rank 1 shape 3 items %d lsb binary data follows\n",
                         noNodes);
  
  for(i=0;i<noNodes;i++){

	 p = grid->GetNode(i)->GetCoord();
	 
	 x = float(p.x);
	 y = float(p.y);
	 z = float(p.z);
	 
	 fwrite(&x,sizeof(float),1,fp);
	 fwrite(&y,sizeof(float),1,fp);
	 fwrite(&z,sizeof(float),1,fp);
         
  }
  fprintf(fp,"\n\n");

  // Write cells
  
  fprintf(fp,"# A list of all elements (connections)\n");
  fprintf(fp,"object 2 class array type int rank 1 shape 4 items %d lsb binary data follows\n",noElems);
  
  for(i=0;i<noElems;i++){

	 c = grid->GetCell(i);
	 
	 n1 = c->GetNode(0);
	 n2 = c->GetNode(1);
	 n3 = c->GetNode(2);
	 n4 = c->GetNode(3);

	 fwrite(&n1,sizeof(int),1,fp);
	 fwrite(&n2,sizeof(int),1,fp);
	 fwrite(&n3,sizeof(int),1,fp);
	 fwrite(&n4,sizeof(int),1,fp);
	 
  }
  fprintf(fp,"\n");
  fprintf(fp,"attribute \"element type\" string \"tetrahedra\"\n");
  fprintf(fp,"attribute \"ref\" string \"positions\"\n");
  fprintf(fp,"\n\n");  

  // Increase the object counter
  iObject += 2;
  
}
//-----------------------------------------------------------------------------
// void OpenDX::WriteField(Grid *grid, MV_ColMat<real> *data)
// {
//   int noNodes = grid->GetNoNodes();
//   float fValue;
//   int iComponent = 0;
  
//   for (int i=0;i<iNumberOfData;i++){

// 	 //cout << "Saving data for " << cDataLabels[i] << endl;
	 
// 	 // Write header for object
// 	 fprintf(fp,"# Values for [%s] at nodal points, frame %d\n",cDataLabels[i],iFrame);
// 	 fprintf(fp,"object %d class array type float rank %d shape %d items %d lsb binary data follows\n",
// 				iObject,1,iDataDimensions[i],noNodes);
	 
// 	 // Write data
// 	 for (int j=0;j<noNodes;j++){

// 		for (int k=0;k<iDataDimensions[i];k++){

// 		  // Get the value
// 		  fValue = float( (*data)(j,iComponent+k) );

// 		  // Write the value
// 		  fwrite(&fValue,sizeof(float),1,fp);
		  
// 		}

// 	 }
// 	 fprintf(fp,"\n");
// 	 fprintf(fp,"attribute \"dep\" string \"positions\"\n\n");
	 
// 	 // Increase the object counter
// 	 iObject += 1;
	 
// 	 // Write field
// 	 fprintf(fp,"# Field for [%s], frame %d\n",cDataLabels[i],iFrame);
// 	 fprintf(fp,"object %d class field\n",iObject);
// 	 fprintf(fp,"component \"positions\" value 1\n");
// 	 fprintf(fp,"component \"connections\" value 2\n");
// 	 fprintf(fp,"component \"data\" value %d\n\n",iObject-1);

// 	 // Increase the object counter
// 	 iObject += 1;	 

// 	 // Step to start of next component
// 	 iComponent += iDataDimensions[i];
	 
//   }
  
//   // Increase the frame counter
//   iFrame +=1;
// }
//-----------------------------------------------------------------------------
void OpenDX::WriteField(Grid *grid, Vector *u)
{
  int noNodes = grid->GetNoNodes();
  float fValue;
  int iComponent = 0;

  for (int i=0;i<iNumberOfData;i++){

	 //cout << "Saving data for " << cDataLabels[i] << endl;
	 
	 // Write header for object
	 fprintf(fp,"# Values for [%s] at nodal points, frame %d\n",cDataLabels[i],iFrame);
	 fprintf(fp,"object %d class array type float rank %d shape %d items %d lsb binary data follows\n",
				iObject,1,iDataDimensions[i],noNodes);
	 
	 // Write data
	 for (int j=0;j<noNodes;j++){

		for (int k=0;k<iDataDimensions[i];k++){

		  // Get the value
		  //display->Message(0,"j=%d datasize=%d iComponent=%d k=%d noNodes=%d u->Size()=%d",
		  //						 j,datasize,iComponent,k,noNodes,u->Size());
		  fValue = float( u->Get(j*datasize+iComponent+k) );
		  
		  // Write the value
		  fwrite(&fValue,sizeof(float),1,fp);
		  
		}

	 }
	 fprintf(fp,"\n");
	 fprintf(fp,"attribute \"dep\" string \"positions\"\n\n");
	 
	 // Increase the object counter
	 iObject += 1;
	 
	 // Write field
	 fprintf(fp,"# Field for [%s], frame %d\n",cDataLabels[i],iFrame);
	 fprintf(fp,"object %d class field\n",iObject);
	 fprintf(fp,"component \"positions\" value 1\n");
	 fprintf(fp,"component \"connections\" value 2\n");
	 fprintf(fp,"component \"data\" value %d\n\n",iObject-1);

	 // Increase the object counter
	 iObject += 1;	 

	 // Step to start of next component
	 iComponent += iDataDimensions[i];
	 
  }
  
  // Increase the frame counter
  iFrame +=1;
}
//-----------------------------------------------------------------------------
void OpenDX::RemoveSeries()
{
  // Remove the previous series (the new one will be placed at the end).
  // This makes sure that we have a valid dx-file even if we kill the
  // program after a few frames.

  char c;
  int iNumberOfSeriesRemoved = 0;
  
  // Step back one step from the end of the file
  fseek(fp,-5L,SEEK_END);
  
  while ( true ){

	 // One step back and read a character
	 fseek(fp,-1,SEEK_CUR);
	 c = fgetc(fp);

	 // Check if we reached a '#'
	 // Warning: if someone puts a # inside a comment we are in trouble
	 if ( c == '#' )
		iNumberOfSeriesRemoved += 1;

	 // Check if we removed all series
	 if ( iNumberOfSeriesRemoved == iNumberOfData ){
		fflush(fp);
		break;
	 }
	 
	 // Step back again to before the character
	 fseek(fp,-1,SEEK_CUR);

  }
  
}
//-----------------------------------------------------------------------------
void OpenDX::SaveTime(real t)
{
  // Check if we have to allocate more space
  if ( iFrame >= (iTimeValuesSize-1) ){
	 cout << "Reached maximum number of frames. Reallocating..." << endl;

	 real *tmp = new real[iTimeValuesSize];
	 for (int i=0;i<iFrame;i++)
		tmp[i] = rTimeValues[i];
	 delete rTimeValues;
	 iTimeValuesSize *= 2;
	 rTimeValues = new real[iTimeValuesSize];
	 for (int i=0;i<iFrame;i++)
		rTimeValues[i] = tmp[i];
	 delete tmp;

	 cout << "Reallocation ok" << endl;
	 
  }

  // Save the value
  rTimeValues[iFrame-1] = t;
}
//-----------------------------------------------------------------------------
void OpenDX::WriteSeries()
{
  // Write the time series

  for (int i=0;i<iNumberOfData;i++){
  
	 fprintf(fp,"# Time series for [%s]\n",cDataLabels[i]);
	 fprintf(fp,"object \"%s series\" class series\n",cDataLabels[i]);
	 
	 for (int j=0;j<(iFrame-1);j++)
		fprintf(fp,"member %d value %d position %f\n",
				  j,
				  3 + 2*iNumberOfData*j + 2*i + 1,
				  rTimeValues[j]);
	 
	 fprintf(fp,"\n");
	 
  }
  
}
//-----------------------------------------------------------------------------
void OpenDX::ReadGrid(Grid *grid)
{
  int noNodes = grid->GetNoNodes();
  int noElems = grid->GetNoCells();

  int noNodes_fromfile = 0;
  int noElems_fromfile = 0;
  int dim_fromfile     = 0;
  
  float x,y,z;
  int n1,n2,n3,n4;

  // Step to the beginning of the file
  fseek(fp,0L,SEEK_SET);
  
  // Read object description for nodes
  if ( !ReadObject("1",&noNodes_fromfile,&dim_fromfile) ){
	 cout << "Error: Unable to read grid. Exiting." << endl;
	 exit(1);
  }
  if ( noNodes_fromfile != noNodes ){
	 cout << "Error: Grid from file does not match." << endl;
	 cout << "- Number of nodes read: " << noNodes_fromfile << " (should be " << noNodes << "." << endl;
	 cout << "Exiting." << endl;
 	 exit(1);
  }
  if ( dim_fromfile != 3 ){
	 cout << "Error: Grid from file does not match." << endl;
	 cout << "- Dimension read: " << dim_fromfile << " (should be 3)." << endl;
	 cout << "Exiting." << endl;
	 exit(1);
  }

  // Read positions
  for (int i=0;i<noNodes;i++){

	 // Read values
	 fread(&x,sizeof(float),1,fp);
	 fread(&y,sizeof(float),1,fp);
	 fread(&z,sizeof(float),1,fp);

	 // Save values: so far we just skip these values.
	 // The grid should be the same anyway...

  }

  // Read object description for elements
  if ( !ReadObject("2",&noElems_fromfile,&dim_fromfile) ){
	 cout << "Error: Unable to read grid. Exiting." << endl;
	 exit(1);
  }
  if ( noElems_fromfile != noElems ){
	 cout << "Error: Grid from file does not match." << endl;
	 cout << "- Number of elements read: " << noElems_fromfile << " (should be " << noElems << "." << endl;
	 cout << "Exiting." << endl;
	 exit(1);
  }
  if ( dim_fromfile != 4 ){
	 cout << "Error: Grid from file does not match." << endl;
	 cout << "- Element size read: " << dim_fromfile << " (should be 4)." << endl;
	 cout << "Exiting." << endl;
	 exit(1);
  }
  
  // Read elements
  for (int i=0;i<noElems;i++){

	 // Read values
	 fread(&n1,sizeof(int),1,fp);
	 fread(&n2,sizeof(int),1,fp);
	 fread(&n3,sizeof(int),1,fp);
	 fread(&n4,sizeof(int),1,fp);
	 
	 // Save values: so far we just skip these values.
	 // The grid should be the same anyway...
	 
  }

}
//-----------------------------------------------------------------------------
// void OpenDX::ReadField(Grid *grid, MV_ColMat<real> *data)
// {
  
  
  
// }
//-----------------------------------------------------------------------------
bool OpenDX::ReadObject(const char *object_name, int *items, int *dimension)
{
  char cWord1[KW_LINELENGTH];
  char cWord2[KW_LINELENGTH];
  char cObjectName[KW_LINELENGTH];
  char cType[KW_LINELENGTH];
  char c;
  int bytes;
  
  // This function tries to step to the position just after the object
  // name [object_name].
  
  while ( true ){

	 // Look for the 'object' keyword
	 while ( (c=getc(fp)) != EOF ){
		
		// Put the character back
		ungetc(c,fp);
		
		// Read a word
		fscanf(fp,"%s",cWord1);

		// Look for 'object'
		if ( strcasecmp(cWord1,"object") == 0 ){
		  fscanf(fp,"%s",cObjectName);
		  break;
		}
		  
	 }

	 // Return false if we couldn't find the object
	 if ( c == EOF )
		return false;
	 
	 // Read some more info
	 while ( (c=getc(fp)) != EOF ){
		
		// Put the character back
		ungetc(c,fp);
		
		// Read two words
		fscanf(fp,"%s %s",cWord1,cWord2);

		// Get data
		if ( strcasecmp(cWord1,"type") == 0 )
		  sprintf(cType,cWord2);
		
		if ( strcasecmp(cWord1,"items") == 0 )
		  *items = int(atof(cWord2));

		if ( strcasecmp(cWord1,"shape") == 0 )
		  *dimension = int(atof(cWord2));
		
		if ( strcasecmp(cWord2,"follows") == 0 )
		  break;

		if ( strcasecmp(cWord2,"field") == 0 )
		  break;
		
	 }

	 // Continue ascii-scanning if we find a field
	 if ( strcasecmp(cWord2,"field") == 0 )
		continue;
	 
	 // Return false if we couldn't find the rest of the object
	 if ( c == EOF )
		return false;

	 // Read the newline
	 getc(fp);
	 
	 // Return if we found the object or skip
	 if ( strcasecmp(cObjectName,object_name) == 0 )
		return true;
		
	 // Skip the object
	 if ( strcasecmp(cType,"float") == 0 )
		bytes = (*dimension)*(*items) * sizeof(float);
	 else if ( strcasecmp(cType,"int") == 0 )
		bytes = (*dimension)*(*items) * sizeof(int);
	 else{
		cout << "Error: Unknown data type '" << cType << "'. Exiting." << endl;
		exit(1);
	 }
	 fseek(fp,bytes,SEEK_CUR);
	 
  }
	 
  return false;
}
//-----------------------------------------------------------------------------
// void OpenDX::ReadData(MV_ColMat<real> *data,
// 							 int component, int obj1, int obj2,
// 							 float t1, float t2, real t)
// {
//   char tmp[KW_LINELENGTH];
//   int iDataSizeFromFile = 0;
//   int iDimensionFromFile = 0;
//   int iComponent = 0;

//   float fValue;
//   float w1 = 1.0;
//   float w2 = 0.0;

//   // Compute weights
//   if ( obj1 != obj2 ){
// 	 w1 = (t2 - float(t)) / (t2-t1);
// 	 w2 = 1.0 - w1;
//   }

//   // Step to the beginning of the file
//   fseek(fp,0L,SEEK_SET);

//   // Step to the correct position in the vector
//   for (int i=0;i<component;i++)
// 	 iComponent += iDataDimensions[i];
  
//   // Read first object
//   sprintf(tmp,"%d",obj1);
//   if ( !ReadObject(tmp,&iDataSizeFromFile,&iDimensionFromFile) ){
// 	 cout << "Error: Unable to read object " << obj1 << ". Exiting." << endl;
// 	 exit(1);
//   }
//   // Check dimensions
//   if ( iDimensionFromFile != iDataDimensions[component] ){
// 	 cout << "Error: Data dimensions don't match. Exiting." << endl;
// 	 exit(1);
//   }
//   // Read data
//   for (int i=0;i<iDataSizeFromFile;i++){
// 	 for (int j=0;j<iDimensionFromFile;j++){
// 		fread(&fValue,sizeof(float),1,fp);
// 		(*data)(i,iComponent+j) = real(w1*fValue);
// 	 }
//   }

//   // Return if we only want to read one component
//   if ( obj1 == obj2 )
// 	 return;

//   // Read second object
//   sprintf(tmp,"%d",obj2);
//   if ( !ReadObject(tmp,&iDataSizeFromFile,&iDimensionFromFile) ){
// 	 cout << "Error: Unable to read object " << obj1 << ". Exiting." << endl;
// 	 exit(1);
//   }
//   // Check dimensions
//   if ( iDimensionFromFile != iDataDimensions[component] ){
// 	 cout << "Error: Data dimensions don't match. Exiting." << endl;
// 	 exit(1);
//   }
//   // Read data
//   for (int i=0;i<iDataSizeFromFile;i++){
// 	 for (int j=0;j<iDimensionFromFile;j++){
// 		fread(&fValue,sizeof(float),1,fp);
// 		(*data)(i,iComponent+j) += real(w2*fValue);
// 	 }
//   }
// }
//-----------------------------------------------------------------------------
void OpenDX::FindObject(real t, int *obj1, int *obj2, float *t1, float *t2)
{
  char c;
  char cWord1[KW_LINELENGTH];
  char cWord2[KW_LINELENGTH];
  char cWord3[KW_LINELENGTH];
  int  noFrames;

  // Reset the times and objects
  *t1 = 0.0;
  *t2 = 0.0;
  *obj1 = 0;
  *obj2 = 0;
  
  // Step to the end of the file
  fseek(fp,0L,SEEK_END);

  // Step back to before the series
  while ( true ){
	 
	 // One step back and read a character
	 if ( fseek(fp,-1,SEEK_CUR) != 0 ){
		cout << "Error: Unable to read time series. Exiting." << endl;
		exit(1);
	 }
	 c = fgetc(fp);

	 // Step back again to before the character
	 fseek(fp,-1,SEEK_CUR);
	 
	 // Check if we reached a '#'
	 // Warning: if someone puts a # inside a comment we are in trouble
	 if ( c == '#' )
		break;
	 
  }  

  // Reset the two position indices
  *obj1 = 0;
  *obj2 = 0;
  
  // Step to the first keyword 'member'
  while ( true ){

	 // Check that we didn't reach end of file
	 if ( (c=getc(fp)) == EOF ){
		cout << "Error: Unable to read solution at time t = " << t << ". Exiting." << endl;
		exit(1);
	 }
	 
	 // Read next word
	 if ( !fscanf(fp,"%s",cWord1) ){
		cout << "Error: Unable to read solution at time t = " << t << ". Exiting." << endl;
		exit(1);
	 }

	 // Check if we found the keyword
	 if ( strcasecmp(cWord1,"member") == 0 ){

		// Set t1 to the previous value of t2
		*t1 = *t2;
		*obj1 = *obj2;
		
		// Read the rest of the line
		if ( !fscanf(fp,"%d %s %d %s %f",
						 &noFrames,
						 cWord1,
						 obj2,
						 cWord2,
						 t2) ){
		  cout << "Error: Unexpected data format." << endl;
		}
			  
		// Set t1 = t2 first time
		if ( noFrames == 0 ){
		  *t1 = *t2;
		  *obj1 = *obj2;
		}
		
		// Check if we found the time we want
		if ( ( (*t1 - KW_EPS) < float(t) ) &&
			  ( (*t2 + KW_EPS) > float(t) )){

		  // Compute object numbers
		  *obj1 -= (2*(iNumberOfData-1) + 1);
		  *obj2 -= (2*(iNumberOfData-1) + 1);

		  //cout << "t1 = " << (*t1) << " " << (*obj1) << endl;
		  //cout << "t2 = " << (*t2) << " " << (*obj2) << endl;
		  
		  return;
		  
		}
		
	 }
	 
  }

  // We shouldn't get here
  cout << "Error: Unable to read field at time t = " << t << ". Exiting." << endl;
  exit(1);
  
}
//-----------------------------------------------------------------------------
