#include "Output.hh"
#include <Vector.hh>
#include <Grid.hh>
#include <Display.hh>
#include <Settings.hh>
#include <utils.h>

extern Settings *settings;

//-----------------------------------------------------------------------------
Output::Output(const char *filename, int no_data, ...)
{
  // Determine file type
  if ( suffix(filename,".dx") ){
	 display->Error("OpenDX not implemented.");
	 filetype = opendx;
  }
  else if ( suffix(filename,".m") ){
	 filetype = matlab;
  }
  else
	 display->Error("Unable to save data to file %s. Unknown file type.",filename);
  
  // Save the file name
  sprintf(this->filename,filename);
  
  // Save the number of data
  this->no_data = no_data;

  // Allocate memory for the dimension array
  dimensions = new int[no_data];
  
  // Allocate memory for the data labels
  labels = new (char *)[no_data];
  for (int i=0;i<no_data;i++){
	 labels[i] = new char[KW_LINELENGTH];
	 sprintf(labels[i],"data component %d",i);
  }
  
  // Get dimensions of the data
  va_list aptr;
  va_start(aptr,no_data);
  for (int i=0;i<no_data;i++)
	 dimensions[i] = va_arg(aptr,int);
  va_end(aptr);

  // Compute total data size
  datasize = 0;
  for (int i=0;i<no_data;i++)
	 datasize += dimensions[i];

  // Check if the file exists
  FILE *fp = fopen(filename,"r");
  if ( fp ){
	 fclose(fp);
	 display->Error("Output file \"%s\" already exists. Remove och rename the file first.",
						 filename);
  }

  // Reset: clear file and reset counter
  Reset();
}
//-----------------------------------------------------------------------------
Output::~Output()
{
  // Clear the dimension array
  delete dimensions;

  // Clear the labels
  for (int i=0;i<no_data;i++)
	 delete labels[i];
  delete labels;
}
//-----------------------------------------------------------------------------
void Output::SetLabel(int component, const char *label)
{
  // Check that we got a proper index
  if ( (component < 0) || (component > (no_data-1)) )
	 display->InternalError("Output::SetLabel()","Illegal data component: %d",
									component);
  
  // Set the label
  sprintf(labels[component],label);
}
//----------------------------------------------------------------------------
void Output::AddFrame(Grid *grid, Vector *u, real t)
{
  // Open the file
  FILE *fp = fopen(filename,"a");
  if ( !fp )
	 display->Error("Unable to write to output file \"%s\".",filename);

  // Update system variables
  UpdateSystemData();

  // Update the problem descriptionx
  settings->Get("problem description",problem_description);
  
  switch ( filetype ){
  case opendx:

	 break;
  case matlab:
	 MatlabAddFrame(fp,grid,u,t);
	 break;
  default:
	 display->InternalError("Output::AddFrame()","Unknown file type.");
  }

  // Increase frame counter
  current_frame += 1;

  // Close file
  fclose(fp);
}
//----------------------------------------------------------------------------
void Output::Reset()
{
  // Clear the file
  FILE *fp = fopen(filename,"w");
  if ( !fp )
	 display->Error("Unable to write to output file \"%s\".",filename);
  fclose(fp);
  
  // Reset frame counter
  current_frame = 0;

  // No header yet
  done_header = false;
  
  // No grid yet
  done_grid = false;
}
//----------------------------------------------------------------------------
void Output::MatlabAddFrame(FILE *fp, Grid *grid, Vector *u, real t)
{
  // Write header
  if ( !done_header ){
	 __MatlabWriteHeader(fp);
	 done_header = true;
  }
  
  // Write grid
  if ( !done_grid ){
	 __MatlabWriteGrid(fp,grid);
	 done_grid = true;
  }

  // Write field
  __MatlabWriteField(fp,u,t);
}
//----------------------------------------------------------------------------
void Output::__MatlabWriteHeader(FILE *fp)
{
  fprintf(fp,"%% Output from Kraftwerk version %s.\n",KW_VERSION);
  fprintf(fp,"%%\n");
  fprintf(fp,"%% Saved by %s at %s\n",sys_user,sys_time);
  fprintf(fp,"%% on %s (%s) running %s version %s.\n",
			 sys_host,sys_machine,sys_sysname,sys_version);
  fprintf(fp,"%%\n");
  fprintf(fp,"%% %s\n",problem_description);
  fprintf(fp,"%%\n");
  fprintf(fp,"%%    p : coordinates for nodes\n");
  fprintf(fp,"%%    e : edges (temporary fix)\n");
  fprintf(fp,"%%    c : nodes in cells\n");
  fprintf(fp,"%%    t     : time values\n");
  fprintf(fp,"%%    u{i}  : solution vector at time t(i)\n");
  fprintf(fp,"\n");
}
//----------------------------------------------------------------------------
void Output::__MatlabWriteGrid(FILE *fp, Grid *grid)
{
  Point p;
  Cell *c;

  // Write nodes
  
  fprintf(fp,"p = [");
  for (int i=0;i<grid->GetNoNodes();i++){
	 p = grid->GetNode(i)->GetCoord();
	 if ( i < (grid->GetNoNodes()-1) )
		fprintf(fp,"%f %f\n",p.x,p.y );
	 else
		fprintf(fp,"%f %f]';\n",p.x,p.y);
  }
  fprintf(fp,"\n");

  // Write cells
  
  fprintf(fp,"c = [");
  for (int i=0;i<grid->GetNoCells();i++){
	 c = grid->GetCell(i);
	 for (int j=0;j<c->GetSize();j++)
		if ( j < (c->GetSize()-1) )
		  fprintf(fp,"%d ",c->GetNode(j)+1);
		else
		  fprintf(fp,"%d",c->GetNode(j)+1);
	 if ( i < (grid->GetNoCells()-1) )
		fprintf(fp," 1\n");
	 else
		fprintf(fp," 1]';\n");
  }
  fprintf(fp,"\n");

  // Write edges (to make the pdeplot routines happy)

  fprintf(fp,"e = [1;2;0;0;0;0;0];\n");
  
}
//----------------------------------------------------------------------------
void Output::__MatlabWriteField(FILE *fp, Vector *u, real t)
{
  fprintf(fp,"t(%d) = %f;\n",current_frame+1,t);
  fprintf(fp,"u{%d} = [",current_frame+1,t);

  for (int i=0;i<u->Size();i++)
	 if ( i < (u->Size()-1) )
		fprintf(fp,"%1.5e\n",u->Get(i));
	 else
		fprintf(fp,"%1.5e];\n",float(u->Get(i)));

  fprintf(fp,"\n");
}
//----------------------------------------------------------------------------
void Output::UpdateSystemData()
{
  env_get_user(sys_user);
  env_get_time(sys_time);
  env_get_host(sys_host);
  env_get_machine(sys_machine);
  env_get_sysname(sys_sysname);
  env_get_version(sys_version);

  display->Message(10,"Output: Reading system data:");
  display->Message(10,"  user    = %s",sys_user);
  display->Message(10,"  time    = %s",sys_time);
  display->Message(10,"  host    = %s",sys_host);
  display->Message(10,"  machine = %s",sys_machine);
  display->Message(10,"  sysname = %s",sys_sysname);
  display->Message(10,"  version = %s",sys_version);
}
//----------------------------------------------------------------------------
