// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.
//
// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-28.

#include <Display.hh>
#include "KrylovSolver.hh"
#include "SISolver.hh"
#include <unistd.h>

//-----------------------------------------------------------------------------
KrylovSolver::KrylovSolver()
{
  A = new SparseMatrix();

  krylov_method = gmres;

  pc = pc_none;

  tol = 1.0e-10;
}
//-----------------------------------------------------------------------------
void KrylovSolver::SetMethod( KrylovMethod krylov_method )
{
  this->krylov_method = krylov_method;
}
//-----------------------------------------------------------------------------
void KrylovSolver::Solve(SparseMatrix *A, Vector *x, Vector *b)
{
  SetMatrix(A);
  Solve(x,b);
}
//-----------------------------------------------------------------------------
void KrylovSolver::Solve(Vector* x, Vector* b)
{
  if (x->Size()!=b->Size())
    x->Resize(b->Size());
  
  norm_b = b->Norm();
  if ( norm_b == 0 ){
    x->SetToConstant(0.0);
    return;
  }

  switch( krylov_method ){ 
  case gmres:
    SolveGMRES(x,b);
    break;
  case cg:
    SolveCG(x,b);
    break;
  default:
    display->InternalError("KrylovSolver::Solve()","Krylov method not implemented");
  }
}
//-----------------------------------------------------------------------------
void KrylovSolver::SolveGMRES(Vector* x, Vector* b)
{
  int max_no_Krylov_vectors = 20;
  
  int max_no_iterations = 100;

  norm_residual = GetResidual(x,b);
  for (int i=0;i<max_no_iterations;i++){
    SolveGMRES_restart_k(x,b,max_no_Krylov_vectors);
    norm_residual = GetResidual(x,b);
    if ( norm_residual < tol * norm_b ){
      display->Status(2,"Restarted GMRES converged after %i (x%i) iterations",i+1,max_no_Krylov_vectors);
      display->Status(2,"Residual = %1.4e, Residual/||b|| = %1.4e",norm_residual,norm_residual/norm_b);
      break;
    }
    if (i == max_no_iterations-1) 
      display->Error("GMRES iterations did not converge, ||res|| = %1.4e",norm_residual);
  }
}
//-----------------------------------------------------------------------------
real KrylovSolver::SolveGMRES_restart_k(Vector* x, Vector* b, int k_max)
{
  // Solve preconditioned problem Ax = AP^(-1)Px = b, 
  // by first solving AP^(-1)v=b through GMRES, then 
  // solve Px = v to get x. Extrem cases is P = A (changes nothing) 
  // and P = I (corresponds to no preconditioner). 
  // AP^(-1)v=b is solved to a tolerance rho/|b| < tol, 
  // where rho is the norm of the residual b-AP^(-1)v = b-Ax,
  // at a maximal number of iterations kmax, 
  // starting from startvector v = Px. 

  int sz = x->Size();
  
  real **H;
  H = new(real *)[k_max+1];
  for (int i = 0; i < k_max+1; i++)
    H[i] = new real[k_max+1];

  for (int i = 0; i < k_max+1; i++){
    for (int j = 0; j < k_max+1; j++) H[i][j] = 0.0;
  }

  real **r;
  r = new(real *)[k_max+1];
  for (int i = 0; i < k_max+1; i++)
    r[i] = new real[k_max+1];

  for (int i = 0; i < k_max+1; i++){
    for (int j = 0; j < k_max+1; j++) r[i][j] = 0.0;
  }

  real **v;
  v = new(real *)[sz];
  for (int i = 0; i < sz; i++)
    v[i] = new real[k_max+1];

  for (int i = 0; i < sz; i++){
    for (int j = 0; j < k_max+1; j++) v[i][j] = 0.0;
  }
  
  real *s;
  s = new real[k_max+1]; 
  for (int i = 0; i < k_max+1; i++) s[i] = 0.0;

  real *y;
  y = new real[k_max+1]; 
  for (int i = 0; i < k_max+1; i++) y[i] = 0.0;

  real *w;
  w = new real[k_max+1]; 
  for (int i = 0; i < k_max+1; i++) w[i] = 0.0;

  real *c;
  c = new real[k_max+1];
  for (int i = 0; i < k_max+1; i++) c[i] = 0.0;

  real *g;
  g = new real[k_max+1]; 
  for (int i = 0; i < k_max+1; i++) g[i] = 0.0;

  real norm_b = 0.0;
  for (int i=0; i < sz; i++) norm_b += sqr(b->Get(i));
  norm_b = sqrt(norm_b);
  
  // Compute start residual = b-AP^(-1)u = b-Ax.
  Vector residual(sz);
  ComputeResidual(x,b,&residual);
  norm_residual = residual.Norm();

  for (int i=0; i < sz; i++) v[i][0] = residual(i)/norm_residual;
      
  Vector tmpvec(sz);

  real rho  = norm_residual;
  real beta = rho;

  real tmp1,tmp2;
  
  for (int i=0; i < k_max+1; i++) g[i] = 0.0;
  g[0] = rho;

  real tmp,nu,norm_v,htmp;
  
  int k,k_end;
  for (k=0; k < k_max; k++){
    
    display->Message(5,"Start GMRES iteration number %i",k+1);
    
    // Compute Krylov vector AP^(-1) v(k) 
    // 2 steps: First solve Px = v, then apply A to x.
    ApplyMatrix(v,k);
    
    for (int j=0; j < k+1; j++){
      H[j][k] = 0.0;
      for (int i=0; i < sz; i++) H[j][k]   += v[i][k+1]*v[i][j];
      for (int i=0; i < sz; i++) v[i][k+1] -= H[j][k]*v[i][j];
    }
    
    norm_v = 0.0;
    for (int i=0; i < sz; i++) norm_v += sqr(v[i][k+1]);
    norm_v = sqrt(norm_v);
    
    H[k+1][k] = norm_v;
    
    // Test for non orthogonality and reorthogonalise if necessary
    if ( !TestForOrthogonality(v) ){
      for (int j=0; j < k+1; j++){
	for (int i=0; i < sz; i++) htmp = v[i][k+1]*v[i][j];
	H[j][k] += htmp;
	for (int i=0; i < sz; i++) v[i][k+1] -= htmp*v[i][j];
      }	  
    }
    
    for (int i=0; i < sz; i++) v[i][k+1] /= norm_v;

    if ( k > 0 ){
      for (int j=0; j < k; j++){
	tmp1 = H[j][k];
	tmp2 = H[j+1][k];
	H[j][k]   = c[j]*tmp1 - s[j]*tmp2;
	H[j+1][k] = s[j]*tmp1 + c[j]*tmp2;
      }
    }
    
    nu   = sqrt( sqr(H[k][k]) + sqr(H[k+1][k]) );
    
    c[k] = H[k][k] / nu;
    s[k] = - H[k+1][k] / nu;
    
    H[k][k]   = c[k]*H[k][k] - s[k]*H[k+1][k];
    H[k+1][k] = 0.0;
    
    tmp    = c[k]*g[k] - s[k]*g[k+1];
    g[k+1] = s[k]*g[k] + c[k]*g[k+1];
    g[k]   = tmp;
    
    rho = fabs(g[k+1]);
    
    display->Message(5,"GMRES residual rho = %1.4e, rho/||b|| = %1.4e, tol = %1.4e",rho,rho/norm_b,tol);

    if ( (rho < tol * norm_b) || (k == k_max-1) ){
      //display->Status(2,"GMRES iteration converged after %i iterations",k);
      k_end = k;
      break;
    }
  }
  
  k = k_end;

  display->Message(5,"Postprocess to obtain solution");

  for (int i=0; i < k+1; i++){ 
    w[i] = g[i];
    for (int j=0; j < k+1; j++){ 
      r[i][j] = H[i][j];
    }
  } 
     

  // Solve triangular system ry = w
  y[k] = w[k] / r[k][k];
  for (int i=1; i < k+1; i++){
    y[k-i] = w[k-i];
    for (int j=0; j < i; j++) y[k-i] -= r[k-i][k-j]*y[k-j];
    y[k-i] /= r[k-i][k-i];
  }
  
  tmpvec.SetToConstant(0.0);
  for (int i=0; i < sz; i++){
    for (int j=0; j < k+1; j++) tmpvec.Add(i,v[i][j]*y[j]);
  }

  if ( pc != pc_none ){
    // Get solution from preconditioned problem (get u from Pu=x)
    SolvePxv(&tmpvec);
  }

  for (int i=0; i < sz; i++) x->Add(i,tmpvec(i));

  norm_residual = GetResidual(x,b);

  //display->Status(2,"Residual = %1.4e, Residual/||b|| = %1.4e",norm_residual,norm_residual/norm_b);

  for (int i = 0; i < k_max+1; i++)
    delete H[i];
  delete H;
  
  for (int i = 0; i < k_max+1; i++)
    delete r[i];
  delete r;
  
  for (int i = 0; i < sz; i++)
    delete v[i];
  delete v;

  delete g;
  delete c;
  delete s;
  delete y;
  delete w;
  
  return norm_residual;
}
//-----------------------------------------------------------------------------
void KrylovSolver::SolveCG(Vector* x, Vector* b)
{
  // Only for symmetric, positive definit problems. 
  // Does not work for the standard way of applying 
  // dirichlet boundary conditions, since then the 
  // symmetry of the matrix is destroyed.

  int sz = x->Size();
  int k_max = 20;
  
  real norm_b = b->Norm();
  
  // Compute start residual = b-Ax.
  Vector residual(sz);
  norm_residual = GetResidual(x,b);

  real *rho;
  rho = new real[k_max+1]; 
  for (int i = 0; i < k_max+1; i++) rho[i] = 0.0;
  rho[0] = sqr(norm_residual);

  Vector w(sz);
  w.SetToConstant(0.0);
  Vector p(sz);
  p.SetToConstant(0.0);

  real alpha,beta,tmp;

  /*
  real diag = 0.0;
  for (int i=0;i<sz;i++){
    diag = A->GetDiagonal(i);
    if ( fabs(diag) > KW_EPS )
      A->ScaleRow(i,1.0/diag);
  }
  */

  for (int k=1; k < k_max; k++){
    
    display->Message(0,"Start CG iteration number %i",k);
    
    display->Message(0,"rho(k-1) = %f, rho(k-2) = %f",rho[k-1],rho[k-2]);
    
    if ( k==1 ){
      p.CopyFrom(&residual);
    } else{
      beta = rho[k-1]/rho[k-2];
      for (int i=0; i < sz; i++) p.Set(i,residual(i) + beta*p(i));
    }      
    
    ApplyMatrix(&p,&w);
    
    tmp = p.Dot(&w);
    alpha = rho[k-1]/tmp;
    
    display->Message(0,"tmp = %f, alpha = %f, norm_p = %f, norm_w = %f",tmp,alpha,p.Norm(),w.Norm());
    
    x->Add(alpha,&p);
    
    residual.Add(-alpha,&w);
    
    rho[k] = sqr(residual.Norm());
    
    display->Message(0,"CG residual rho = %.5f, rho/||b|| = %.5f, tol = %.5f",rho[k-1],rho[k-1]/norm_b,tol);
    
    if ( sqrt(rho[k-1]) < tol*norm_b ){
      display->Message(0,"GMRES iteration converged, ||res|| = %f",rho[k-1]);
      break;
    }
    
  }

  norm_residual = GetResidual(x,b);

  display->Message(0,"Residual/||b|| = %f",norm_residual/norm_b);
  display->Message(0,"Residual = %f",norm_residual);

  delete rho;
}
//-----------------------------------------------------------------------------
void KrylovSolver::ApplyMatrix( Vector *x, Vector *Ax )
{
  A->Mult(x,Ax);
}
//-----------------------------------------------------------------------------
void KrylovSolver::ApplyMatrix( real **x, int comp )
{
  // Preconditioner
  SISolver sisolver;

  switch( pc ){ 
  case pc_richardson:
    sisolver.SetMethod(richardson);
    break;
  case pc_jacobi:
    sisolver.SetMethod(jacobi);
    break;
  case pc_gaussseidel:
    sisolver.SetMethod(gaussseidel);
    break;
  case pc_sor:
    sisolver.SetMethod(sor);
    break;
  case pc_none:
    break;
  default:
    display->InternalError("SISolver::Solve()","Preconditioner not implemented");
  }

  Vector tmp1(A->Size(0)); 
  Vector tmp2(A->Size(0)); 

  for (int i=0; i<A->Size(0); i++) tmp1.Set(i,x[i][comp]);

  // Precondition 
  no_pc_sweeps = 1;
  if ( pc != pc_none ){
    sisolver.SetNoIterations(no_pc_sweeps);
    sisolver.Solve(A,&tmp2,&tmp1);
    tmp1.CopyFrom(&tmp2);
  }      
    
  A->Mult(&tmp1,&tmp2);

  for (int i=0; i<A->Size(0); i++) x[i][comp+1] = tmp2(i);
}
//-----------------------------------------------------------------------------
void KrylovSolver::SetMatrix(SparseMatrix* A)
{
  this->A = A;
}
//-----------------------------------------------------------------------------
void KrylovSolver::SolvePxv( Vector *x )
{
  // Preconditioner
  SISolver sisolver;
  
  switch( pc ){ 
  case pc_richardson:
    sisolver.SetMethod(richardson);
    break;
  case pc_jacobi:
    sisolver.SetMethod(jacobi);
    break;
  case pc_gaussseidel:
    sisolver.SetMethod(gaussseidel);
    break;
  case pc_sor:
    sisolver.SetMethod(sor);
    break;
  case pc_none:
    break;
  default:
    display->InternalError("SISolver::Solve()","Preconditioner not implemented");
  }

  Vector tmp1(A->Size(0)); 

  for (int i=0; i<A->Size(0); i++) tmp1.Set(i,x->Get(i));

  // Solve preconditioned problem 
  sisolver.SetNoIterations(no_pc_sweeps);
  sisolver.Solve(A,x,&tmp1);
}
//-----------------------------------------------------------------------------
void KrylovSolver::ComputeResidual( Vector *x, Vector *b, Vector *res )
{
  real Axrow;
  for (int i=0;i<A->Size(0);i++){
    Axrow  = A->Mult(i,x);
    res->Set(i,b->Get(i)-Axrow);
  }
}
//-----------------------------------------------------------------------------
real KrylovSolver::GetResidual( Vector *x, Vector *b )
{
  real Axrow;
  norm_residual = 0.0;
  for (int i=0;i<A->Size(0);i++){
    Axrow  = A->Mult(i,x);
    norm_residual += sqr(b->Get(i)-Axrow);
  }
  return sqrt(norm_residual);
}
//-----------------------------------------------------------------------------
bool KrylovSolver::TestForOrthogonality( real **v )
{
  // if (||Av_k||+delta*||v_k+1||=||Av_k||)  delta \approx 10^-3

  display->Message(5,"Reorthogonalization needed");

  return false;
}
//-----------------------------------------------------------------------------


