// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "ProblemConvDiff.hh"
#include "EquationConvDiff.hh"

//-----------------------------------------------------------------------------
const char *ProblemConvDiff::Description()
{
  return "Convection-Diffusion equation (3D)";
}
//-----------------------------------------------------------------------------
void ProblemConvDiff::Solve()
{
  no_nodes = grid->GetNoNodes();
  no_cells = grid->GetNoCells();

  real T0,T;
  EquationConvDiff equation;
  SparseMatrix A;
  Vector b;
  Vector xp(no_nodes);
  Vector u(no_nodes);
  KrylovSolver solver;

  // Output of data
  OpenDX output("convdiff.dx",1,1);
  output.SetLabel(0,"Temperature");

  // Convection
  Vector bx(no_nodes);
  Vector by(no_nodes);
  Vector bz(no_nodes);
  for (int i=0;i<no_nodes;i++){
    bx.Set(i,5.0); 
    by.Set(i,0.0); 
    bz.Set(i,0.0); 
  }  

  // Global fields
  GlobalField BETAx(grid,&bx);
  GlobalField BETAy(grid,&by);
  GlobalField BETAz(grid,&bz);
  GlobalField eps(grid,1.0);
  GlobalField up(grid,&xp);
  GlobalField f(grid,1.0);
  
  // Attach fields to equation
  equation.AttachField(0,&BETAx);
  equation.AttachField(1,&BETAy);
  equation.AttachField(2,&BETAz);
  equation.AttachField(3,&up);
  equation.AttachField(4,&eps);
  equation.AttachField(5,&f);

  // Set up the discretiser
  Discretiser discretiser(grid,&equation);
  
  settings->Get("start time",&T0);
  settings->Get("final time",&T);

  // Initial value
  u.SetToConstant(0.0);
  Point pnt;
  real x,y,z;
  for (int i=0;i<no_nodes;i++){
    pnt = grid->GetNode(i)->GetCoord();
    x = real(pnt.x);
    y = real(pnt.y); 
    z = real(pnt.z);
    u.Set(i,y*(1.0-y)*z*(1.0-y)*drand48());
  }
  u.SetToConstant(0.0);
  
  int time_step = 0;
  real t        = T0;
  real dt       = grid->GetSmallestDiameter();
  
  // Save the solution
  output.AddFrame(grid,&u,t);

  // Start time stepping
  while (t < T){

    time_step++;
	 t += dt;
    
    // Set solution to previous solution
    xp.CopyFrom(&u);

	 // Write info
    display->Progress(0,(t-T0)/(T-T0),"i=%i t=%f",time_step,t);
    display->Message(0,"up norm = %f, u norm = %f",xp.Norm(),u.Norm());

	 // Set time and time-step
	 equation.SetTime(t);
	 equation.SetTimeStep(dt);
	 
    // Discretise Convection-Diffusion equation
    discretiser.Assemble(&A,&b);
  
    // Solve the linear system
    solver.Solve(&A,&u,&b);

    // Save the solution
    output.AddFrame(grid,&u,t);
  }
  
}
//-----------------------------------------------------------------------------
