#ifndef __EQUATION_VEL_CG1_CG1_HH
#define __EQUATION_VEL_CG1_CG1_HH

#include <ShapeFunction.hh>
#include <LocalField.hh>
#include <GlobalField.hh>
#include <Settings.hh>
#include <EquationSystem.hh>
#include <TetLinSpace.hh>
#include <math.h>

extern Settings *settings;

class EquationVel_cG1cG1:public EquationSystem {
private:

  int cell_no;

  real delta1, delta2, r;
  real C1, C2, K;

  real Re, dt, t;

  real Cs, S_abs, nu, nu_T;

  real normU;

  real x[4];
  real y[4];
  real z[4];

  real f1[4];
  real f2[4];
  real f3[4];

  real MASS, NL, NLSD, LAP, DIV, FORCE, FORCESD, PRE, PRESD;

  LocalField *u_fine;
  LocalField *p_fine;
  LocalField *u_coarse;
  LocalField *p_coarse;

  LocalField *upTS;
  LocalField *ppTS;
  LocalField *upNL;
  LocalField *ppNL;
  LocalField *unow;
  LocalField *pnow;

  LocalField *force;

  LocalField w11;
  LocalField w22;
  LocalField w33;

public:

  EquationVel_cG1cG1() : EquationSystem(3,3) {

    start_vector_component = 0;

    cell_no = -1;

    pi = KW_PI;

    C1 = 1.0;
    C2 = 1.0;
    K = 20.0;

    Cs = 0.1;
    
    settings->Get("variational multscale eddy viscosity",
		  &variational_multiscale_eddy_viscosity);
    settings->Get("smagorinsky eddy viscosity",
		  &smagorinsky_eddy_viscosity);
    

    variational_multiscale_eddy_viscosity = false;
    smagorinsky_eddy_viscosity            = false;
  }
  
  real InitLocalFields(FiniteElement *fe){

    upTS = new LocalField[3];
    for (int i=0;i<3;i++) upTS[i].AttachFiniteElement(fe);

    ppTS = new LocalField[1];
    ppTS[0].AttachFiniteElement(fe);

    upNL = new LocalField[3];
    for (int i=0;i<3;i++) upNL[i].AttachFiniteElement(fe);

    ppNL = new LocalField[1];
    ppNL[0].AttachFiniteElement(fe);

    unow = new LocalField[3];
    for (int i=0;i<3;i++) unow[i].AttachFiniteElement(fe);

    pnow = new LocalField[1];
    pnow[0].AttachFiniteElement(fe);

    u_fine = new LocalField[3];
    for (int i=0;i<3;i++) u_fine[i].AttachFiniteElement(fe);

    p_fine = new LocalField[1];
    p_fine[0].AttachFiniteElement(fe);

    u_coarse = new LocalField[3];
    for (int i=0;i<3;i++) u_coarse[i].AttachFiniteElement(fe);

    p_coarse = new LocalField[1];
    p_coarse[0].AttachFiniteElement(fe);
    
    force = new LocalField[3];
    for (int i=0;i<3;i++) force[i].AttachFiniteElement(fe);

    w11.AttachFiniteElement(fe);
    w22.AttachFiniteElement(fe);
    w33.AttachFiniteElement(fe);
  }

  real DeleteLocalFields(){

    delete [] upTS;
    delete [] ppTS;
    delete [] upNL;
    delete [] ppNL;
    delete [] unow;
    delete [] pnow;

    delete [] u_fine;
    delete [] p_fine;
    delete [] u_coarse;
    delete [] p_coarse;
    
    delete [] force;
  }

  real IntegrateLHS(ShapeFunction *u, ShapeFunction *v){

    cell_no = u->GetCellNo();

    if ( (cell_no == 0) && u[0].Active() && v[0].Active() && (u[0].GetDof() == 0) && (v[0].GetDof() == 0) ) 
      InitLocalFields(u->GetFiniteElement());

    if ( u[0].Active() && v[0].Active() && (u[0].GetDof() == 0) && (v[0].GetDof() == 0) ) {

      h = 2.0 * u->GetCircumRadius();
      
      for (int i=0;i<3;i++) upTS[i].Refill(cell_no,i,coeff[0]);
      //ppTS[0].Refill(cell_no,0,coeff[1]);

      for (int i=0;i<3;i++) upNL[i].Refill(cell_no,i,coeff[2]);
      //ppNL[0].Refill(cell_no,0,coeff[3]);

      for (int i=0;i<3;i++) unow[i].Refill(cell_no,i,coeff[4]);
      //pnow[0].Refill(cell_no,0,coeff[5]);

      dt = coeff[6]->GetConstantValue();
      //t = coeff[7]->GetConstantValue();
      Re = coeff[8]->GetConstantValue();

      if (variational_multiscale_eddy_viscosity) {
	for (int i=0;i<3;i++) u_coarse[i].Refill(cell_no,i,coeff[9]);
	//p_coarse[0].Refill(cell_no,0,coeff[10]);

	for (int i=0;i<3;i++) u_fine[i].Refill(cell_no,i,coeff[11]);
	//p_fine[0].Refill(cell_no,0,coeff[12]);
      }
      
      normU = sqrt(sqr(unow[0].GetMeanValue()) + sqr(unow[1].GetMeanValue()) + sqr(unow[2].GetMeanValue()));

      if (h * Re > 1.0) {
	//delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	delta1 = C1 * h;
      } else {
	delta1 = C1 * sqr(h);
      }
      if (h * Re > 1.0) {
	delta2 = C2 * h;
      } else {
	delta2 = C2 * sqr(h);
      }
    }

     
    MASS  = u[0]*v[0] + u[1]*v[1] + u[2]*v[2];

    MASS /= dt;


    LAP = u[0].dx()*v[0].dx() + u[0].dy()*v[0].dy() + u[0].dz()*v[0].dz() + 
	  u[1].dx()*v[1].dx() + u[1].dy()*v[1].dy() + u[1].dz()*v[1].dz() + 
	  u[2].dx()*v[2].dx() + u[2].dy()*v[2].dy() + u[2].dz()*v[2].dz();    

    LAP *= 0.5;


    w11 = upTS[0]; w11.Add(upNL[0]);
    w22 = upTS[1]; w22.Add(upNL[1]);
    w33 = upTS[2]; w33.Add(upNL[2]);
    
    NL    = u[0].dx() * (w11*v[0]) + u[0].dy() * (w22*v[0]) + u[0].dz() * (w33*v[0]) + 
            u[1].dx() * (w11*v[1]) + u[1].dy() * (w22*v[1]) + u[1].dz() * (w33*v[1]) + 
            u[2].dx() * (w11*v[2]) + u[2].dy() * (w22*v[2]) + u[2].dz() * (w33*v[2]);

    NL *= 0.25;


    NLSD  =  u[0].dx()*v[0].dx()*(w11*w11) + u[0].dx()*v[0].dy()*(w11*w22) + u[0].dx()*v[0].dz()*(w11*w33) + 
             u[0].dy()*v[0].dx()*(w22*w11) + u[0].dy()*v[0].dy()*(w22*w22) + u[0].dy()*v[0].dz()*(w22*w33) +
             u[0].dz()*v[0].dx()*(w33*w11) + u[0].dz()*v[0].dy()*(w33*w22) + u[0].dz()*v[0].dz()*(w33*w33) +

             u[1].dx()*v[1].dx()*(w11*w11) + u[1].dx()*v[1].dy()*(w11*w22) + u[1].dx()*v[1].dz()*(w11*w33) +
             u[1].dy()*v[1].dx()*(w22*w11) + u[1].dy()*v[1].dy()*(w22*w22) + u[1].dy()*v[1].dz()*(w22*w33) +
             u[1].dz()*v[1].dx()*(w33*w11) + u[1].dz()*v[1].dy()*(w33*w22) + u[1].dz()*v[1].dz()*(w33*w33) +

             u[2].dx()*v[2].dx()*(w11*w11) + u[2].dx()*v[2].dy()*(w11*w22) + u[2].dx()*v[2].dz()*(w11*w33) +
             u[2].dy()*v[2].dx()*(w22*w11) + u[2].dy()*v[2].dy()*(w22*w22) + u[2].dy()*v[2].dz()*(w22*w33) +
             u[2].dz()*v[2].dx()*(w33*w11) + u[2].dz()*v[2].dy()*(w33*w22) + u[2].dz()*v[2].dz()*(w33*w33);

    NLSD *= 0.125;


    DIV   = (u[0].dx() + u[1].dy() + u[2].dz()) * (v[0].dx() + v[1].dy() + v[2].dz());

    DIV *= 0.5;


    if (smagorinsky_eddy_viscosity) {

      S_abs = sqrt( sqr(upNL[0].dx()) + 
                    sqr(0.5*(upNL[0].dy()+upNL[1].dx())) + 
                    sqr(0.5*(upNL[0].dz()+upNL[2].dx())) + 
                    sqr(0.5*(upNL[1].dx()+upNL[0].dy())) + 
                    sqr(upNL[1].dy()) + 
                    sqr(0.5*(upNL[1].dz()+upNL[2].dy())) + 
                    sqr(0.5*(upNL[2].dx()+upNL[0].dz())) + 
                    sqr(0.5*(upNL[2].dy()+upNL[1].dz())) + 
                    sqr(upNL[2].dz()) );
 
      nu_T = pow(Cs * h, 2.0) * S_abs;
      nu = (1.0 / Re) + nu_T;
    } else {
      nu = (1.0 / Re);
    }

  LAP *= nu;


    if (variational_multiscale_eddy_viscosity) {

      S_abs = sqrt( sqr(upNL[0].dx()) + 
                    sqr(0.5*(upNL[0].dy()+upNL[1].dx())) + 
                    sqr(0.5*(upNL[0].dz()+upNL[2].dx())) + 
                    sqr(0.5*(upNL[1].dx()+upNL[0].dy())) + 
                    sqr(upNL[1].dy()) + 
                    sqr(0.5*(upNL[1].dz()+upNL[2].dy())) + 
                    sqr(0.5*(upNL[2].dx()+upNL[0].dz())) + 
                    sqr(0.5*(upNL[2].dy()+upNL[1].dz())) + 
                    sqr(upNL[2].dz()) );
 
      nu_T = pow(Cs * h, 2.0) * S_abs;


      LAP += nu_T * ( (u[0].dx() - u_coarse[0].dx()) * v[0].dx() +  
                      (u[0].dy() - u_coarse[0].dy()) * v[0].dy() +  
                      (u[0].dz() - u_coarse[0].dz()) * v[0].dz() +  
                      (u[1].dx() - u_coarse[1].dx()) * v[1].dx() +  
                      (u[1].dy() - u_coarse[1].dy()) * v[1].dy() +  
                      (u[1].dz() - u_coarse[1].dz()) * v[1].dz() +  
                      (u[2].dx() - u_coarse[2].dx()) * v[2].dx() +  
                      (u[2].dy() - u_coarse[2].dy()) * v[2].dy() +  
                      (u[2].dz() - u_coarse[2].dz()) * v[2].dz() );  
      
    }
  
  
  if ( (cell_no >= u[0].GetNoCells()-1) && u[2].Active() && v[2].Active() && (u[2].GetDof() == 3) && (v[2].GetDof() == 3) ) 
    DeleteLocalFields();
  
  /*
    display->Message(0,"LAP = %f",LAP);
    display->Message(0,"NL = %f",NL);
    display->Message(0,"NLSD = %f",NLSD);
    display->Message(0,"DIV = %f",DIV);
    display->Message(0,"delta1 = %f",delta1);
    display->Message(0,"delta2 = %f",delta2);
  */

    //if ( fabs(u[0].dx()*v[0].dx()*(w11*w11)) > 0.0 ) display->Message(0,"NL = %f, %f",u[0].dx()*v[0].dx()*(w11*w11),NL);

    return(
	   
	   MASS + LAP + NL + delta1*NLSD + delta2*DIV
	   
	   );
    
  }
  
  //--------------------------------------------------------------------------------

  real IntegrateRHS(ShapeFunction *v){

    cell_no = v->GetCellNo();

    if ( (cell_no == 0) && v[0].Active() && (v[0].GetDof() == 0) ) 
      InitLocalFields(v->GetFiniteElement());

    if ( v[0].Active() && (v[0].GetDof() == 0) ) {

      h = 2.0 * v->GetCircumRadius();
      
      for (int i=0;i<3;i++) upTS[i].Refill(cell_no,i,coeff[0]);
      //ppTS[0].Refill(cell_no,0,coeff[1]);

      for (int i=0;i<3;i++) upNL[i].Refill(cell_no,i,coeff[2]);
      //ppNL[0].Refill(cell_no,0,coeff[3]);

      for (int i=0;i<3;i++) unow[i].Refill(cell_no,i,coeff[4]);
      pnow[0].Refill(cell_no,0,coeff[5]);

      dt = coeff[6]->GetConstantValue();
      //t = coeff[7]->GetConstantValue();
      Re = coeff[8]->GetConstantValue();

      if (variational_multiscale_eddy_viscosity) {
	for (int i=0;i<3;i++) u_coarse[i].Refill(cell_no,i,coeff[9]);
	//p_coarse[0].Refill(cell_no,0,coeff[10]);

	for (int i=0;i<3;i++) u_fine[i].Refill(cell_no,i,coeff[11]);
	//p_fine[0].Refill(cell_no,0,coeff[12]);
      }
      
      for (int i=0;i<4;i++) x[i] = v->GetCoord(i,0);
      for (int i=0;i<4;i++) y[i] = v->GetCoord(i,1);
      for (int i=0;i<4;i++) z[i] = v->GetCoord(i,2);
      
      //for (int i=0;i<4;i++) f1[i] = (1.0 / Re) * 32.0 * (y[i] * (1.0 - y[i]) + z[i] * (1.0 - z[i])); 
      for (int i=0;i<4;i++) f1[i] = 0.0;
      
      //for (int i=0;i<4;i++) f2[i] = 0.001 * sin(10.0 * pi * x[i]);
      for (int i=0;i<4;i++) f2[i] = 0.0;
      
      //for (int i=0;i<4;i++) f3[i] = 0.001 * sin(10.0 * pi * x[i]);
      for (int i=0;i<4;i++) f3[i] = 0.0;
      
      force[0].Refill(cell_no,f1);
      force[1].Refill(cell_no,f2);
      force[2].Refill(cell_no,f3);

      normU = sqrt(sqr(unow[0].GetMeanValue()) + sqr(unow[1].GetMeanValue()) + sqr(unow[2].GetMeanValue()));

      if (h * Re > 1.0) {
	//delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	delta1 = C1 * h;
      } else {
	delta1 = C1 * sqr(h);
      }
      if (h * Re > 1.0) {
	delta2 = C2 * h;
      } else {
	delta2 = C2 * sqr(h);
      }
    }


    MASS  = upTS[0]*v[0] + upTS[1]*v[1] + upTS[2]*v[2];

    MASS /= dt;


    LAP = upTS[0].dx()*v[0].dx() + upTS[0].dy()*v[0].dy() + upTS[0].dz()*v[0].dz() + 
	  upTS[1].dx()*v[1].dx() + upTS[1].dy()*v[1].dy() + upTS[1].dz()*v[1].dz() + 
	  upTS[2].dx()*v[2].dx() + upTS[2].dy()*v[2].dy() + upTS[2].dz()*v[2].dz();

    LAP *= 0.5;


    w11 = upTS[0]; w11.Add(upNL[0]);
    w22 = upTS[1]; w22.Add(upNL[1]);
    w33 = upTS[2]; w33.Add(upNL[2]);
    
    NL    = upTS[0].dx() * (w11*v[0]) + upTS[0].dy() * (w22*v[0]) + upTS[0].dz() * (w33*v[0]) + 
            upTS[1].dx() * (w11*v[1]) + upTS[1].dy() * (w22*v[1]) + upTS[1].dz() * (w33*v[1]) + 
            upTS[2].dx() * (w11*v[2]) + upTS[2].dy() * (w22*v[2]) + upTS[2].dz() * (w33*v[2]);

    NL *= 0.25;


    NLSD  =  upTS[0].dx()*v[0].dx()*(w11*w11) + upTS[0].dx()*v[0].dy()*(w11*w22) + upTS[0].dx()*v[0].dz()*(w11*w33) + 
             upTS[0].dy()*v[0].dx()*(w22*w11) + upTS[0].dy()*v[0].dy()*(w22*w22) + upTS[0].dy()*v[0].dz()*(w22*w33) +
             upTS[0].dz()*v[0].dx()*(w33*w11) + upTS[0].dz()*v[0].dy()*(w33*w22) + upTS[0].dz()*v[0].dz()*(w33*w33) +

             upTS[1].dx()*v[1].dx()*(w11*w11) + upTS[1].dx()*v[1].dy()*(w11*w22) + upTS[1].dx()*v[1].dz()*(w11*w33) +
             upTS[1].dy()*v[1].dx()*(w22*w11) + upTS[1].dy()*v[1].dy()*(w22*w22) + upTS[1].dy()*v[1].dz()*(w22*w33) +
             upTS[1].dz()*v[1].dx()*(w33*w11) + upTS[1].dz()*v[1].dy()*(w33*w22) + upTS[1].dz()*v[1].dz()*(w33*w33) +

             upTS[2].dx()*v[2].dx()*(w11*w11) + upTS[2].dx()*v[2].dy()*(w11*w22) + upTS[2].dx()*v[2].dz()*(w11*w33) +
             upTS[2].dy()*v[2].dx()*(w22*w11) + upTS[2].dy()*v[2].dy()*(w22*w22) + upTS[2].dy()*v[2].dz()*(w22*w33) +
             upTS[2].dz()*v[2].dx()*(w33*w11) + upTS[2].dz()*v[2].dy()*(w33*w22) + upTS[2].dz()*v[2].dz()*(w33*w33);

    NLSD *= 0.125;


    DIV   = (upTS[0].dx() + upTS[1].dy() + upTS[2].dz()) * (v[0].dx() + v[1].dy() + v[2].dz());
            
    DIV *= 0.5;

    if (smagorinsky_eddy_viscosity) {

      S_abs = sqrt( sqr(upNL[0].dx()) + 
                    sqr(0.5*(upNL[0].dy()+upNL[1].dx())) + 
                    sqr(0.5*(upNL[0].dz()+upNL[2].dx())) + 
                    sqr(0.5*(upNL[1].dx()+upNL[0].dy())) + 
                    sqr(upNL[1].dy()) + 
                    sqr(0.5*(upNL[1].dz()+upNL[2].dy())) + 
                    sqr(0.5*(upNL[2].dx()+upNL[0].dz())) + 
                    sqr(0.5*(upNL[2].dy()+upNL[1].dz())) + 
                    sqr(upNL[2].dz()) );
 
      nu_T = pow(Cs * h, 2.0) * S_abs;
      nu = (1.0 / Re) + nu_T;
    } else {
      nu = (1.0 / Re);
    }
    
    
    LAP *= nu;


    PRE   = pnow[0] * (v[0].dx() + v[1].dy() + v[2].dz());

    PRESD = pnow[0].dx()*v[0].dx()*w11 + 
            pnow[0].dx()*v[0].dy()*w22 + 
            pnow[0].dx()*v[0].dz()*w33 + 
            pnow[0].dy()*v[1].dx()*w11 + 
            pnow[0].dy()*v[1].dy()*w22 + 
            pnow[0].dy()*v[1].dz()*w33 + 
            pnow[0].dz()*v[2].dx()*w11 + 
            pnow[0].dz()*v[2].dy()*w22 + 
            pnow[0].dz()*v[2].dz()*w33;
 
    PRESD *= 0.5;

    FORCE = force[0]*v[0] + force[1]*v[1] + force[2]*v[2];

    FORCESD = v[0].dx()*(force[0]*w11) + 
              v[0].dy()*(force[0]*w22) + 
              v[0].dz()*(force[0]*w33) + 
              v[1].dx()*(force[1]*w11) + 
              v[1].dy()*(force[1]*w22) + 
              v[1].dz()*(force[1]*w33) + 
              v[2].dx()*(force[2]*w11) + 
              v[2].dy()*(force[2]*w22) + 
              v[2].dz()*(force[2]*w33);
    
    FORCESD *= 0.5;
    
    if (smagorinsky_eddy_viscosity) {
      
      S_abs = sqrt( sqr(upNL[0].dx()) + 
                    sqr(0.5*(upNL[0].dy()+upNL[1].dx())) + 
                    sqr(0.5*(upNL[0].dz()+upNL[2].dx())) + 
                    sqr(0.5*(upNL[1].dx()+upNL[0].dy())) + 
                    sqr(upNL[1].dy()) + 
                    sqr(0.5*(upNL[1].dz()+upNL[2].dy())) + 
                    sqr(0.5*(upNL[2].dx()+upNL[0].dz())) + 
                    sqr(0.5*(upNL[2].dy()+upNL[1].dz())) + 
                    sqr(upNL[2].dz()) );
 
      nu_T = pow(Cs * h, 2.0) * S_abs;
      nu = (1.0 / Re) + nu_T;
    } else {
      nu = (1.0 / Re);
    }
    
    
    LAP *= nu;

    if ( (cell_no >= v[0].GetNoCells()-1) && v[2].Active() && (v[2].GetDof() == 3) ) 
      DeleteLocalFields();

    /*
    display->Message(0,"Re = %f",Re);
    display->Message(0,"MASS = %f",MASS);
    display->Message(0,"LAP = %f",LAP);
    display->Message(0,"NLSD = %f",NLSD);
    display->Message(0,"FORCESD = %f",FORCESD);
    display->Message(0,"delta = %f",delta1);
    display->Message(0,"FORCE = %f",FORCE);
    display->Message(0,"NL = %f",NL);
    display->Message(0,"DIV = %f",DIV);
    */

    return(
	   
	   MASS + FORCE - LAP - NL + PRE + delta1*(FORCESD-NLSD-PRESD) + delta2*DIV
	   
           );
    
  }

  
};

#endif
