// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "StatProblemDualPressure.hh"

//-----------------------------------------------------------------------------
StatProblemDualPressure::StatProblemDualPressure(Grid *grid) : StatProblem(grid) 
{
  noeq             = 1;

  // Initialize Fields to attach to Equation
  u1_prev_TS.AttachGrid(grid);
  u2_prev_TS.AttachGrid(grid);
  u3_prev_TS.AttachGrid(grid);
  p_prev_TS.AttachGrid(grid);
  
  U1.AttachGrid(grid);
  U2.AttachGrid(grid);
  U3.AttachGrid(grid);
  
  u1.AttachGrid(grid);
  u2.AttachGrid(grid);
  u3.AttachGrid(grid);
  p.AttachGrid(grid);
  
  dt.AttachGrid(grid);
  t.AttachGrid(grid);
  Re.AttachGrid(grid);
  
  settings->Get("bnd x0", &bnd_x0);
  settings->Get("bnd x1", &bnd_x1);
  settings->Get("bnd y0", &bnd_y0);
  settings->Get("bnd y1", &bnd_y1);
  settings->Get("bnd z0", &bnd_z0);
  settings->Get("bnd z1", &bnd_z1);
 
  settings->Get("reynolds number", &ReNo);

  settings->Get("periodic bc x", &periodic_bc_x);
  settings->Get("periodic bc y", &periodic_bc_y);
  settings->Get("periodic bc z", &periodic_bc_z);

  settings->Get("outflow bc x0", &outflow_bc_x0);
  settings->Get("outflow bc x1", &outflow_bc_x1);

  SetDomain();
}
//-----------------------------------------------------------------------------
int StatProblemDualPressure::Timestep( MV_ColMat<real> &val, real time, real deltat )
{
  Scalar *temp;
  int i,ierr;
  
  Vec u;
  Vec b;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );
      
  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  u3_prev_TS.InitByNodes(uk_prev_TS,2);
  //p_prev_TS.InitByNodes(uk_prev_TS,3);
  
  U1.InitByNodes(Uprimal,0);
  U2.InitByNodes(Uprimal,1);
  U3.InitByNodes(Uprimal,2);
  
  uk = &val;
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  u3.InitByNodes(uk,2);
  p.InitByNodes(uk,3);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);
  
  real int_p;
  p.Integrate( int_p );
  display->Status(2,"Integral of p = %f",int_p);

  MV_Vector<int> DirichletIdx(0); // Should be of size 0
  MV_Vector<real> DirichletVal(0); // Should be of size 0
  
  MV_Vector<int> perList;
  
  // FIXME: Temporary until we get the grid class working again
  // NOTE! Commented lines containing 'MGU' below!
  //MultigridUtilities MGU;
  bool usingMG = false;

  if ( periodic_bc_x || periodic_bc_y ||periodic_bc_z ) usingMG = false;
  
  if ( !A_assembled ){
    
    cout << "Assemble matrix A and vector b." << endl;
    Scalar one = 1.0;
    Mat Atmp;
    A_mat.newsize(compGrid+1,1);	
    int startGrid;
    if ( usingMG ){
      cout << "Using Multigrid Solver" << endl; 
      GridProjections.newsize( compGrid,1);
      startGrid = 0;
    }
    else{
      startGrid = compGrid;
    } 
    
    for ( i=startGrid; i < compGrid+1; i++ ){
      discrete->createSeqRHS( &u, &b, grid, noeq );
      if ( exact_integration ){
		  if ( periodic_bc_x || periodic_bc_y ||periodic_bc_z ){
			 if ( periodic_bc_x ) getPerList( bndx0, bndx1, grid, perList );
			 if ( periodic_bc_y ) getPerList( bndy0, bndy1, grid, perList );
			 if ( periodic_bc_z ) getPerList( bndz0, bndz1, grid, perList );
			 if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
				discrete->createSeqMatPeriodic( &Atmp, grid, noeq, perList );
				discrete->discretiseLHSIntExactPeriodicInitial( *equation, grid, Atmp, noeq, perList );
				discrete->discretiseRHSIntExactPeriodic( *equation, grid, b, noeq, perList );
			 }
		  } else{
			 discrete->createSeqMat( &Atmp, grid, noeq );
			 discrete->discretiseLHSIntExact( *equation, grid, Atmp, noeq );
			 discrete->discretiseRHSIntExact( *equation, grid, b, noeq );
		  }
      } else{
		  discrete->createSeqMat( &Atmp, grid, noeq );
		  discrete->discretiseLHS( *equation, *mapper, grid, *quadrule, *element, Atmp );
		  discrete->discretiseRHS( *equation, *mapper, grid, *quadrule, *element, b );
      }
      discrete->dropZeroElements(Atmp, A_mat(i,0), 1.0e-12);
      if ( outflow_bc_x0 ){
		  if ( i == compGrid ){
			 ierr = outflowBC( A_mat(i,0), b, bndx0, u1, ReNo, grid, 1, 0, DirichletIdx, DirichletVal );
		  } else{
			 ierr = dirichletBC( A_mat(i,0), b, bndx0, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
		  }
      }
      if ( outflow_bc_x1 ){
		  if ( i == compGrid ){
			 ierr = outflowBC( A_mat(i,0), b, bndx1, u1, ReNo, grid, 1, 0, DirichletIdx, DirichletVal );
		  } else{
			 ierr = dirichletBC( A_mat(i,0), b, bndx1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
		  }
      }
      if ( periodic_bc_x ) ierr = dirichletBC( A_mat(i,0), b, bndx1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
      if ( periodic_bc_y ) ierr = dirichletBC( A_mat(i,0), b, bndy1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
      if ( periodic_bc_z ) ierr = dirichletBC( A_mat(i,0), b, bndz1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );

		// FIXME: Here
		//if ( (i > 0) && usingMG ) {
		//  MGU.constructProjectionMatrix( (*GRID)(i-1), grid, GridProjections(i-1,0), grid );
		// }
      
    }
    
    ierr = MatDestroy(Atmp); CHKERRQ(ierr);
    A = A_mat(compGrid,0);
    
    A_assembled = true;
    
  }
  else{
    discrete->createSeqRHS( &u, &b, grid, noeq );
    cout << "Assemble vector b" << endl;
    if ( exact_integration ){
      if ( periodic_bc_x ) getPerList( bndx0, bndx1, grid, perList );
      if ( periodic_bc_y ) getPerList( bndy0, bndy1, grid, perList );
      if ( periodic_bc_z ) getPerList( bndz0, bndz1, grid, perList );
      if ( periodic_bc_x ||  periodic_bc_y ||  periodic_bc_z ){
		  discrete->discretiseRHSIntExactPeriodic( *equation, grid, b, noeq, perList );
      } else{
		  discrete->discretiseRHSIntExact( *equation, grid, b, noeq );
      }
    } else{
      discrete->discretiseRHS( *equation, *mapper, grid, *quadrule, *element, b );
    }
  
    if ( outflow_bc_x0 ){
      ierr = outflowBC( b, bndx0, u1, ReNo, grid, 1, 0, DirichletIdx, DirichletVal );
      //ierr = dirichletBC( b, bndx0, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
    }
    if ( outflow_bc_x1 ){
      ierr = outflowBC( b, bndx1, u1, ReNo, grid, 1, 0, DirichletIdx, DirichletVal );
      //ierr = dirichletBC( b, bndx1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
    }
    if ( periodic_bc_x ) ierr = dirichletBC( b, bndx1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
    if ( periodic_bc_y ) ierr = dirichletBC( b, bndy1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
    if ( periodic_bc_z ) ierr = dirichletBC( b, bndz1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
  }
  
  real norm;
  ierr = VecNorm( b, NORM_2, &norm ); CHKERRQ( ierr ); 
  
  MV_Vector<real> uprev( noNodes );
  for ( i=0; i < noNodes; i++ ) uprev(i) = val(i,3);
  
  MV_Vector<int> idx( noNodes );
  for ( i=0; i < noNodes; i++ ) idx(i) = i;
  
  ierr = VecSetValues(u,noNodes,&idx(0),&uprev(0),INSERT_VALUES); CHKERRQ( ierr ); 
  
  cout << "Solve algebraic system of equations" << endl;

  // FIXME: And here
  //if ( usingMG == 1 ){
  //  ierr = MGU.solveMG(A_mat,GridProjections,b,u,compGrid,1); CHKERRQ(ierr); 
  // }
  //else{
  discrete->solve( A, u, b );
  //}
  
  ierr = VecGetArray(u, &temp);
  for ( i=0; i < noNodes; i++ ){
    if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
      val(i,3) = temp[perList(i)];
    } else{
      val(i,3) = temp[i];
    }	  
  }
  ierr = VecRestoreArray(u, &temp); CHKERRQ( ierr );
  
  ierr = VecDestroy(u); CHKERRQ(ierr);
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
  
}
//-----------------------------------------------------------------------------
int StatProblemDualPressure::ComputeDiscreteResidual( real time, real deltat, real& norm_2, real& norm_inf )
{
  Scalar *temp;
  int ierr;
  
  Vec u;
  Vec b;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );
      
  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  u3_prev_TS.InitByNodes(uk_prev_TS,2);
  //p_prev_TS.InitByNodes(uk_prev_TS,3);
  
  U1.InitByNodes(Uprimal,0);
  U2.InitByNodes(Uprimal,1);
  U3.InitByNodes(Uprimal,2);
  
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  u3.InitByNodes(uk,2);
  p.InitByNodes(uk,3);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);
  
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, grid, perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, grid, perList );
  if ( periodic_bc_z ) getPerList( bndz0, bndz1, grid, perList );
  
  MV_Vector<int> DirichletIdx(0); // Should be of size 0
  MV_Vector<real> DirichletVal(0); // Should be of size 0
  
  discrete->createSeqRHS( &u, &b, grid, noeq );
  if ( exact_integration ){
    if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
      discrete->discretiseRHSIntExactPeriodic( *equation, grid, b, noeq, perList );
    } else{
      discrete->discretiseRHSIntExact( *equation, grid, b, noeq );
    }
  } else{
    discrete->discretiseRHS( *equation, *mapper, grid, *quadrule, *element, b );
  }
  if ( outflow_bc_x0 ){
    ierr = outflowBC( b, bndx0, u1, ReNo, grid, 1, 0, DirichletIdx, DirichletVal );
    //ierr = dirichletBC( b, bndx0, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
  }
  if ( outflow_bc_x1 ){
    ierr = outflowBC( b, bndx1, u1, ReNo, grid, 1, 0, DirichletIdx, DirichletVal );
    //ierr = dirichletBC( b, bndx1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
  }
  if ( periodic_bc_x ) ierr = dirichletBC( b, bndx1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
  if ( periodic_bc_y ) ierr = dirichletBC( b, bndy1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
  if ( periodic_bc_z ) ierr = dirichletBC( b, bndz1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
  
  int i;
  MV_Vector<real> uvec( noNodes );
  for ( i=0; i < noNodes; i++ ) uvec(i) = p.GetNodalValue(i);

  MV_Vector<int> idx( noNodes );
  for ( i=0; i < noNodes; i++ ) idx(i) = i;

  ierr = VecSetValues(u,noNodes,&idx(0),&uvec(0),INSERT_VALUES); CHKERRQ( ierr ); 
  
  if ( periodic_bc_x ) ierr = dirichletBC( u, bndx1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
  if ( periodic_bc_y ) ierr = dirichletBC( u, bndy1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
  if ( periodic_bc_z ) ierr = dirichletBC( u, bndz1, 0.0, grid, 1, 0, DirichletIdx, DirichletVal );
  
  if ( debug_mode ) cout << "Compute residual" << endl;
  Vec res;
  VecDuplicate( b, &res);
  Scalar minus = - 1.0;
  VecScale( &minus, b );
  MatMultAdd( A_mat(compGrid,0), u, b, res );
  ierr = VecNorm( res, NORM_2, &norm_2 );
  ierr = VecNorm( res, NORM_INFINITY, &norm_inf );
  
  ierr = VecDestroy(res); CHKERRQ(ierr);
  ierr = VecDestroy(u); CHKERRQ(ierr);
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
}
//-----------------------------------------------------------------------------
int StatProblemDualPressure::InitEquationData( real time, real deltat )
{      
  data.newsize(15);
  
  data(0)  = &u1_prev_TS;
  data(1)  = &u2_prev_TS;
  data(2)  = &u3_prev_TS;
  data(3)  = &p_prev_TS;
  
  data(4)  = &U1;
  data(5)  = &U2;
  data(6)  = &U3;
  //data(7)  = &p_prev_NL;
  
  data(8) = &u1;
  data(9) = &u2;
  data(10) = &u3;
  data(11) = &p;
  
  data(12) = &dt;
  data(13) = &t;
  data(14) = &Re;
  
  equation->attachField(data);
}
//-----------------------------------------------------------------------------
int StatProblemDualPressure::SetDomain()
{
  bndx0.newsize(nsd,2);
  bndx1.newsize(nsd,2);
  bndy0.newsize(nsd,2);
  bndy1.newsize(nsd,2);
  bndz0.newsize(nsd,2);
  bndz1.newsize(nsd,2);
  
  bndx0(0,0) = bnd_x0;
  bndx0(0,1) = bnd_x0;
  bndx0(1,0) = bnd_y0;
  bndx0(1,1) = bnd_y1;
  bndx0(2,0) = bnd_z0;
  bndx0(2,1) = bnd_z1;
  
  bndx1(0,0) = bnd_x1;
  bndx1(0,1) = bnd_x1;
  bndx1(1,0) = bnd_y0;
  bndx1(1,1) = bnd_y1;
  bndx1(2,0) = bnd_z0;
  bndx1(2,1) = bnd_z1;
  
  bndy0(0,0) = bnd_x0;
  bndy0(0,1) = bnd_x1;
  bndy0(1,0) = bnd_y0;
  bndy0(1,1) = bnd_y0;
  bndy0(2,0) = bnd_z0;
  bndy0(2,1) = bnd_z1;
  
  bndy1(0,0) = bnd_x0;
  bndy1(0,1) = bnd_x1;
  bndy1(1,0) = bnd_y1;
  bndy1(1,1) = bnd_y1;
  bndy1(2,0) = bnd_z0;
  bndy1(2,1) = bnd_z1;
  
  bndz0(0,0) = bnd_x0;
  bndz0(0,1) = bnd_x1;
  bndz0(1,0) = bnd_y0;
  bndz0(1,1) = bnd_y1;
  bndz0(2,0) = bnd_z0;
  bndz0(2,1) = bnd_z0;
  
  bndz1(0,0) = bnd_x0;
  bndz1(0,1) = bnd_x1;
  bndz1(1,0) = bnd_y0;
  bndz1(1,1) = bnd_y1;
  bndz1(2,0) = bnd_z1;
  bndz1(2,1) = bnd_z1;
  
  return 0;
}
//-----------------------------------------------------------------------------  
