// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "StatProblemDualVelocity.hh"

//----------------------------------------------------------------------------------------------------
StatProblemDualVelocity::StatProblemDualVelocity(Grid *grid) : StatProblem(grid)
{ 
  noeq = 3;

  re_assemble_matrix = true; 
  
  // Initialize Fields to attach to Equation
  u1_prev_TS.AttachGrid(grid);
  u2_prev_TS.AttachGrid(grid);
  u3_prev_TS.AttachGrid(grid);
  p_prev_TS.AttachGrid(grid);
  
  U1.AttachGrid(grid);
  U2.AttachGrid(grid);
  U3.AttachGrid(grid);
  
  u1.AttachGrid(grid);
  u2.AttachGrid(grid);
  u3.AttachGrid(grid);
  p.AttachGrid(grid);
  
  dt.AttachGrid(grid);
  t.AttachGrid(grid);
  Re.AttachGrid(grid);
  
  settings->Get("bnd x0", &bnd_x0);
  settings->Get("bnd x1", &bnd_x1);
  settings->Get("bnd y0", &bnd_y0);
  settings->Get("bnd y1", &bnd_y1);
  settings->Get("bnd z0", &bnd_z0);
  settings->Get("bnd z1", &bnd_z1);
 
  settings->Get("reynolds number", &ReNo);

  settings->Get("driven cavity", &driven_cavity);
  settings->Get("backward facing step", &backward_facing_step);
  settings->Get("bluff body", &bluff_body);
  settings->Get("jet", &jet);
  settings->Get("LES benchmark", &LES_benchmark);

  settings->Get("slip bc x0", &slip_bc_x0);
  settings->Get("slip bc x1", &slip_bc_x1);
  settings->Get("slip bc y0", &slip_bc_y0);
  settings->Get("slip bc y1", &slip_bc_y1);
  settings->Get("slip bc z0", &slip_bc_z0);
  settings->Get("slip bc z1", &slip_bc_z1);

  settings->Get("no slip bc x0", &no_slip_bc_x0);
  settings->Get("no slip bc x1", &no_slip_bc_x1);
  settings->Get("no slip bc y0", &no_slip_bc_y0);
  settings->Get("no slip bc y1", &no_slip_bc_y1);
  settings->Get("no slip bc z0", &no_slip_bc_z0);
  settings->Get("no slip bc z1", &no_slip_bc_z1);

  settings->Get("periodic bc x", &periodic_bc_x);
  settings->Get("periodic bc y", &periodic_bc_y);
  settings->Get("periodic bc z", &periodic_bc_z);

  settings->Get("inflow bc x0", &inflow_bc_x0);
  settings->Get("outflow bc x0", &outflow_bc_x0);
  settings->Get("outflow bc x1", &outflow_bc_x1);

  settings->Get("couette bc y", &couette_bc_y);

  settings->Get("driven cavity bc x0", &driven_cavity_bc_x0);
  settings->Get("driven cavity bc y1", &driven_cavity_bc_y1);

  settings->Get("dual drag bc", &dual_drag_bc);

  SetDomain();
}
//----------------------------------------------------------------------------------------------------
int StatProblemDualVelocity::Timestep( MV_ColMat<real>& val, real time, real deltat )
{
  Scalar *temp;
  int ierr;
  
  Vec u;
  Vec b;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );
      
  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  u3_prev_TS.InitByNodes(uk_prev_TS,2);
  //p_prev_TS.InitByNodes(uk_prev_TS,3);
  
  U1.InitByNodes(Uprimal,0);
  U2.InitByNodes(Uprimal,1);
  U3.InitByNodes(Uprimal,2);
  
  uk = &val;
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  u3.InitByNodes(uk,2);
  p.InitByNodes(uk,3);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  // get master-slave info if periodic bc
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, grid, perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, grid, perList );
  if ( periodic_bc_z ) getPerList( bndz0, bndz1, grid, perList );

  // assemble matrix if not assembled
  if ( !A_assembled ){
    if ( debug_mode ) cout << "Assemble matrix A" << endl;
    // check if exact integration / quadrature is used
    if ( exact_integration ){
      // check if periodic bc
      if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
	discrete->createSeqMatPeriodic( &A, grid, noeq, perList );
	discrete->discretiseLHSIntExactPeriodicInitial( *equation, grid, A, noeq, perList );
      } else{
	discrete->createSeqMat( &A, grid, noeq );
	discrete->discretiseLHSIntExact( *equation, grid, A, noeq );
      }
    } else{
      discrete->createSeqMat( &A, grid, noeq );
      discrete->discretiseLHS( *equation, *mapper, grid, *quadrule, *element, A );
    }
    // if assembled, check if to reassemble
  } else if ( A_assembled && re_assemble_matrix ){
    if (debug_mode ) cout << "Assemble matrix A" << endl;
    ierr = MatZeroEntries(A);
    // check if exact integration / quadrature is used
    if ( exact_integration ){
      // check if periodic bc
      if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
	discrete->discretiseLHSIntExactPeriodic( *equation, grid, A, noeq, perList );
      } else{
	discrete->discretiseLHSIntExact( *equation, grid, A, noeq );
      }
    } else{
      discrete->discretiseLHS( *equation, *mapper, grid, *quadrule, *element, A );
    }
  }
  
  if ( debug_mode ) cout << "Assemble vector b" << endl;
  discrete->createSeqRHS( &u, &b, grid, noeq );
  if ( exact_integration ){
    // check if periodic bc
    if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
      discrete->discretiseRHSIntExactPeriodic( *equation, grid, b, noeq, perList );
    } else{
      discrete->discretiseRHSIntExact( *equation, grid, b, noeq );
    }
  } else{
    discrete->discretiseRHS( *equation, *mapper, grid, *quadrule, *element, b );
  }

  // compute norm of rhs 
  real norm;
  ierr = VecNorm( b, NORM_2, &norm );
  cout << "Load norm: " << norm << endl<<flush;
  
  // set dirichlet bc
  SetBoundaryConditions(A,b);
      
  // compute norm of rhs 
  ierr = VecNorm( b, NORM_2, &norm );
  cout << "Load norm: " << norm << " (with BC)" << endl << flush;
  
  // set start solution to solution from previous iteration 
  MV_Vector<real> uprev( noeq * noNodes );
  int i;
  for ( i=0; i < noNodes; i++ ){
    uprev(i*noeq)     = val(i,0);
    uprev(i*noeq + 1) = val(i,1);
    uprev(i*noeq + 2) = val(i,2);
  }
  MV_Vector<int> idx( noeq * noNodes );
  for ( i=0; i < noeq * noNodes; i++ ) idx(i) = i;
  
  ierr = VecSetValues(u,noeq*noNodes,&idx(0),&uprev(0),INSERT_VALUES); CHKERRQ( ierr ); 

  // solve discrete system 
  if ( debug_mode ) cout << "Solve algebraic system of equations" << endl;
  discrete->solve( A, u, b );
  
  // compute norm of solution
  VecNorm(u,NORM_INFINITY,&norm);
  if ( debug_mode ) cout << "NORM of u = " << norm << endl;

  // update solution vector
  ierr = VecGetArray(u, &temp); CHKERRQ( ierr );
  int j;
  for ( i=0; i < noNodes; i++ ){
    for ( j=0; j < nsd; j++ ){ 
      if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
	val(i,j) = temp[j+perList(i)*noeq];
      } else{
	val(i,j) = temp[j+i*noeq];
      }
    }
  }
  ierr = VecRestoreArray(u, &temp); CHKERRQ( ierr );

  // destroy solution variables 
  ierr = VecDestroy(u); CHKERRQ(ierr);
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemDualVelocity::ComputeDiscreteResidual( real time, real deltat, real& norm_2, real& norm_inf )
{
  int ierr;
  
  Vec u;
  Vec b;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );
      
  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  u3_prev_TS.InitByNodes(uk_prev_TS,2);
  //p_prev_TS.InitByNodes(uk_prev_TS,3);
  
  U1.InitByNodes(Uprimal,0);
  U2.InitByNodes(Uprimal,1);
  U3.InitByNodes(Uprimal,2);
  
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  u3.InitByNodes(uk,2);
  p.InitByNodes(uk,3);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  // get master-slave info if periodic bc
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, grid, perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, grid, perList );
  if ( periodic_bc_z ) getPerList( bndz0, bndz1, grid, perList );
  
  if ( re_assemble_matrix ){
    if ( debug_mode ) cout << "Assemble matrix A" << endl;
    ierr = MatZeroEntries(A);
    if ( exact_integration ){
      if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
	discrete->discretiseLHSIntExactPeriodic( *equation, grid, A, noeq, perList );
      } else{
	discrete->discretiseLHSIntExact( *equation, grid, A, noeq );
      }
    } else{
      discrete->discretiseLHS( *equation, *mapper, grid, *quadrule, *element, A );
    }
  }
  
  cout << "Assemble vector b" << endl;
  discrete->createSeqRHS( &u, &b, grid, noeq );
  if ( exact_integration ){
    if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
      discrete->discretiseRHSIntExactPeriodic( *equation, grid, b, noeq, perList );
    } else{
      discrete->discretiseRHSIntExact( *equation, grid, b, noeq );
    }
  } else{
    discrete->discretiseRHS( *equation, *mapper, grid, *quadrule, *element, b );	
  }
  
  // set dirichlet bc
  SetBoundaryConditions(A,b);
  
  MV_Vector<real> uprev( noeq * noNodes );
  int i;
  for ( i=0; i < noNodes; i++ ){
    uprev(i*noeq)     = u1.GetNodalValue(i);
    uprev(i*noeq + 1) = u2.GetNodalValue(i);
    uprev(i*noeq + 2) = u3.GetNodalValue(i);
  }
  MV_Vector<int> idx( noeq * noNodes );
  for ( i=0; i < noeq * noNodes; i++ ) idx(i) = i;
  ierr = VecSetValues(u,noeq*noNodes,&idx(0),&uprev(0),INSERT_VALUES); CHKERRQ( ierr ); 
  
  if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
    MV_Vector<int> DirichletIdx(0);  // Should be of size 0
    MV_Vector<real> DirichletVal(0); // Should be of size 0
    ierr = dirichletBC( u, bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
    ierr = dirichletBC( u, bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
    ierr = dirichletBC( u, bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );		
  }
  
  Vec res;
  VecDuplicate( b, &res);
  Scalar minus = - 1.0;
  VecScale( &minus, b ); 
  MatMultAdd( A, u, b, res );
  ierr = VecNorm( res, NORM_2, &norm_2 );
  ierr = VecNorm( res, NORM_INFINITY, &norm_inf );
  
  ierr = VecDestroy(res); CHKERRQ(ierr);
  ierr = VecDestroy(u); CHKERRQ(ierr);
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemDualVelocity::InitEquationData( real time, real deltat )
{      
  data.newsize(15);
  
  data(0)  = &u1_prev_TS;
  data(1)  = &u2_prev_TS;
  data(2)  = &u3_prev_TS;
  data(3)  = &p_prev_TS;
  
  data(4)  = &U1;
  data(5)  = &U2;
  data(6)  = &U3;
  //data(7)  = &p_prev_NL;
  
  data(8) = &u1;
  data(9) = &u2;
  data(10) = &u3;
  data(11) = &p;
  
  data(12) = &dt;
  data(13) = &t;
  data(14) = &Re;
  
  equation->attachField(data);
}
//----------------------------------------------------------------------------------------------------
int StatProblemDualVelocity::SetBoundaryConditions( Mat& A, Vec& b )
{
  // Set Dirichlet BC
  MV_Vector<int> DirichletIdx(0); // Should be of size 0
  MV_Vector<real> DirichletVal(0); // Should be of size 0
  
  int ierr;
  
  if ( !re_assemble_matrix && A_assembled ){
    
    if ( backward_facing_step ){
      ierr = dirichletBC( b, bfs_bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, bfs_bndy1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }	  
    
    if ( jet ){
      ierr = dirichletBC_GRID( b, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( b, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( b, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    
    if ( bluff_body ){
      ierr = dirichletBC( b, ob_bndx0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }  
    
    if ( LES_benchmark ){
      ierr = dirichletBC( b, les_bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndx0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, les_bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, les_bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, les_bndz0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndz0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, les_bndz1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }  
    
    if ( couette_bc_y ){
      ierr = dirichletBC( b, bndy1, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, bndy0, -1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0,  0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0,  0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 1.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_y1 ){
      ierr = dirichletBC( b, bndy1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 1.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    
    if ( slip_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_x1 ){
      ierr = dirichletBC( b, bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x1 ){ 
      ierr = dirichletBC( b, bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y0 ){
      ierr = dirichletBC( b, bndy0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y0 ){
      ierr = dirichletBC( b, bndy0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y1 ){
      ierr = dirichletBC( b, bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y1 ){
      ierr = dirichletBC( b, bndy1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_z0 ){
      ierr = dirichletBC( b, bndz0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_z0 ){
      ierr = dirichletBC( b, bndz0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_z1 ){
      ierr = dirichletBC( b, bndz1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_z1 ){
      ierr = dirichletBC( b, bndz1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
        
    if ( periodic_bc_x ){
      ierr = dirichletBC( b, bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );		
    }
    
    if ( periodic_bc_y ){
      ierr = dirichletBC( b, bndy1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );		
    }
    
    if ( periodic_bc_z ){
      ierr = dirichletBC( b, bndz1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );		
    }
    
    if ( dual_drag_bc ){
      ierr = dirichletBC( b, ob_bndx0, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndx1, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy0, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy1, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz0, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz1, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }  
    
  } else {
    
    
    if ( backward_facing_step ){
      ierr = dirichletBC( A, b, bfs_bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bfs_bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bfs_bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, bfs_bndy1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bfs_bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bfs_bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }	  
    
    if ( jet ){
      ierr = dirichletBC_GRID( A, b, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( A, b, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( A, b, grid, 3, 2, DirichletIdx, DirichletVal );
    }

    if ( bluff_body ){
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndz0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndz1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }  
    
    if ( LES_benchmark ){
      ierr = dirichletBC( A, b, les_bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndx0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, les_bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, les_bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, les_bndz0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndz0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, les_bndz1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }  
    
    if ( couette_bc_y ){
      ierr = dirichletBC( A, b, bndy1, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, bndy0, -1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy0,  0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy0,  0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_x0 ){
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx0, 1.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_y1 ){
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 1.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    
    if ( slip_bc_x0 ){
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x0 ){
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_x1 ){
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x1 ){ 
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y0 ){
      ierr = dirichletBC( A, b, bndy0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y0 ){
      ierr = dirichletBC( A, b, bndy0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y1 ){
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y1 ){
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_z0 ){
      ierr = dirichletBC( A, b, bndz0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_z0 ){
      ierr = dirichletBC( A, b, bndz0, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_z1 ){
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_z1 ){
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }
        
    if ( periodic_bc_x ){
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );		
    }
   
    if ( periodic_bc_y ){
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );		
    }
   
    if ( periodic_bc_z ){
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );		
    }
   
    if ( dual_drag_bc ){
      ierr = dirichletBC( A, b, ob_bndx0, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndx1, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy0, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy1, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndz0, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz0, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz0, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndz1, 1.0, grid, 3, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );
    }  
    
    A_assembled = true;
    
  }	
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemDualVelocity::SetDomain()
{
  bndx0.newsize(nsd,2);
  bndx1.newsize(nsd,2);
  bndy0.newsize(nsd,2);
  bndy1.newsize(nsd,2);
  bndz0.newsize(nsd,2);
  bndz1.newsize(nsd,2);
  
  ob_bndx0.newsize(nsd,2);
  ob_bndx1.newsize(nsd,2);
  ob_bndy0.newsize(nsd,2);
  ob_bndy1.newsize(nsd,2);
  ob_bndz0.newsize(nsd,2);
  ob_bndz1.newsize(nsd,2);
  
  les_bndx0.newsize(nsd,2);
  les_bndx1.newsize(nsd,2);
  les_bndy1.newsize(nsd,2);
  les_bndz0.newsize(nsd,2);
  les_bndz1.newsize(nsd,2);
  
  bfs_bndx1.newsize(nsd,2);
  bfs_bndy1.newsize(nsd,2);
  
  bndx0(0,0) = bnd_x0;
  bndx0(0,1) = bnd_x0;
  bndx0(1,0) = bnd_y0;
  bndx0(1,1) = bnd_y1;
  bndx0(2,0) = bnd_z0;
  bndx0(2,1) = bnd_z1;
  
  bndx1(0,0) = bnd_x1;
  bndx1(0,1) = bnd_x1;
  bndx1(1,0) = bnd_y0;
  bndx1(1,1) = bnd_y1;
  bndx1(2,0) = bnd_z0;
  bndx1(2,1) = bnd_z1;
  
  bndy0(0,0) = bnd_x0;
  bndy0(0,1) = bnd_x1;
  bndy0(1,0) = bnd_y0;
  bndy0(1,1) = bnd_y0;
  bndy0(2,0) = bnd_z0;
  bndy0(2,1) = bnd_z1;
  
  bndy1(0,0) = bnd_x0;
  bndy1(0,1) = bnd_x1;
  bndy1(1,0) = bnd_y1;
  bndy1(1,1) = bnd_y1;
  bndy1(2,0) = bnd_z0;
  bndy1(2,1) = bnd_z1;
  
  bndz0(0,0) = bnd_x0;
  bndz0(0,1) = bnd_x1;
  bndz0(1,0) = bnd_y0;
  bndz0(1,1) = bnd_y1;
  bndz0(2,0) = bnd_z0;
  bndz0(2,1) = bnd_z0;
  
  bndz1(0,0) = bnd_x0;
  bndz1(0,1) = bnd_x1;
  bndz1(1,0) = bnd_y0;
  bndz1(1,1) = bnd_y1;
  bndz1(2,0) = bnd_z1;
  bndz1(2,1) = bnd_z1;
  
  ob_bndx0(0,0) = 0.375;
  ob_bndx0(0,1) = 0.375;
  ob_bndx0(1,0) = 0.375;
  ob_bndx0(1,1) = 0.625;
  ob_bndx0(2,0) = 0.375;
  ob_bndx0(2,1) = 0.625;
  
  ob_bndx1(0,0) = 0.625;
  ob_bndx1(0,1) = 0.625;
  ob_bndx1(1,0) = 0.375;
  ob_bndx1(1,1) = 0.625;
  ob_bndx1(2,0) = 0.375;
  ob_bndx1(2,1) = 0.625;
  
  ob_bndy0(0,0) = 0.375;
  ob_bndy0(0,1) = 0.625;
  ob_bndy0(1,0) = 0.375;
  ob_bndy0(1,1) = 0.375;
  ob_bndy0(2,0) = 0.375;
  ob_bndy0(2,1) = 0.625;
  
  ob_bndy1(0,0) = 0.375;
  ob_bndy1(0,1) = 0.625;
  ob_bndy1(1,0) = 0.625;
  ob_bndy1(1,1) = 0.625;
  ob_bndy1(2,0) = 0.375;
  ob_bndy1(2,1) = 0.625;
  
  ob_bndz0(0,0) = 0.375;
  ob_bndz0(0,1) = 0.625;
  ob_bndz0(1,0) = 0.375;
  ob_bndz0(1,1) = 0.625;
  ob_bndz0(2,0) = 0.375;
  ob_bndz0(2,1) = 0.375;
  
  ob_bndz1(0,0) = 0.375;
  ob_bndz1(0,1) = 0.625;
  ob_bndz1(1,0) = 0.375;
  ob_bndz1(1,1) = 0.625;
  ob_bndz1(2,0) = 0.625;
  ob_bndz1(2,1) = 0.625;
  
  bfs_bndx1(0,0) = 0.5;
  bfs_bndx1(0,1) = 0.5;
  bfs_bndx1(1,0) = 0.0;
  bfs_bndx1(1,1) = 0.5;
  bfs_bndx1(2,0) = 0.0;
  bfs_bndx1(2,1) = 1.0;
  
  bfs_bndy1(0,0) = 0.0;
  bfs_bndy1(0,1) = 0.5;
  bfs_bndy1(1,0) = 0.5;
  bfs_bndy1(1,1) = 0.5;
  bfs_bndy1(2,0) = 0.0;
  bfs_bndy1(2,1) = 1.0;
    
  les_bndx0(0,0) = 3.0;
  les_bndx0(0,1) = 3.0;
  les_bndx0(1,0) = 0.0;
  les_bndx0(1,1) = 1.0;
  les_bndx0(2,0) = 3.0;
  les_bndx0(2,1) = 4.0;
  
  les_bndx1(0,0) = 4.0;
  les_bndx1(0,1) = 4.0;
  les_bndx1(1,0) = 0.0;
  les_bndx1(1,1) = 1.0;
  les_bndx1(2,0) = 3.0;
  les_bndx1(2,1) = 4.0;
  
  les_bndy1(0,0) = 3.0;
  les_bndy1(0,1) = 4.0;
  les_bndy1(1,0) = 1.0;
  les_bndy1(1,1) = 1.0;
  les_bndy1(2,0) = 3.0;
  les_bndy1(2,1) = 4.0;
  
  les_bndz0(0,0) = 3.0;
  les_bndz0(0,1) = 4.0;
  les_bndz0(1,0) = 0.0;
  les_bndz0(1,1) = 1.0;
  les_bndz0(2,0) = 3.0;
  les_bndz0(2,1) = 3.0;
  
  les_bndz1(0,0) = 3.0;
  les_bndz1(0,1) = 4.0;
  les_bndz1(1,0) = 0.0;
  les_bndz1(1,1) = 1.0;
  les_bndz1(2,0) = 4.0;
  les_bndz1(2,1) = 4.0;
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemDualVelocity::ReAssembleMatrix( bool am )
{
  re_assemble_matrix = am;
  
  return 0;
}
  

