// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "StatProblemVelocity.hh"

//----------------------------------------------------------------------------------------------------
StatProblemVelocity::StatProblemVelocity(Grid *grid) : StatProblem(grid)
{ 
  noeq = 3;

  re_assemble_matrix = true;
  
  // Initialize Fields to attach to Equation
  u1_prev_TS.AttachGrid(grid);
  u2_prev_TS.AttachGrid(grid);
  u3_prev_TS.AttachGrid(grid);
  p_prev_TS.AttachGrid(grid);
  
  u1_prev_NL.AttachGrid(grid);
  u2_prev_NL.AttachGrid(grid);
  u3_prev_NL.AttachGrid(grid);
  p_prev_NL.AttachGrid(grid);
  
  u1.AttachGrid(grid);
  u2.AttachGrid(grid);
  u3.AttachGrid(grid);
  p.AttachGrid(grid);
  
  dt.AttachGrid(grid);
  t.AttachGrid(grid);
  Re.AttachGrid(grid);
  
  settings->Get("turbulent flow",&turbulent_flow);

  if ( turbulent_flow ){
    tau11.AttachGrid(grid);
    tau12.AttachGrid(grid);
    tau13.AttachGrid(grid);
    tau22.AttachGrid(grid);
    tau23.AttachGrid(grid);
    tau33.AttachGrid(grid);
  }

  settings->Get("compute projections",&compute_projections);

  if ( compute_projections ){
    Field u1_coarse(grid);
    Field u2_coarse(grid);
    Field u3_coarse(grid);
    Field p_coarse(grid);

    Field u1_fine(grid);
    Field u2_fine(grid);
    Field u3_fine(grid);
    Field p_fine(grid);
  }

  settings->Get("bnd x0", &bnd_x0);
  settings->Get("bnd x1", &bnd_x1);
  settings->Get("bnd y0", &bnd_y0);
  settings->Get("bnd y1", &bnd_y1);
  settings->Get("bnd z0", &bnd_z0);
  settings->Get("bnd z1", &bnd_z1);
 
  settings->Get("reynolds number", &ReNo);

  settings->Get("driven cavity", &driven_cavity);
  settings->Get("backward facing step", &backward_facing_step);
  settings->Get("bluff body", &bluff_body);
  settings->Get("jet", &jet);
  settings->Get("LES benchmark", &LES_benchmark);

  settings->Get("slip bc x0", &slip_bc_x0);
  settings->Get("slip bc x1", &slip_bc_x1);
  settings->Get("slip bc y0", &slip_bc_y0);
  settings->Get("slip bc y1", &slip_bc_y1);
  settings->Get("slip bc z0", &slip_bc_z0);
  settings->Get("slip bc z1", &slip_bc_z1);

  settings->Get("no slip bc x0", &no_slip_bc_x0);
  settings->Get("no slip bc x1", &no_slip_bc_x1);
  settings->Get("no slip bc y0", &no_slip_bc_y0);
  settings->Get("no slip bc y1", &no_slip_bc_y1);
  settings->Get("no slip bc z0", &no_slip_bc_z0);
  settings->Get("no slip bc z1", &no_slip_bc_z1);

  settings->Get("periodic bc x", &periodic_bc_x);
  settings->Get("periodic bc y", &periodic_bc_y);
  settings->Get("periodic bc z", &periodic_bc_z);

  settings->Get("inflow bc x0", &inflow_bc_x0);
  settings->Get("outflow bc x0", &outflow_bc_x0);
  settings->Get("outflow bc x1", &outflow_bc_x1);

  settings->Get("couette bc y", &couette_bc_y);

  settings->Get("driven cavity bc x0", &driven_cavity_bc_x0);
  settings->Get("driven cavity bc y1", &driven_cavity_bc_y1);

  settings->Get("dual drag bc", &dual_drag_bc);

  // define the computational domain
  SetDomain();
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity::Timestep( MV_ColMat<real>& val, real time, real deltat )
{
  // initialize inner variables 
  int ierr;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );

  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  u3_prev_TS.InitByNodes(uk_prev_TS,2);
  //p_prev_TS.InitByNodes(uk_prev_TS,3);

  u1_prev_NL.InitByNodes(uk_prev_NL,0);
  u2_prev_NL.InitByNodes(uk_prev_NL,1);
  u3_prev_NL.InitByNodes(uk_prev_NL,2);
  //p_prev_NL.InitByNodes(uk_prev_NL,3);
  
  uk = &val;
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  u3.InitByNodes(uk,2);
  p.InitByNodes(uk,3);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  if ( turbulent_flow ){
    tau11.InitByNodes(tau,0);
    tau12.InitByNodes(tau,1);
    tau13.InitByNodes(tau,2);
    tau22.InitByNodes(tau,3);
    tau23.InitByNodes(tau,4);
    tau33.InitByNodes(tau,5);
  }

  if ( compute_projections ){
    u1_coarse.InitByNodes(uk_coarse,0);
    u2_coarse.InitByNodes(uk_coarse,1);
    u3_coarse.InitByNodes(uk_coarse,2);
    p_coarse.InitByNodes(uk_coarse,3);

    u1_fine.InitByNodes(uk_fine,0);
    u2_fine.InitByNodes(uk_fine,1);
    u3_fine.InitByNodes(uk_fine,2);
    p_fine.InitByNodes(uk_fine,3);
  }

  // get master-slave info if periodic bc
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, grid, perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, grid, perList );
  if ( periodic_bc_z ) getPerList( bndz0, bndz1, grid, perList );

  Vector b(noNodes*noeq);
  
  // assemble matrix if not assembled
  if ( !A_assembled ){
    display->Status(2,"Assemble matrix A");
    // check if periodic bc
    if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
      //discrete->createSeqMatPeriodic( &A, grid, noeq, perList );
      //discrete->discretiseLHSIntExactPeriodicInitial( *equation, grid, A, noeq, perList );
    } else{
      discrete->_Allocate(A);
      discrete->_AssembleLHS(A);
    }
    // if assembled, check if to reassemble
  } else if ( A_assembled && re_assemble_matrix ){
    display->Status(2,"Re-assembling matrix A");
    // check if periodic bc
    if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
      //discrete->discretiseLHSIntExactPeriodic( *equation, grid, A, noeq, perList );
    } else{
      discrete->_AssembleLHS(A);
    }
  }
  
  display->Status(2,"Assemble vector b");
  // check if periodic bc
  if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
    //discrete->discretiseRHSIntExactPeriodic( *equation, grid, b, noeq, perList );
  } else{
    discrete->_AssembleRHS(&b);
  }

  // compute norm of rhs 
  real norm = b.Norm();
  display->Status(2,"Load norm: %f",norm);
  
  // set dirichlet bc
  SetBoundaryConditions(A,&b);
  
  // compute norm of rhs 
  norm = b.Norm();
  display->Status(2,"Load norm: %f (with b.c.)",norm);
  
  // set start solution to solution from previous iteration 
  Vector u( noeq * noNodes );
  for (int i=0; i < noNodes; i++){
    u.Set(i*noeq,  val(i,0));
    u.Set(i*noeq+1,val(i,1));
    u.Set(i*noeq+2,val(i,2));
  }

  // solve discrete system  
  display->Status(0,"Solve algebraic system of equations");

  //SISolver solver;
  //solver.Solve( A, &u, &b );
  KrylovSolver solver;
  solver.SetMatrix(A);
  solver.SetMethod(gmres);
  solver.Solve( &u, &b );
  
  // compute norm of solution
  norm = u.Norm();
  display->Status(4,"Norm of u = %f",norm);
  norm = A->Norm();
  display->Status(4,"Norm of A = %f",norm);

  // update solution vector
  for (int i=0; i < noNodes; i++){
    for (int j=0; j < nsd; j++){ 
      if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
	val(i,j) = u(j+perList(i)*noeq);
      } else{
	val(i,j) = u(j+i*noeq);
      }
    }
  }

  // destroy solution variables 
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity::ComputeDiscreteResidual( real time, real deltat, real& norm_2, real& norm_inf )
{
  int ierr;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );

  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  u3_prev_TS.InitByNodes(uk_prev_TS,2);
  //p_prev_TS.InitByNodes(uk_prev_TS,3);

  u1_prev_NL.InitByNodes(uk_prev_NL,0);
  u2_prev_NL.InitByNodes(uk_prev_NL,1);
  u3_prev_NL.InitByNodes(uk_prev_NL,2);
  //p_prev_NL.InitByNodes(uk_prev_NL,3);
  
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  u3.InitByNodes(uk,2);
  p.InitByNodes(uk,3);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  if ( turbulent_flow ){
    tau11.InitByNodes(tau,0);
    tau12.InitByNodes(tau,1);
    tau13.InitByNodes(tau,2);
    tau22.InitByNodes(tau,3);
    tau23.InitByNodes(tau,4);
    tau33.InitByNodes(tau,5);
  }

  if ( compute_projections ){
    u1_coarse.InitByNodes(uk_coarse,0);
    u2_coarse.InitByNodes(uk_coarse,1);
    u3_coarse.InitByNodes(uk_coarse,2);
    p_coarse.InitByNodes(uk_coarse,3);

    u1_fine.InitByNodes(uk_fine,0);
    u2_fine.InitByNodes(uk_fine,1);
    u3_fine.InitByNodes(uk_fine,2);
    p_fine.InitByNodes(uk_fine,3);
  }

  // get master-slave info if periodic bc
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, grid, perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, grid, perList );
  if ( periodic_bc_z ) getPerList( bndz0, bndz1, grid, perList );

  Vector b(noNodes*noeq);
  
  if ( re_assemble_matrix ){
    display->Status(2,"Re-assembling matrix A");
    // check if periodic bc
    if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
      //discrete->discretiseLHSIntExactPeriodic( *equation, grid, A, noeq, perList );
    } else{
      discrete->_AssembleLHS(A);
    }
  }
  
  display->Status(2,"Assemble vector b");
  // check if periodic bc
  if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
    //discrete->discretiseRHSIntExactPeriodic( *equation, grid, b, noeq, perList );
  } else{
    discrete->_AssembleRHS(&b);
  }

  // compute norm of rhs 
  real norm = b.Norm();
  display->Status(2,"Load norm: %f",norm);
  
  // set dirichlet bc
  SetBoundaryConditions(A,&b);
  
  // compute norm of rhs 
  norm = b.Norm();
  display->Status(2,"Load norm: %f (with b.c.)",norm);
  
  // set start solution to solution from previous iteration 
  Vector u( noeq * noNodes );
  for (int i=0; i < noNodes; i++){
    u.Set(i*noeq,  u1.GetNodalValue(i));
    u.Set(i*noeq+1,u2.GetNodalValue(i));
    u.Set(i*noeq+2,u3.GetNodalValue(i));
  }

  if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
    MV_Vector<int> DirichletIdx(0);  // Should be of size 0
    MV_Vector<real> DirichletVal(0); // Should be of size 0
    ierr = dirichletBC( &u, bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
    ierr = dirichletBC( &u, bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
    ierr = dirichletBC( &u, bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );		
  }
  
  Vector Au(noNodes*noeq);
  A->Mult(&u,&Au);
  
  Vector res(noNodes*noeq);
  for (int i=0; i < noNodes*noeq; i++) res.Set(i,b(i) - Au(i));

  norm_2   = res.Norm();
  norm_inf = res.Norm();

  return 0;
}
//----------------------------------------------------------------------------------------------------
/*
int StatProblemVelocity::Timestep( MV_ColMat<real>& val, real time, real deltat )
{
  // initialize inner variables 
  int ierr;
  
  Vec u;
  Vec b;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );

  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  u3_prev_TS.InitByNodes(uk_prev_TS,2);
  //p_prev_TS.InitByNodes(uk_prev_TS,3);

  u1_prev_NL.InitByNodes(uk_prev_NL,0);
  u2_prev_NL.InitByNodes(uk_prev_NL,1);
  u3_prev_NL.InitByNodes(uk_prev_NL,2);
  //p_prev_NL.InitByNodes(uk_prev_NL,3);
  
  uk = &val;
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  u3.InitByNodes(uk,2);
  p.InitByNodes(uk,3);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  if ( turbulent_flow ){
    tau11.InitByNodes(tau,0);
    tau12.InitByNodes(tau,1);
    tau13.InitByNodes(tau,2);
    tau22.InitByNodes(tau,3);
    tau23.InitByNodes(tau,4);
    tau33.InitByNodes(tau,5);
  }

  if ( compute_projections ){
    u1_coarse.InitByNodes(uk_coarse,0);
    u2_coarse.InitByNodes(uk_coarse,1);
    u3_coarse.InitByNodes(uk_coarse,2);
    p_coarse.InitByNodes(uk_coarse,3);

    u1_fine.InitByNodes(uk_fine,0);
    u2_fine.InitByNodes(uk_fine,1);
    u3_fine.InitByNodes(uk_fine,2);
    p_fine.InitByNodes(uk_fine,3);
  }

  // get master-slave info if periodic bc
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, grid, perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, grid, perList );
  if ( periodic_bc_z ) getPerList( bndz0, bndz1, grid, perList );

  // assemble matrix if not assembled
  if ( !A_assembled ){
    display->Status(2,"Assemble matrix A");
    // check if exact integration / quadrature is used
    if ( exact_integration ){
      // check if periodic bc
      if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
	discrete->createSeqMatPeriodic( &A, grid, noeq, perList );
	discrete->discretiseLHSIntExactPeriodicInitial( *equation, grid, A, noeq, perList );
      } else{
	discrete->createSeqMat( &A, grid, noeq );
	discrete->discretiseLHSIntExact( *equation, grid, A, noeq );
      }
    } else{
      discrete->createSeqMat( &A, grid, noeq );
      discrete->discretiseLHS( *equation, *mapper, grid, *quadrule, *element, A );
    }
    // if assembled, check if to reassemble
  } else if ( A_assembled && re_assemble_matrix ){
    display->Status(2,"Re-assembling matrix A");
    ierr = MatZeroEntries(A);
    // check if exact integration / quadrature is used
    if ( exact_integration ){
      // check if periodic bc
      if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
	discrete->discretiseLHSIntExactPeriodic( *equation, grid, A, noeq, perList );
      } else{
	discrete->discretiseLHSIntExact( *equation, grid, A, noeq );
      }
    } else{
      discrete->discretiseLHS( *equation, *mapper, grid, *quadrule, *element, A );
    }
  }
  
  display->Status(2,"Assemble vector b");
  discrete->createSeqRHS( &u, &b, grid, noeq );
  if ( exact_integration ){
    // check if periodic bc
    if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
      discrete->discretiseRHSIntExactPeriodic( *equation, grid, b, noeq, perList );
    } else{
      discrete->discretiseRHSIntExact( *equation, grid, b, noeq );
    }
  } else{
    discrete->discretiseRHS( *equation, *mapper, grid, *quadrule, *element, b );
  }

  // compute norm of rhs 
  real norm;
  ierr = VecNorm( b, NORM_2, &norm );
  display->Status(2,"Load norm: %f",norm);
  
  // set dirichlet bc
  SetBoundaryConditions(A,b);
  
  // compute norm of rhs 
  ierr = VecNorm( b, NORM_2, &norm );
  display->Status(2,"Load norm: %f (with b.c.)",norm);
  
  // set start solution to solution from previous iteration 
  MV_Vector<real> uprev( noeq * noNodes );
  int i;
  for ( i=0; i < noNodes; i++ ){
    uprev(i*noeq)     = val(i,0);
    uprev(i*noeq + 1) = val(i,1);
    uprev(i*noeq + 2) = val(i,2);
  }
  MV_Vector<int> idx( noeq * noNodes );
  for ( i=0; i < noeq * noNodes; i++ ) idx(i) = i;
  
  ierr = VecSetValues(u,noeq*noNodes,&idx(0),&uprev(0),INSERT_VALUES); CHKERRQ( ierr ); 

  // solve discrete system  
  display->Status(2,"Solve algebraic system of equations");
  discrete->solve( A, u, b );
  
  // compute norm of solution
  VecNorm(u,NORM_INFINITY,&norm);
  display->Status(4,"Norm of u = %f",norm);
  MatNorm(A,NORM_INFINITY,&norm);
  display->Status(4,"Norm of A = %f",norm);

  // update solution vector
  ierr = VecGetArray(u, &temp); CHKERRQ( ierr );
  int j;
  for ( i=0; i < noNodes; i++ ){
    for ( j=0; j < nsd; j++ ){ 
      if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
	val(i,j) = temp[j+perList(i)*noeq];
      } else{
	val(i,j) = temp[j+i*noeq];
      }
    }
  }
  ierr = VecRestoreArray(u, &temp); CHKERRQ( ierr );

  // destroy solution variables 
  ierr = VecDestroy(u); CHKERRQ(ierr);
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity::ComputeDiscreteResidual( real time, real deltat, real& norm_2, real& norm_inf )
{
  int ierr;
  
  Vec u;
  Vec b;

  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );

  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  u3_prev_TS.InitByNodes(uk_prev_TS,2);
  //p_prev_TS.InitByNodes(uk_prev_TS,3);

  u1_prev_NL.InitByNodes(uk_prev_NL,0);
  u2_prev_NL.InitByNodes(uk_prev_NL,1);
  u3_prev_NL.InitByNodes(uk_prev_NL,2);
  //p_prev_NL.InitByNodes(uk_prev_NL,3);
  
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  u3.InitByNodes(uk,2);
  p.InitByNodes(uk,3);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  if ( turbulent_flow ){
    tau11.InitByNodes(tau,0);
    tau12.InitByNodes(tau,1);
    tau13.InitByNodes(tau,2);
    tau22.InitByNodes(tau,3);
    tau23.InitByNodes(tau,4);
    tau33.InitByNodes(tau,5);
  }

  if ( compute_projections ){
    u1_coarse.InitByNodes(uk_coarse,0);
    u2_coarse.InitByNodes(uk_coarse,1);
    u3_coarse.InitByNodes(uk_coarse,2);
    p_coarse.InitByNodes(uk_coarse,3);

    u1_fine.InitByNodes(uk_fine,0);
    u2_fine.InitByNodes(uk_fine,1);
    u3_fine.InitByNodes(uk_fine,2);
    p_fine.InitByNodes(uk_fine,3);
  }

  // get master-slave info if periodic bc
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, grid, perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, grid, perList );
  if ( periodic_bc_z ) getPerList( bndz0, bndz1, grid, perList );

  if ( re_assemble_matrix ){
    display->Status(2,"Re-assembling matrix A for residual");
    ierr = MatZeroEntries(A);
    if ( exact_integration ){
      if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
	discrete->discretiseLHSIntExactPeriodic( *equation, grid, A, noeq, perList );
      } else{
	discrete->discretiseLHSIntExact( *equation, grid, A, noeq );
      }
    } else{
      discrete->discretiseLHS( *equation, *mapper, grid, *quadrule, *element, A );
    }
  }
  
  display->Status(2,"Assemble vector b");
  discrete->createSeqRHS( &u, &b, grid, noeq );
  if ( exact_integration ){
    if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
      discrete->discretiseRHSIntExactPeriodic( *equation, grid, b, noeq, perList );
    } else{
      discrete->discretiseRHSIntExact( *equation, grid, b, noeq );
    }
  } else{
    discrete->discretiseRHS( *equation, *mapper, grid, *quadrule, *element, b );	
  }
  
  // set dirichlet bc
  SetBoundaryConditions(A,b);
  
  MV_Vector<real> uprev( noeq * noNodes );
  int i;
  for ( i=0; i < noNodes; i++ ){
    uprev(i*noeq)     = u1.GetNodalValue(i);
    uprev(i*noeq + 1) = u2.GetNodalValue(i);
    uprev(i*noeq + 2) = u3.GetNodalValue(i);
  }
  MV_Vector<int> idx( noeq * noNodes );
  for ( i=0; i < noeq * noNodes; i++ ) idx(i) = i;
  ierr = VecSetValues(u,noeq*noNodes,&idx(0),&uprev(0),INSERT_VALUES); CHKERRQ( ierr ); 
  
  if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
    MV_Vector<int> DirichletIdx(0);  // Should be of size 0
    MV_Vector<real> DirichletVal(0); // Should be of size 0
    ierr = dirichletBC( u, bndx1, 0.0, grid, 3, 0, DirichletIdx, DirichletVal );
    ierr = dirichletBC( u, bndx1, 0.0, grid, 3, 1, DirichletIdx, DirichletVal );
    ierr = dirichletBC( u, bndx1, 0.0, grid, 3, 2, DirichletIdx, DirichletVal );		
  }
  
  Vec res;
  VecDuplicate( b, &res);
  Scalar minus = - 1.0;
  VecScale( &minus, b ); 
  MatMultAdd( A, u, b, res );
  ierr = VecNorm( res, NORM_2, &norm_2 );
  ierr = VecNorm( res, NORM_INFINITY, &norm_inf );
  
  ierr = VecDestroy(res); CHKERRQ(ierr);
  ierr = VecDestroy(u); CHKERRQ(ierr);
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
}
*/
//----------------------------------------------------------------------------------------------------
  /*
int StatProblemVelocity::AssembleMatrix( MV_ColMat<real>& val, real time, real deltat, Mat& Matrix )
{
  // initialize inner variables 
  Scalar *temp;
  int ierr;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );

  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  u3_prev_TS.InitByNodes(uk_prev_TS,2);
  //p_prev_TS.InitByNodes(uk_prev_TS,3);

  u1_prev_NL.InitByNodes(uk_prev_NL,0);
  u2_prev_NL.InitByNodes(uk_prev_NL,1);
  u3_prev_NL.InitByNodes(uk_prev_NL,2);
  //p_prev_NL.InitByNodes(uk_prev_NL,3);
  
  uk = &val;
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  u3.InitByNodes(uk,2);
  p.InitByNodes(uk,3);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  if ( turbulent_flow ){
    tau11.InitByNodes(tau,0);
    tau12.InitByNodes(tau,1);
    tau13.InitByNodes(tau,2);
    tau22.InitByNodes(tau,3);
    tau23.InitByNodes(tau,4);
    tau33.InitByNodes(tau,5);
  }

  if ( compute_projections ){
    u1_coarse.InitByNodes(uk_coarse,0);
    u2_coarse.InitByNodes(uk_coarse,1);
    u3_coarse.InitByNodes(uk_coarse,2);
    p_coarse.InitByNodes(uk_coarse,3);

    u1_fine.InitByNodes(uk_fine,0);
    u2_fine.InitByNodes(uk_fine,1);
    u3_fine.InitByNodes(uk_fine,2);
    p_fine.InitByNodes(uk_fine,3);
  }

  // get master-slave info if periodic bc
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, grid, perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, grid, perList );
  if ( periodic_bc_z ) getPerList( bndz0, bndz1, grid, perList );

  // assemble matrix if not assembled
  if ( !A_assembled ){
    display->Status(2,"Assemble matrix A");
    // check if exact integration / quadrature is used
    if ( exact_integration ){
      // check if periodic bc
      if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
	discrete->createSeqMatPeriodic( &A, grid, noeq, perList );
	discrete->discretiseLHSIntExactPeriodicInitial( *equation, grid, A, noeq, perList );
      } else{
	discrete->createSeqMat( &A, grid, noeq );
	discrete->discretiseLHSIntExact( *equation, grid, A, noeq );
      }
    } else{
      discrete->createSeqMat( &A, grid, noeq );
      discrete->discretiseLHS( *equation, *mapper, grid, *quadrule, *element, A );
    }
    // if assembled, check if to reassemble
  } else if ( A_assembled && re_assemble_matrix ){
    display->Status(2,"Re-assembling matrix A");
    ierr = MatZeroEntries(A);
    // check if exact integration / quadrature is used
    if ( exact_integration ){
      // check if periodic bc
      if ( periodic_bc_x || periodic_bc_y || periodic_bc_z ){
	discrete->discretiseLHSIntExactPeriodic( *equation, grid, A, noeq, perList );
      } else{
	discrete->discretiseLHSIntExact( *equation, grid, A, noeq );
      }
    } else{
      discrete->discretiseLHS( *equation, *mapper, grid, *quadrule, *element, A );
    }
  }

  // FIXME: Temp sol.
  Vec b;
  discrete->createSeqRHS( &b, grid, noeq );

  // set dirichlet bc
  SetBoundaryConditions(A,b);

  Matrix = A;
  
  // destroy solution variables 
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
}
  */
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity::InitEquationData( real time, real deltat )
{      
  int sz;
  data.newsize(15);
  sz = 15;
  if ( turbulent_flow ) data.newsize(sz+6);
  sz = data.size();
  if ( compute_projections ) data.newsize(sz+8);
  sz = 0;

  data(sz++)  = &u1_prev_TS;
  data(sz++)  = &u2_prev_TS;
  data(sz++)  = &u3_prev_TS;
  data(sz++)  = &p_prev_TS;
  
  data(sz++)  = &u1_prev_NL;
  data(sz++)  = &u2_prev_NL;
  data(sz++)  = &u3_prev_NL;
  data(sz++)  = &p_prev_NL;
  
  data(sz++)  = &u1;
  data(sz++)  = &u2;
  data(sz++) = &u3;
  data(sz++) = &p;
  
  data(sz++) = &dt;
  data(sz++) = &t;
  data(sz++) = &Re;
  
  if ( turbulent_flow ){
    data(sz++) = &tau11;
    data(sz++) = &tau12;
    data(sz++) = &tau13;
    data(sz++) = &tau22;
    data(sz++) = &tau23;
    data(sz++) = &tau33;
  }

  if ( compute_projections ){
    data(sz++) = &u1_coarse;
    data(sz++) = &u2_coarse;
    data(sz++) = &u3_coarse;
    data(sz++) = &p_coarse;

    data(sz++) = &u1_fine;
    data(sz++) = &u2_fine;
    data(sz++) = &u3_fine;
    data(sz++) = &p_fine;
  }

  equation->attachField(data);
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity::SetBoundaryConditions( SparseMatrix *A, Vector *b )
{
  // set dirichlet bc
  MV_Vector<int> DirichletIdx(0); // Should be of size 0
  MV_Vector<real> DirichletVal(0); // Should be of size 0
  
  int ierr;
  
  if ( !re_assemble_matrix && A_assembled ){
    
    if ( backward_facing_step ){
      ierr = dirichletBC( b, bfs_bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, bfs_bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }	  
    
    if ( jet ){
      ierr = dirichletBC_GRID( b, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( b, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( b, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    
    if ( bluff_body ){
      ierr = dirichletBC( b, ob_bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }  
    
    if ( LES_benchmark ){
      ierr = dirichletBC( b, les_bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndx0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, les_bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, les_bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, les_bndz0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndz0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, les_bndz1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, les_bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }  
    
    if ( couette_bc_y ){
      ierr = dirichletBC( b, bndy1, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, bndy0, -1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0,  0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0,  0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    
    if ( inflow_bc_x0 ){
      ierr = dirichletBCInflow( b, bndx0, grid, noeq, 0, 1.0, &turbulence, DirichletIdx, DirichletVal );
      ierr = dirichletBCInflow( b, bndx0, grid, noeq, 1, 0.0, &turbulence, DirichletIdx, DirichletVal );
      ierr = dirichletBCInflow( b, bndx0, grid, noeq, 2, 0.0, &turbulence, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 1.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_y1 ){
      ierr = dirichletBC( b, bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 1.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    
    if ( slip_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_x1 ){
      ierr = dirichletBC( b, bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x1 ){ 
      ierr = dirichletBC( b, bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y0 ){
      ierr = dirichletBC( b, bndy0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y0 ){
      ierr = dirichletBC( b, bndy0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y1 ){
      ierr = dirichletBC( b, bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y1 ){
      ierr = dirichletBC( b, bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_z0 ){
      ierr = dirichletBC( b, bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_z0 ){
      ierr = dirichletBC( b, bndz0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_z1 ){
      ierr = dirichletBC( b, bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_z1 ){
      ierr = dirichletBC( b, bndz1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
        
    if ( periodic_bc_x ){
      ierr = dirichletBC( b, bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );		
    }
    
    if ( periodic_bc_y ){
      ierr = dirichletBC( b, bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );		
    }
    
    if ( periodic_bc_z ){
      ierr = dirichletBC( b, bndz1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );		
    }
    
    if ( dual_drag_bc ){
      ierr = dirichletBC( b, ob_bndx0, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndx1, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy0, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy1, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz0, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz1, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }  
    
  } else {
    
    
    if ( backward_facing_step ){
      ierr = dirichletBC( A, b, bfs_bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bfs_bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bfs_bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, bfs_bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bfs_bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bfs_bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }	  
    
    if ( jet ){
      ierr = dirichletBC_GRID( A, b, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( A, b, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( A, b, grid, noeq, 2, DirichletIdx, DirichletVal );
    }

    if ( bluff_body ){
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndz0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndz1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }  
    
    if ( LES_benchmark ){
      ierr = dirichletBC( A, b, les_bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndx0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, les_bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, les_bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, les_bndz0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndz0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, les_bndz1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, les_bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }  
    
    if ( couette_bc_y ){
      ierr = dirichletBC( A, b, bndy1, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, bndy0, -1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy0,  0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy0,  0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    
    if ( inflow_bc_x0 ){
      ierr = dirichletBCInflow( A, b, bndx0, grid, noeq, 0, 1.0, &turbulence, DirichletIdx, DirichletVal );
      ierr = dirichletBCInflow( A, b, bndx0, grid, noeq, 1, 0.0, &turbulence, DirichletIdx, DirichletVal );
      ierr = dirichletBCInflow( A, b, bndx0, grid, noeq, 2, 0.0, &turbulence, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_x0 ){
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx0, 1.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_y1 ){
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 1.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    
    if ( slip_bc_x0 ){
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x0 ){
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_x1 ){
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x1 ){ 
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y0 ){
      ierr = dirichletBC( A, b, bndy0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y0 ){
      ierr = dirichletBC( A, b, bndy0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y1 ){
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y1 ){
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_z0 ){
      ierr = dirichletBC( A, b, bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_z0 ){
      ierr = dirichletBC( A, b, bndz0, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_z1 ){
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_z1 ){
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }
        
    if ( periodic_bc_x ){
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );		
    }
   
    if ( periodic_bc_y ){
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );		
    }
   
    if ( periodic_bc_z ){
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );		
    }
   
    if ( dual_drag_bc ){
      ierr = dirichletBC( A, b, ob_bndx0, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndx1, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy0, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy1, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndz0, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz0, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz0, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndz1, 1.0, grid, noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz1, 0.0, grid, noeq, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndz1, 0.0, grid, noeq, 2, DirichletIdx, DirichletVal );
    }  
    
    A_assembled = true;
    
  }	
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity::SetDomain()
{
  
  bndx0.newsize(nsd,2);
  bndx1.newsize(nsd,2);
  bndy0.newsize(nsd,2);
  bndy1.newsize(nsd,2);
  bndz0.newsize(nsd,2);
  bndz1.newsize(nsd,2);
  
  ob_bndx0.newsize(nsd,2);
  ob_bndx1.newsize(nsd,2);
  ob_bndy0.newsize(nsd,2);
  ob_bndy1.newsize(nsd,2);
  ob_bndz0.newsize(nsd,2);
  ob_bndz1.newsize(nsd,2);
  
  les_bndx0.newsize(nsd,2);
  les_bndx1.newsize(nsd,2);
  les_bndy1.newsize(nsd,2);
  les_bndz0.newsize(nsd,2);
  les_bndz1.newsize(nsd,2);
  
  bfs_bndx1.newsize(nsd,2);
  bfs_bndy1.newsize(nsd,2);
  
  bndx0(0,0) = bnd_x0;
  bndx0(0,1) = bnd_x0;
  bndx0(1,0) = bnd_y0;
  bndx0(1,1) = bnd_y1;
  bndx0(2,0) = bnd_z0;
  bndx0(2,1) = bnd_z1;
  
  bndx1(0,0) = bnd_x1;
  bndx1(0,1) = bnd_x1;
  bndx1(1,0) = bnd_y0;
  bndx1(1,1) = bnd_y1;
  bndx1(2,0) = bnd_z0;
  bndx1(2,1) = bnd_z1;
  
  bndy0(0,0) = bnd_x0;
  bndy0(0,1) = bnd_x1;
  bndy0(1,0) = bnd_y0;
  bndy0(1,1) = bnd_y0;
  bndy0(2,0) = bnd_z0;
  bndy0(2,1) = bnd_z1;
  
  bndy1(0,0) = bnd_x0;
  bndy1(0,1) = bnd_x1;
  bndy1(1,0) = bnd_y1;
  bndy1(1,1) = bnd_y1;
  bndy1(2,0) = bnd_z0;
  bndy1(2,1) = bnd_z1;
  
  bndz0(0,0) = bnd_x0;
  bndz0(0,1) = bnd_x1;
  bndz0(1,0) = bnd_y0;
  bndz0(1,1) = bnd_y1;
  bndz0(2,0) = bnd_z0;
  bndz0(2,1) = bnd_z0;
  
  bndz1(0,0) = bnd_x0;
  bndz1(0,1) = bnd_x1;
  bndz1(1,0) = bnd_y0;
  bndz1(1,1) = bnd_y1;
  bndz1(2,0) = bnd_z1;
  bndz1(2,1) = bnd_z1;
  
  ob_bndx0(0,0) = 0.375;
  ob_bndx0(0,1) = 0.375;
  ob_bndx0(1,0) = 0.375;
  ob_bndx0(1,1) = 0.625;
  ob_bndx0(2,0) = 0.375;
  ob_bndx0(2,1) = 0.625;
  
  ob_bndx1(0,0) = 0.625;
  ob_bndx1(0,1) = 0.625;
  ob_bndx1(1,0) = 0.375;
  ob_bndx1(1,1) = 0.625;
  ob_bndx1(2,0) = 0.375;
  ob_bndx1(2,1) = 0.625;
  
  ob_bndy0(0,0) = 0.375;
  ob_bndy0(0,1) = 0.625;
  ob_bndy0(1,0) = 0.375;
  ob_bndy0(1,1) = 0.375;
  ob_bndy0(2,0) = 0.375;
  ob_bndy0(2,1) = 0.625;
  
  ob_bndy1(0,0) = 0.375;
  ob_bndy1(0,1) = 0.625;
  ob_bndy1(1,0) = 0.625;
  ob_bndy1(1,1) = 0.625;
  ob_bndy1(2,0) = 0.375;
  ob_bndy1(2,1) = 0.625;
  
  ob_bndz0(0,0) = 0.375;
  ob_bndz0(0,1) = 0.625;
  ob_bndz0(1,0) = 0.375;
  ob_bndz0(1,1) = 0.625;
  ob_bndz0(2,0) = 0.375;
  ob_bndz0(2,1) = 0.375;
  
  ob_bndz1(0,0) = 0.375;
  ob_bndz1(0,1) = 0.625;
  ob_bndz1(1,0) = 0.375;
  ob_bndz1(1,1) = 0.625;
  ob_bndz1(2,0) = 0.625;
  ob_bndz1(2,1) = 0.625;
  
  bfs_bndx1(0,0) = 0.5;
  bfs_bndx1(0,1) = 0.5;
  bfs_bndx1(1,0) = 0.0;
  bfs_bndx1(1,1) = 0.5;
  bfs_bndx1(2,0) = 0.0;
  bfs_bndx1(2,1) = 1.0;
  
  bfs_bndy1(0,0) = 0.0;
  bfs_bndy1(0,1) = 0.5;
  bfs_bndy1(1,0) = 0.5;
  bfs_bndy1(1,1) = 0.5;
  bfs_bndy1(2,0) = 0.0;
  bfs_bndy1(2,1) = 1.0;
  
  les_bndx0(0,0) = 3.0;
  les_bndx0(0,1) = 3.0;
  les_bndx0(1,0) = 0.0;
  les_bndx0(1,1) = 1.0;
  les_bndx0(2,0) = 3.0;
  les_bndx0(2,1) = 4.0;
  
  les_bndx1(0,0) = 4.0;
  les_bndx1(0,1) = 4.0;
  les_bndx1(1,0) = 0.0;
  les_bndx1(1,1) = 1.0;
  les_bndx1(2,0) = 3.0;
  les_bndx1(2,1) = 4.0;
  
  les_bndy1(0,0) = 3.0;
  les_bndy1(0,1) = 4.0;
  les_bndy1(1,0) = 1.0;
  les_bndy1(1,1) = 1.0;
  les_bndy1(2,0) = 3.0;
  les_bndy1(2,1) = 4.0;
  
  les_bndz0(0,0) = 3.0;
  les_bndz0(0,1) = 4.0;
  les_bndz0(1,0) = 0.0;
  les_bndz0(1,1) = 1.0;
  les_bndz0(2,0) = 3.0;
  les_bndz0(2,1) = 3.0;
  
  les_bndz1(0,0) = 3.0;
  les_bndz1(0,1) = 4.0;
  les_bndz1(1,0) = 0.0;
  les_bndz1(1,1) = 1.0;
  les_bndz1(2,0) = 4.0;
  les_bndz1(2,1) = 4.0;
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity::ReAssembleMatrix( bool am )
{
  re_assemble_matrix = am;
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity::SetTurbulentInflow( real trb )
{
  int ierr;
  int i,j;
  int *bndNodes; 
  MV_Vector<int> bNodes(noNodes);
  int noBndNodes;

  ierr = getBndNodes(bNodes, &noBndNodes, bndx0, grid);

  turbulence.newsize(noBndNodes,noeq);
  for ( i=0; i < noBndNodes; i++ ){
    for ( j=0; j < noeq; j++ ){
      turbulence(i,j) = trb * 2.0*(drand48() - 0.5);
    }
  }
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity::SetReynoldsStresses( MV_ColMat<real>* taupnt )
{
  tau = taupnt;
  
  return 0;
}
//----------------------------------------------------------------------------------------------------


