// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-28.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "TranProblemDualNS.hh"

//----------------------------------------------------------------------------------------------------
TranProblemDualNS::TranProblemDualNS(Grid *grid, MV_Vector < GridRelation > *P):TranProblem(grid,P)
{
	char output_file[KW_LINELENGTH];

	// Data file for the dual 
	settings->Get("output file dual", output_file);
	opendx = new OpenDX(output_file, 2, 3, 1);
	opendx->SetLabel(0, "Velocity");
	opendx->SetLabel(1, "Pressure");

	// Data file for the primal
	settings->Get("output file primal", output_file);
	opendx_primal = new OpenDX(output_file, 2, 3, 1);
	opendx_primal->SetLabel(0, "Velocity");
	opendx_primal->SetLabel(1, "Pressure");

	// Get inner variables
	settings->Get("reynolds number", &reynolds_number);

	matlab_data = new Matlab(39);

	matlab_data->SetLabel(0, "|| u ||_1");
	matlab_data->SetLabel(1, "|| u ||_2");
	matlab_data->SetLabel(2, "|| u ||_{max}");
	matlab_data->SetLabel(3, "| u |_{1,1}");
	matlab_data->SetLabel(4, "| u |_{1,2}");
	matlab_data->SetLabel(5, "| u |_{1,max}");
	matlab_data->SetLabel(6, "|| u ||_{1,1}");
	matlab_data->SetLabel(7, "|| u ||_{1,2}");
	matlab_data->SetLabel(8, "|| u ||_{1,max}");
	matlab_data->SetLabel(9, "| u |_{2,1}");
	matlab_data->SetLabel(10, "| u |_{2,2}");
	matlab_data->SetLabel(11, "| u |_{2,max}");
	matlab_data->SetLabel(12, "|| u ||_{2,1}");
	matlab_data->SetLabel(13, "|| u ||_{2,2}");
	matlab_data->SetLabel(14, "|| u ||_{2,max}");
	matlab_data->SetLabel(15, "|| p ||_1");
	matlab_data->SetLabel(16, "|| p ||_2");
	matlab_data->SetLabel(17, "|| p ||_{max}");
	matlab_data->SetLabel(18, "| p |_{1,1}");
	matlab_data->SetLabel(19, "| p |_{1,2}");
	matlab_data->SetLabel(20, "| p |_{1,max}");
	matlab_data->SetLabel(21, "|| p ||_{1,1}");
	matlab_data->SetLabel(22, "|| p ||_{1,2}");
	matlab_data->SetLabel(23, "|| p ||_{1,max}");
	matlab_data->SetLabel(24, "| p |_{2,1}");
	matlab_data->SetLabel(25, "| p |_{2,2}");
	matlab_data->SetLabel(26, "| p |_{2,max}");
   matlab_data->SetLabel(27, "|| p ||_{2,1}");
	matlab_data->SetLabel(28, "|| p ||_{2,2}");
	matlab_data->SetLabel(29, "|| p ||_{2,max}");
	matlab_data->SetLabel(30, "|| div u ||_1");
	matlab_data->SetLabel(31, "|| div u ||_2");
	matlab_data->SetLabel(32, "|| div u ||_{max}");
	matlab_data->SetLabel(33, "|| du/dt ||_1");
	matlab_data->SetLabel(34, "|| du/dt ||_2");
	matlab_data->SetLabel(35, "|| du/dt ||_{max}");
	matlab_data->SetLabel(36, "|| dp/dt ||_1");
	matlab_data->SetLabel(37, "|| dp/dt ||_2");
	matlab_data->SetLabel(38, "|| dp/dt ||_{max}");
}

//----------------------------------------------------------------------------------------------------
TranProblemDualNS::~TranProblemDualNS()
{
	// Delete the OpenDX files
	delete opendx;
	delete opendx_primal;
}

//----------------------------------------------------------------------------------------------------
int
 TranProblemDualNS::Solve()
{
	real progress = 0.0;

	display->Status(0, "Solving Navier-Stokes dual problem.");

	// Update internal variables
	Update();

	// Reset OpenDX files
	opendx->Reset();

	// Reset Matlab file
	matlab_data->Reset();

	// Set time t to starting time T0
	t = T;

	// Set tolerances for solution method
	SetTolerances();

	// Check output levels
	CheckOutputLevel();

	// Reset output files
	opendx->Reset();
	opendx_residuals->Reset();

	// Set FEM discretisation
	Discretiser discrete;
	Mapper mapper(nsd);
	ElementTetLin element;
	QuadratureRules quadrule(TET, 1);
	EquationDualVel_cG1cG1 eq_vel;
	EquationDualPre_cG1cG1 eq_pre;

	// Set number of equations 
	noeq_vel = eq_vel.noEq();
	noeq_pre = eq_pre.noEq();
	noeq = noeq_vel + noeq_pre;

	// Initialize solution vectors
	uk.newsize(no_nodes, noeq);
	uk_prev_TS.newsize(no_nodes, noeq);
	print_uk.newsize(printNoNodes, noeq);

	// Set initial data
	initial_data = zero_initial_data_dual;
	SetInitialData();

	// Initialize stationary problem
	sp_vel = new StatProblemDualVelocity(grid);
	sp_pre = new StatProblemDualPressure(grid);

	// Attach FEM discretisation to stationary problem
	sp_pre->AttachFEM(&discrete, &eq_pre);
	sp_vel->AttachFEM(&discrete, &eq_vel);

	// Read primal solution from file 
	globals->Get("current input level primal", &input_level_primal);
	
	//MV_ColMat < real > U((*GRID) (input_level_primal).GetNoNodes(),noeq);
	MV_ColMat <real> U(grid->GetNoNodes(),noeq);

	int i, j, k;
	char tmp[256];

	// Write data to file
	if (write_data)
		WriteData(&uk, &uk_prev_TS, t, dt);
	if (write_residuals)
		CompContResidual(&uk, &uk_prev_TS, t, dt, contResidual);

	// Start time marching (backwards)
	display->Status(2, "Start computation");
	while (t >= T0) {

		// Save solution if we should
		if (SaveSample()) {

			// Write solution to file
			for (i = 0; i < printNoNodes; i++) {
				for (j = 0; j < noeq; j++)
					print_uk(i, j) = uk(i, j);
			}
			//opendx->AddFrame(&(*GRID) (printGrid), &print_uk, 0.0);
			opendx->AddFrame(grid, &print_uk, 0.0);

		}
		// Set time step
		GetMeshSize(grid_h);
		dt = -2.0 * grid_h;
		t += dt;

		// Write some output for gdisp
		progress = (T - t) / (T - T0);
		display->Progress(0, progress,
				  "Dual solution: %.1f %%. Currently at t = %f",
				  progress * 100.0, t);

		// Set solution from previous time step
		uk_prev_TS = uk;
		sp_pre->Set_uk_prev_TS(&uk_prev_TS);
		sp_vel->Set_uk_prev_TS(&uk_prev_TS);

		// Read primal solution from file
		display->Status(2,
				"Reading data for primal problem at t = %f, on level %i",
				t, input_level_primal);
		//opendx_primal->GetData(&(*GRID) (input_level_primal), &U, t);
		opendx_primal->GetData(grid, &U, t);

		// FIXME: Gridrelations should be created when needed (the first time)
		// FIXME: Should be done automatically
		// FIXME: This projection is not general

		if (U.size(0) != no_nodes)
			(*GRIDREL) (0).projNodalData(U);
		sp_pre->Set_Uprimal(&U);
		sp_vel->Set_Uprimal(&U);

		sp_pre->Set_uk(&uk);
		sp_vel->Set_uk(&uk);

		// Re assemble the momentum matrix or not
		sp_vel->ReAssembleMatrix(true);

		// Start iteration 
		display->Status(2,
				"Start U-P iterations to solve linear problem");
		for (UP_it = 0; UP_it < max_no_UP_iter; UP_it++) {

			// Solve dual continuity equation 
			sp_pre->Timestep(uk, t, dt);

			// Solve dual momentum equation 
			sp_vel->Timestep(uk, t, dt);

			// Re assemble the momentum matrix or not
			sp_vel->ReAssembleMatrix(false);

			// Compute residuals 
			sp_pre->Set_uk(&uk);
			sp_vel->Set_uk(&uk);
			sp_pre->ComputeDiscreteResidual(t, dt,
							pre_res_norm_2,
							pre_res_norm_inf);
			sp_vel->ComputeDiscreteResidual(t, dt,
							vel_res_norm_2,
							vel_res_norm_inf);

			// Check if iteration has converged
			if (CheckUPConvergence())
				break;
		}

		// Write solution to file
		for (i = 0; i < printNoNodes; i++) {
			for (j = 0; j < noeq; j++)
				print_uk(i, j) = uk(i, j);
		}
		//sprintf(tmp,"udualnow.inp"); writeInp(tmp, &(*GRID)(compGrid-2), print_uk, 3, 0);
		//sprintf(tmp,"pdualnow.inp"); writeInp(tmp, &(*GRID)(compGrid-2), print_uk, 1, 3);
		//sprintf(tmp,"u_restart.inp"); writeInp(tmp, &(*GRID)(compGrid), uk, 3, 0);

		// Write data to file
		if (write_data)
			WriteData(&uk, &uk_prev_TS, t, dt);
		if (write_residuals)
			CompContResidual(&uk, &uk_prev_TS, t, dt,
					 contResidual);
	}

	// Solution is computed to end time
	display->Status(0,
			"Dual solution has been computed to endtime T = %f",
			t);

	return 0;
}

//----------------------------------------------------------------------------------------------------
bool TranProblemDualNS::CheckUPConvergence()
{
	bool UP_convergence = false;

	UP_res_norm_2 = sqrt(sqr(pre_res_norm_2) + sqr(vel_res_norm_2));
	if (pre_res_norm_inf > vel_res_norm_inf)
		UP_res_norm_inf = pre_res_norm_inf;
	else
		UP_res_norm_inf = vel_res_norm_inf;

	display->Status(1, "VAL5=R(UP) %.6f", UP_res_norm_2);

	display->Status(4, "l2-norm of UP-residuals: ");
	display->Status(4, "%.6f (momentum equation)", vel_res_norm_2);
	display->Status(4, "%.6f (continuity equation)", pre_res_norm_2);
	display->Status(4, "%.6f (total l2-residual)", UP_res_norm_2);
	display->Status(4, "%.6f (total max-residual)", UP_res_norm_inf);

	if (UP_res_norm_2 < UP_tol)
		UP_convergence = true;
	if ((UP_res_norm_2 > UP_max_tol)
	    && (UP_it == (max_no_UP_iter - 1))) {
		display->Error("UP residual too large.");
		exit(1);
	}

	return UP_convergence;
}

//----------------------------------------------------------------------------------------------------
int TranProblemDualNS::SetInitialData()
{
	display->Status(2, "Set initial conditions");

	uk = 0.0;

	MV_ColMat < real > read_uk;

	char tmp[256];
	real x, y, z, pert, noRolls;
	real pi = KW_PI;
	int i, j;

	switch (initial_data) {
	case zero_initial_data_dual:
		break;
	default:
		display->Error("The initial condition is not implemented");
		exit(1);
	}

	return 0;
}

//----------------------------------------------------------------------------------------------------
int TranProblemDualNS::SetTolerances()
{
	UP_tol = 1.0e-2;
	UP_max_tol = UP_tol;
	max_no_UP_iter = 100;
	UP_conv = true;

	stat_tol = 1.0e-10;

	return 0;
}

//----------------------------------------------------------------------------------------------------
int TranProblemDualNS::CompContResidual(MV_ColMat < real > *U,
					MV_ColMat < real > *Uprev, real t,
					real dt,
					MV_Vector < real > &contResidual)
{
	display->InternalError("TranProblemDualNS::CompContResidual()",
			       "This function should not be used.");
	//neighbor.init((*GRID)(compGrid));

	return 0;
}

//----------------------------------------------------------------------------------------------------
int TranProblemDualNS::WriteData(MV_ColMat < real > *U,
				 MV_ColMat < real > *Uprev, real t,
				 real dt)
{
	Update();

	//Field Ufld_1(&(*GRID) (compGrid));
	//Field Ufld_2(&(*GRID) (compGrid));
	//Field Ufld_3(&(*GRID) (compGrid));
	//Field Pfld(&(*GRID) (compGrid));

	Field Ufld_1(grid);
	Field Ufld_2(grid);
	Field Ufld_3(grid);
	Field Pfld(grid);

	Ufld_1.InitByNodes(U, 0);
	Ufld_2.InitByNodes(U, 1);
	Ufld_3.InitByNodes(U, 2);
	Pfld.InitByNodes(U, 3);

	real L1_norm_U =
	    Ufld_1.Lp_norm(1) + Ufld_2.Lp_norm(1) + Ufld_3.Lp_norm(1);
	real L2_norm_U =
	    sqrt(sqr(Ufld_1.Lp_norm(2)) + sqr(Ufld_2.Lp_norm(2)) +
		 sqr(Ufld_3.Lp_norm(2)));
	real Lmax_norm_U =
	    max(max(Ufld_1.Lmax_norm(), Ufld_2.Lmax_norm()),
		Ufld_3.Lmax_norm());

	real W11_seminorm_U =
	    Ufld_1.Wkp_seminorm(1, 1) + Ufld_2.Wkp_seminorm(1,
							    1) +
	    Ufld_3.Wkp_seminorm(1, 1);
	real W12_seminorm_U =
	    sqrt(sqr(Ufld_1.Wkp_seminorm(1, 2)) +
		 sqr(Ufld_2.Wkp_seminorm(1, 2)) +
		 sqr(Ufld_3.Wkp_seminorm(1, 2)));
	real W1max_seminorm_U =
	    max(max(Ufld_1.Wkmax_seminorm(1), Ufld_2.Wkmax_seminorm(1)),
		Ufld_3.Wkmax_seminorm(1));

	real W11_norm_U = L1_norm_U + W11_seminorm_U;
	real W12_norm_U = sqrt(sqr(L2_norm_U) + sqr(W12_seminorm_U));
	real W1max_norm_U = max(Lmax_norm_U, W1max_seminorm_U);

	real W21_seminorm_U, W22_seminorm_U, W2max_seminorm_U, W21_norm_U,
	    W22_norm_U, W2max_norm_U;
	W21_seminorm_U = W22_seminorm_U = W2max_seminorm_U = W21_norm_U =
	    W22_norm_U = W2max_norm_U = 0.0;
	/*
	   W21_seminorm_U   = Ufld_1.Wkp_seminorm(2,1) + Ufld_2.Wkp_seminorm(2,1) + Ufld_3.Wkp_seminorm(2,1);
	   W22_seminorm_U   = sqrt( sqr(Ufld_1.Wkp_seminorm(2,2)) + sqr(Ufld_2.Wkp_seminorm(2,2)) + sqr(Ufld_3.Wkp_seminorm(2,2)) );
	   W2max_seminorm_U = max( max(Ufld_1.Wkmax_seminorm(2),Ufld_2.Wkmax_seminorm(2)), Ufld_3.Wkmax_seminorm(2) );

	   W21_norm_U   = W11_norm_U + W21_seminorm_U;
	   W22_norm_U   = sqrt( sqr(W12_norm_U) + sqr(W22_seminorm_U) );
	   W2max_norm_U = max( W1max_norm_U, W2max_seminorm_U );
	 */

	real L1_norm_P = Pfld.Lp_norm(1);
	real L2_norm_P = Pfld.Lp_norm(2);
	real Lmax_norm_P = Pfld.Lmax_norm();

	real W11_seminorm_P = Pfld.Wkp_seminorm(1, 1);
	real W12_seminorm_P = Pfld.Wkp_seminorm(1, 2);
	real W1max_seminorm_P = Pfld.Wkmax_seminorm(1);

	real W11_norm_P = L1_norm_P + W11_seminorm_P;
	real W12_norm_P = sqrt(sqr(L2_norm_P) + sqr(W12_seminorm_P));
	real W1max_norm_P = max(Lmax_norm_P, W1max_seminorm_P);

	real W21_seminorm_P, W22_seminorm_P, W2max_seminorm_P, W21_norm_P,
	    W22_norm_P, W2max_norm_P;
	W21_seminorm_P = W22_seminorm_P = W2max_seminorm_P = W21_norm_P =
	    W22_norm_P = W2max_norm_P = 0.0;
	/*
	   W21_seminorm_P   = Pfld.Wkp_seminorm(2,1);
	   W22_seminorm_P   = Pfld.Wkp_seminorm(2,2);
	   W2max_seminorm_P = Pfld.Wkmax_seminorm(2);

	   W21_norm_P   = W11_norm_P + W21_seminorm_P;
	   W22_norm_P   = sqrt( sqr(W12_norm_P) + sqr(W22_seminorm_P) );
	   W2max_norm_P = max( W1max_norm_P, W2max_seminorm_P );
	 */

	// Compute the divergence of the velocity field 
	MV_Vector < real > divU(no_elms);
	int el;
	for (el = 0; el < no_elms; el++)
		divU(el) =
		    Ufld_1.EvalGradient(el, 0) + Ufld_2.EvalGradient(el,
								     1) +
		    Ufld_3.EvalGradient(el, 2);

				  //Field Ufld_div(&(*GRID) (compGrid));
				  Field Ufld_div(grid);
	Ufld_div.InitByElms(&divU);

	real L1_norm_divU = Ufld_div.Lp_norm(1);
	real L2_norm_divU = Ufld_div.Lp_norm(2);
	real Lmax_norm_divU = Ufld_div.Lmax_norm();

	// Compute time derivative   
	MV_ColMat < real > Udot(no_nodes, (*U).size(1));
	int nod, i;
	for (nod = 0; nod < no_nodes; nod++) {
		for (i = 0; i < (*U).size(1); i++)
			Udot(nod, i) =
			    ((*U) (nod, i) - (*Uprev) (nod, i)) / dt;
	}
				  
				  //Field Udotfld_1(&(*GRID) (compGrid));
				  //Field Udotfld_2(&(*GRID) (compGrid));
				  //Field Udotfld_3(&(*GRID) (compGrid));
				  //Field Pdotfld(&(*GRID) (compGrid));

				  Field Udotfld_1(grid);
				  Field Udotfld_2(grid);
				  Field Udotfld_3(grid);
				  Field Pdotfld(grid);

				  
				  
	Udotfld_1.InitByNodes(&Udot, 0);
	Udotfld_2.InitByNodes(&Udot, 1);
	Udotfld_3.InitByNodes(&Udot, 2);
	Pdotfld.InitByNodes(&Udot, 3);

	real L1_norm_Udot =
	    Udotfld_1.Lp_norm(1) + Udotfld_2.Lp_norm(1) +
	    Udotfld_3.Lp_norm(1);
	real L2_norm_Udot =
	    sqrt(sqr(Udotfld_1.Lp_norm(2)) + sqr(Udotfld_2.Lp_norm(2)) +
		 sqr(Udotfld_3.Lp_norm(2)));
	real Lmax_norm_Udot =
	    max(max(Udotfld_1.Lmax_norm(), Udotfld_2.Lmax_norm()),
		Udotfld_3.Lmax_norm());

	real L1_norm_Pdot = Pdotfld.Lp_norm(1);
	real L2_norm_Pdot = Pdotfld.Lp_norm(2);
	real Lmax_norm_Pdot = Pdotfld.Lmax_norm();

	matlab_data->SetTime(t);

	matlab_data->Set(0, L1_norm_U);
	matlab_data->Set(1, L2_norm_U);
	matlab_data->Set(2, Lmax_norm_U);
	matlab_data->Set(3, W11_seminorm_U);
	matlab_data->Set(4, W12_seminorm_U);
	matlab_data->Set(5, W1max_seminorm_U);
	matlab_data->Set(6, W11_norm_U);
	matlab_data->Set(7, W12_norm_U);
	matlab_data->Set(8, W1max_norm_U);
	matlab_data->Set(9, W21_seminorm_U);
	matlab_data->Set(10, W22_seminorm_U);
	matlab_data->Set(11, W2max_seminorm_U);
	matlab_data->Set(12, W21_norm_U);
	matlab_data->Set(13, W22_norm_U);
	matlab_data->Set(14, W2max_norm_U);
	matlab_data->Set(15, L1_norm_P);
	matlab_data->Set(16, L2_norm_P);
	matlab_data->Set(17, Lmax_norm_P);
	matlab_data->Set(18, W11_seminorm_P);
	matlab_data->Set(19, W12_seminorm_P);
	matlab_data->Set(20, W1max_seminorm_P);
	matlab_data->Set(21, W11_norm_P);
	matlab_data->Set(22, W12_norm_P);
	matlab_data->Set(23, W1max_norm_P);
	matlab_data->Set(24, W21_seminorm_P);
	matlab_data->Set(25, W22_seminorm_P);
	matlab_data->Set(26, W2max_seminorm_P);
	matlab_data->Set(27, W21_norm_P);
	matlab_data->Set(28, W22_norm_P);
	matlab_data->Set(29, W2max_norm_P);
	matlab_data->Set(30, L1_norm_divU);
	matlab_data->Set(31, L2_norm_divU);
	matlab_data->Set(32, Lmax_norm_divU);
	matlab_data->Set(33, L1_norm_Udot);
	matlab_data->Set(34, L2_norm_Udot);
	matlab_data->Set(35, Lmax_norm_Udot);
	matlab_data->Set(36, L1_norm_Pdot);
	matlab_data->Set(37, L2_norm_Pdot);
	matlab_data->Set(38, Lmax_norm_Pdot);
	matlab_data->Save();

	return 0;
}

//----------------------------------------------------------------------------------------------------
