// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "StatProblemPressure2d.hh"

//----------------------------------------------------------------------------------------------------
StatProblemPressure2d::StatProblemPressure2d( MV_Vector<Grid>* g ) : StatProblem( g ) 
{
  noeq = 1;

  // Initialize Fields to attach to Equation
  u1_prev_TS.AttachGrid(&(*GRID)(compGrid));
  u2_prev_TS.AttachGrid(&(*GRID)(compGrid));
  p_prev_TS.AttachGrid(&(*GRID)(compGrid));
  
  u1_prev_NL.AttachGrid(&(*GRID)(compGrid));
  u2_prev_NL.AttachGrid(&(*GRID)(compGrid));
  p_prev_NL.AttachGrid(&(*GRID)(compGrid));
  
  u1.AttachGrid(&(*GRID)(compGrid));
  u2.AttachGrid(&(*GRID)(compGrid));
  p.AttachGrid(&(*GRID)(compGrid));
  
  dt.AttachGrid(&(*GRID)(compGrid));
  t.AttachGrid(&(*GRID)(compGrid));
  Re.AttachGrid(&(*GRID)(compGrid));
  
  settings->Get("bnd x0", &bnd_x0);
  settings->Get("bnd x1", &bnd_x1);
  settings->Get("bnd y0", &bnd_y0);
  settings->Get("bnd y1", &bnd_y1);
  settings->Get("bnd z0", &bnd_z0);
  settings->Get("bnd z1", &bnd_z1);
 
  settings->Get("reynolds number", &ReNo);

  settings->Get("periodic bc x", &periodic_bc_x);
  settings->Get("periodic bc y", &periodic_bc_y);

  settings->Get("outflow bc x1", &outflow_bc_x1);

  SetDomain();
}
//----------------------------------------------------------------------------------------------------
int StatProblemPressure2d::Timestep( MV_ColMat<real> &val, real time, real deltat )
{
  
  // define internal variables
  Scalar *temp;
  int ierr,i;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );

  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  //p_prev_TS.InitByNodes(uk_prev_TS,2);

  u1_prev_NL.InitByNodes(uk_prev_NL,0);
  u2_prev_NL.InitByNodes(uk_prev_NL,1);
  //p_prev_NL.InitByNodes(uk_prev_NL,2);
  
  uk = &val;
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  p.InitByNodes(uk,2);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  real int_p;
  p.Integrate( int_p );
  display->Status(2,"Integral of p = %f",int_p);

  Vec u;
  Vec b;
  
  MV_Vector<int> DirichletIdx(0); // Should be of size 0
  MV_Vector<real> DirichletVal(0); // Should be of size 0

  MV_Vector<int> perList;
  

  // FIXME: Removed Multigrid until new Grid class works.
  // FIXME: Also uncomment MGU below!
  bool usingMG = false;
  //MultigridUtilities MGU;
  
  if ( periodic_bc_x || periodic_bc_y ) usingMG = false;
  
  if ( !A_assembled ){
    
    display->Status(2,"Assemble matrix A and vector b.");
    Scalar one = 1.0;
    Mat Atmp;
    A_mat.newsize(compGrid+1,1);	
    int startGrid;
    if ( usingMG ){
      display->Status(4,"Using Multigrid Solver");
      GridProjections.newsize( compGrid,1);
      startGrid = 0;
    }
    else{
      startGrid = compGrid;
    } 
    
    for ( i=startGrid; i < compGrid+1; i++ ){
      discrete->createSeqRHS( &u, &b, (*GRID)(i), noeq );
      if ( exact_integration ){
	if ( periodic_bc_x || periodic_bc_y ){
	  if ( periodic_bc_x ) getPerList( bndx0, bndx1, &(*GRID)(i), perList );
	  if ( periodic_bc_y ) getPerList( bndy0, bndy1, &(*GRID)(i), perList );
	  if ( periodic_bc_x || periodic_bc_y ){
	    discrete->createSeqMatPeriodic( &Atmp, (*GRID)(i), noeq, perList );
	    discrete->discretiseLHSIntExactPeriodicInitial( *equation, (*GRID)(i), Atmp, noeq, perList );
	    discrete->discretiseRHSIntExactPeriodic( *equation, (*GRID)(i), b, noeq, perList );
	  }
	} else{
	  discrete->createSeqMat( &Atmp, (*GRID)(i), noeq );
	  discrete->discretiseLHSIntExact( *equation, (*GRID)(i), Atmp, noeq );
	  discrete->discretiseRHSIntExact( *equation, (*GRID)(i), b, noeq );
	}
      } else{
	discrete->createSeqMat( &Atmp, (*GRID)(i), noeq );
	discrete->discretiseLHS( *equation, *mapper, (*GRID)(i), *quadrule, *element, Atmp );
	discrete->discretiseRHS( *equation, *mapper, (*GRID)(i), *quadrule, *element, b );
      }
      discrete->dropZeroElements(Atmp, A_mat(i,0), 1.0e-12);
      if ( outflow_bc_x1 ){
	if ( i == compGrid ){
	  ierr = outflowBC( A_mat(i,0), b, bndx1, u1, ReNo, &(*GRID)(i), 1, 0, DirichletIdx, DirichletVal );
	} else{
	  ierr = dirichletBC( A_mat(i,0), b, bndx1, 0.0, &(*GRID)(i), 1, 0, DirichletIdx, DirichletVal );
	}
      }
      if ( periodic_bc_x ) ierr = dirichletBC( A_mat(i,0), b, bndx1, 0.0, &(*GRID)(i), 1, 0, DirichletIdx, DirichletVal );
      if ( periodic_bc_y ) ierr = dirichletBC( A_mat(i,0), b, bndy1, 0.0, &(*GRID)(i), 1, 0, DirichletIdx, DirichletVal );
		// FIXME: Here
		//		if ( (i > 0) && usingMG ) {
		//cout << "check 2, " << i << endl;
		//MGU.constructProjectionMatrix( (*GRID)(i-1), (*GRID)(i), GridProjections(i-1,0), (*GRID)(i) );
		//cout << "check 2" << endl;
		// }
      
    }
    
    ierr = MatDestroy(Atmp); CHKERRQ(ierr);
    A = A_mat(compGrid,0);
    /*
      discrete->createSeqMat( &M, (*GRID)(compGrid), noeq );
      discrete->discretiseMatMassExact( (*GRID)(compGrid), *equation, M, noeq );
    */
    
    A_assembled = true;
    
  }
  else{
    discrete->createSeqRHS( &u, &b, (*GRID)(compGrid), noeq );
    display->Status(2,"Assemble vector b");
    if ( exact_integration ){
      if ( periodic_bc_x ) getPerList( bndx0, bndx1, &(*GRID)(compGrid), perList );
      if ( periodic_bc_y ) getPerList( bndy0, bndy1, &(*GRID)(compGrid), perList );
      if ( periodic_bc_x ||  periodic_bc_y ){
	discrete->discretiseRHSIntExactPeriodic( *equation, (*GRID)(compGrid), b, noeq, perList );
      } else{
	discrete->discretiseRHSIntExact( *equation, (*GRID)(compGrid), b, noeq );
      }
    } else{
      discrete->discretiseRHS( *equation, *mapper, (*GRID)(compGrid), *quadrule, *element, b );
    }
  
    if ( outflow_bc_x1 ){
      ierr = outflowBC( b, bndx1, u1, ReNo, &(*GRID)(compGrid), 1, 0, DirichletIdx, DirichletVal );
      //ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 1, 0, DirichletIdx, DirichletVal );
    }
    if ( periodic_bc_x ) ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 1, 0, DirichletIdx, DirichletVal );
    if ( periodic_bc_y ) ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 1, 0, DirichletIdx, DirichletVal );
  }
  
  real norm;
  ierr = VecNorm( b, NORM_2, &norm ); CHKERRQ( ierr ); 
  
  
  // set start solution to previous solution
  MV_Vector<real> uprev( noNodes );
  for ( i=0; i < noNodes; i++ ) uprev(i) = val(i,3);

  MV_Vector<int> idx( noNodes );
  for ( i=0; i < noNodes; i++ ) idx(i) = i;

  ierr = VecSetValues(u,noNodes,&idx(0),&uprev(0),INSERT_VALUES); CHKERRQ( ierr ); 
  
  display->Status(2,"Solve algebraic system of equations");

  // FIXME: And here
  //if ( usingMG ){
  //  ierr = MGU.solveMG(A_mat,GridProjections,b,u,compGrid,1); CHKERRQ(ierr); 
  // }
  //else{
  discrete->solve( A, u, b );
  // }
  
  VecNorm(u,NORM_INFINITY,&norm);
  
  ierr = VecGetArray(u, &temp);
  for ( i=0; i < noNodes; i++ ){
    if ( periodic_bc_x || periodic_bc_y ){
      val(i,2) = temp[perList(i)];
    } else{
      val(i,2) = temp[i];
    }	  
  }
  ierr = VecRestoreArray(u, &temp); CHKERRQ( ierr );
  
  ierr = VecDestroy(u); CHKERRQ(ierr);
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemPressure2d::ComputeDiscreteResidual( real time, real deltat, real& norm_2, real& norm_inf )
{      
  Scalar *temp;
  int ierr;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );

  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  //p_prev_TS.InitByNodes(uk_prev_TS,2);

  u1_prev_NL.InitByNodes(uk_prev_NL,0);
  u2_prev_NL.InitByNodes(uk_prev_NL,1);
  //p_prev_NL.InitByNodes(uk_prev_NL,2);
  
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  p.InitByNodes(uk,2);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  Vec u;
  Vec b;
  
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, &(*GRID)(compGrid), perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, &(*GRID)(compGrid), perList );
  
  MV_Vector<int> DirichletIdx(0); // Should be of size 0
  MV_Vector<real> DirichletVal(0); // Should be of size 0
  
  discrete->createSeqRHS( &u, &b, (*GRID)(compGrid), noeq );
  if ( exact_integration ){
    if ( periodic_bc_x || periodic_bc_y ){
      discrete->discretiseRHSIntExactPeriodic( *equation, (*GRID)(compGrid), b, noeq, perList );
    } else{
      discrete->discretiseRHSIntExact( *equation, (*GRID)(compGrid), b, noeq );
    }
  } else{
    discrete->discretiseRHS( *equation, *mapper, (*GRID)(compGrid), *quadrule, *element, b );
  }
  if ( outflow_bc_x1 ){
    ierr = outflowBC( b, bndx1, u1, ReNo, &(*GRID)(compGrid), 1, 0, DirichletIdx, DirichletVal );
    //ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 1, 0, DirichletIdx, DirichletVal );
  }
  if ( periodic_bc_x ) ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 1, 0, DirichletIdx, DirichletVal );
  if ( periodic_bc_y ) ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 1, 0, DirichletIdx, DirichletVal );
  
  int i;
  MV_Vector<real> uvec( noNodes );
  for ( i=0; i < noNodes; i++ ) uvec(i) = p.GetNodalValue(i);

  MV_Vector<int> idx( noNodes );
  for ( i=0; i < noNodes; i++ ) idx(i) = i;

  ierr = VecSetValues(u,noNodes,&idx(0),&uvec(0),INSERT_VALUES); CHKERRQ( ierr ); 
  
  if ( periodic_bc_x ) ierr = dirichletBC( u, bndx1, 0.0, &(*GRID)(compGrid), 1, 0, DirichletIdx, DirichletVal );
  if ( periodic_bc_y ) ierr = dirichletBC( u, bndy1, 0.0, &(*GRID)(compGrid), 1, 0, DirichletIdx, DirichletVal );
  
  display->Status(2,"Compute residual");
  Vec res;
  VecDuplicate( b, &res);
  Scalar minus = - 1.0;
  VecScale( &minus, b );
  MatMultAdd( A_mat(compGrid,0), u, b, res );
  ierr = VecNorm( res, NORM_2, &norm_2 );
  ierr = VecNorm( res, NORM_INFINITY, &norm_inf );
  
  ierr = VecDestroy(res); CHKERRQ(ierr);
  ierr = VecDestroy(u); CHKERRQ(ierr);
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemPressure2d::InitEquationData( real time, real deltat )
{      
  data.newsize(12);
  
  data(0)  = &u1_prev_TS;
  data(1)  = &u2_prev_TS;
  data(2)  = &p_prev_TS;
  
  data(3)  = &u1_prev_NL;
  data(4)  = &u2_prev_NL;
  data(5)  = &p_prev_NL;
 
  data(6)  = &u1;
  data(7)  = &u2;
  data(8)  = &p;
  
  data(9)  = &dt;
  data(10) = &t;
  data(11) = &Re;
  
  equation->attachField(data);
}
//----------------------------------------------------------------------------------------------------
int StatProblemPressure2d::SetDomain()
{
  bndx0.newsize(nsd,2);
  bndx1.newsize(nsd,2);
  bndy0.newsize(nsd,2);
  bndy1.newsize(nsd,2);
  
  bndx0(0,0) = bnd_x0;
  bndx0(0,1) = bnd_x0;
  bndx0(1,0) = bnd_y0;
  bndx0(1,1) = bnd_y1;
  bndx0(2,0) = bnd_z0;
  bndx0(2,1) = bnd_z1;
  
  bndx1(0,0) = bnd_x1;
  bndx1(0,1) = bnd_x1;
  bndx1(1,0) = bnd_y0;
  bndx1(1,1) = bnd_y1;
  bndx1(2,0) = bnd_z0;
  bndx1(2,1) = bnd_z1;
  
  bndy0(0,0) = bnd_x0;
  bndy0(0,1) = bnd_x1;
  bndy0(1,0) = bnd_y0;
  bndy0(1,1) = bnd_y0;
  bndy0(2,0) = bnd_z0;
  bndy0(2,1) = bnd_z1;
  
  bndy1(0,0) = bnd_x0;
  bndy1(0,1) = bnd_x1;
  bndy1(1,0) = bnd_y1;
  bndy1(1,1) = bnd_y1;
  bndy1(2,0) = bnd_z0;
  bndy1(2,1) = bnd_z1;
  
  return 0;
}
//----------------------------------------------------------------------------------------------------




