// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "StatProblemVelocity2d.hh"

//----------------------------------------------------------------------------------------------------
StatProblemVelocity2d::StatProblemVelocity2d( MV_Vector<Grid>* g ) : StatProblem( g )
{ 
  noeq = 2;

  re_assemble_matrix = true;
  
  // Initialize Fields to attach to Equation
  u1_prev_TS.AttachGrid(&(*GRID)(compGrid));
  u2_prev_TS.AttachGrid(&(*GRID)(compGrid));
  p_prev_TS.AttachGrid(&(*GRID)(compGrid));
  
  u1_prev_NL.AttachGrid(&(*GRID)(compGrid));
  u2_prev_NL.AttachGrid(&(*GRID)(compGrid));
  p_prev_NL.AttachGrid(&(*GRID)(compGrid));
  
  u1.AttachGrid(&(*GRID)(compGrid));
  u2.AttachGrid(&(*GRID)(compGrid));
  p.AttachGrid(&(*GRID)(compGrid));
  
  dt.AttachGrid(&(*GRID)(compGrid));
  t.AttachGrid(&(*GRID)(compGrid));
  Re.AttachGrid(&(*GRID)(compGrid));
  
  settings->Get("compute projections",&compute_projections);

  if ( compute_projections ){
    Field u1_coarse(&(*GRID)(compGrid));
    Field u2_coarse(&(*GRID)(compGrid));
    Field p_coarse(&(*GRID)(compGrid));

    Field u1_fine(&(*GRID)(compGrid));
    Field u2_fine(&(*GRID)(compGrid));
    Field p_fine(&(*GRID)(compGrid));
  }

  settings->Get("bnd x0", &bnd_x0);
  settings->Get("bnd x1", &bnd_x1);
  settings->Get("bnd y0", &bnd_y0);
  settings->Get("bnd y1", &bnd_y1);
  settings->Get("bnd z0", &bnd_z0);
  settings->Get("bnd z1", &bnd_z1);
 
  settings->Get("reynolds number", &ReNo);

  settings->Get("driven cavity", &driven_cavity);
  settings->Get("backward facing step", &backward_facing_step);
  settings->Get("bluff body", &bluff_body);
  settings->Get("jet", &jet);

  settings->Get("slip bc x0", &slip_bc_x0);
  settings->Get("slip bc x1", &slip_bc_x1);
  settings->Get("slip bc y0", &slip_bc_y0);
  settings->Get("slip bc y1", &slip_bc_y1);

  settings->Get("no slip bc x0", &no_slip_bc_x0);
  settings->Get("no slip bc x1", &no_slip_bc_x1);
  settings->Get("no slip bc y0", &no_slip_bc_y0);
  settings->Get("no slip bc y1", &no_slip_bc_y1);

  settings->Get("periodic bc x", &periodic_bc_x);
  settings->Get("periodic bc y", &periodic_bc_y);

  settings->Get("inflow bc x0", &inflow_bc_x0);
  settings->Get("outflow bc x0", &outflow_bc_x0);
  settings->Get("outflow bc x1", &outflow_bc_x1);

  settings->Get("couette bc y", &couette_bc_y);

  settings->Get("driven cavity bc x0", &driven_cavity_bc_x0);
  settings->Get("driven cavity bc y1", &driven_cavity_bc_y1);

  settings->Get("dual drag bc", &dual_drag_bc);

  // define the computational domain
  SetDomain();
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity2d::Timestep( MV_ColMat<real>& val, real time, real deltat )
{
  // initialize inner variables 
  Scalar *temp;
  int ierr;
  
  Vec u;
  Vec b;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );

  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  //p_prev_TS.InitByNodes(uk_prev_TS,2);

  u1_prev_NL.InitByNodes(uk_prev_NL,0);
  u2_prev_NL.InitByNodes(uk_prev_NL,1);
  //p_prev_NL.InitByNodes(uk_prev_NL,2);
  
  uk = &val;
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  p.InitByNodes(uk,2);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  if ( compute_projections ){
    u1_coarse.InitByNodes(uk_coarse,0);
    u2_coarse.InitByNodes(uk_coarse,1);
    p_coarse.InitByNodes(uk_coarse,2);

    u1_fine.InitByNodes(uk_fine,0);
    u2_fine.InitByNodes(uk_fine,1);
    p_fine.InitByNodes(uk_fine,2);
  }

  // get master-slave info if periodic bc
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, &(*GRID)(compGrid), perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, &(*GRID)(compGrid), perList );

  // assemble matrix if not assembled
  if ( !A_assembled ){
    display->Status(2,"Assemble matrix A");
    // check if exact integration / quadrature is used
    if ( exact_integration ){
      // check if periodic bc
      if ( periodic_bc_x || periodic_bc_y ){
	discrete->createSeqMatPeriodic( &A, (*GRID)(compGrid), noeq, perList );
	discrete->discretiseLHSIntExactPeriodicInitial( *equation, (*GRID)(compGrid), A, noeq, perList );
      } else{
	discrete->createSeqMat( &A, (*GRID)(compGrid), noeq );
	discrete->discretiseLHSIntExact( *equation, (*GRID)(compGrid), A, noeq );
      }
    } else{
      discrete->createSeqMat( &A, (*GRID)(compGrid), noeq );
      discrete->discretiseLHS( *equation, *mapper, (*GRID)(compGrid), *quadrule, *element, A );
    }
    // if assembled, check if to reassemble
  } else if ( A_assembled && re_assemble_matrix ){
    display->Status(2,"Assemble matrix A");
    ierr = MatZeroEntries(A);
    // check if exact integration / quadrature is used
    if ( exact_integration ){
      // check if periodic bc
      if ( periodic_bc_x || periodic_bc_y ){
	discrete->discretiseLHSIntExactPeriodic( *equation, (*GRID)(compGrid), A, noeq, perList );
      } else{
	discrete->discretiseLHSIntExact( *equation, (*GRID)(compGrid), A, noeq );
      }
    } else{
      discrete->discretiseLHS( *equation, *mapper, (*GRID)(compGrid), *quadrule, *element, A );
    }
  }
  
  display->Status(2,"Assemble vector b");
  discrete->createSeqRHS( &u, &b, (*GRID)(compGrid), noeq );
  if ( exact_integration ){
    // check if periodic bc
    if ( periodic_bc_x || periodic_bc_y ){
      discrete->discretiseRHSIntExactPeriodic( *equation, (*GRID)(compGrid), b, noeq, perList );
    } else{
      discrete->discretiseRHSIntExact( *equation, (*GRID)(compGrid), b, noeq );
    }
  } else{
    discrete->discretiseRHS( *equation, *mapper, (*GRID)(compGrid), *quadrule, *element, b );
  }

  // compute norm of rhs 
  real norm;
  ierr = VecNorm( b, NORM_2, &norm );
  display->Status(2,"Load norm: %f",norm);
  
  // set dirichlet bc
  SetBoundaryConditions(A,b);
  
  // compute norm of rhs 
  ierr = VecNorm( b, NORM_2, &norm );
  display->Status(2,"Load norm: %f (with b.c.)",norm);
  
  // set start solution to solution from previous iteration 
  MV_Vector<real> uprev( noeq * noNodes );
  int i;
  for ( i=0; i < noNodes; i++ ){
    uprev(i*noeq)     = val(i,0);
    uprev(i*noeq + 1) = val(i,1);
  }
  MV_Vector<int> idx( noeq * noNodes );
  for ( i=0; i < noeq * noNodes; i++ ) idx(i) = i;
  
  ierr = VecSetValues(u,noeq*noNodes,&idx(0),&uprev(0),INSERT_VALUES); CHKERRQ( ierr ); 

  // solve discrete system  
  display->Status(2,"Solve algebraic system of equations");
  discrete->solve( A, u, b );
  
  // compute norm of solution
  VecNorm(u,NORM_INFINITY,&norm);
  display->Status(4,"Norm of u = %f",norm);
  MatNorm(A,NORM_INFINITY,&norm);
  display->Status(4,"Norm of A = %f",norm);

  // update solution vector
  ierr = VecGetArray(u, &temp); CHKERRQ( ierr );
  int j;
  for ( i=0; i < noNodes; i++ ){
    for ( j=0; j < nsd; j++ ){ 
      if ( periodic_bc_x || periodic_bc_y ){
	val(i,j) = temp[j+perList(i)*noeq];
      } else{
	val(i,j) = temp[j+i*noeq];
      }
    }
  }
  ierr = VecRestoreArray(u, &temp); CHKERRQ( ierr );

  // destroy solution variables 
  ierr = VecDestroy(u); CHKERRQ(ierr);
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity2d::AssembleMatrix( MV_ColMat<real>& val, real time, real deltat, Mat& Matrix )
{
  // initialize inner variables 
  Scalar *temp;
  int ierr;
  
  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );

  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  //p_prev_TS.InitByNodes(uk_prev_TS,2);

  u1_prev_NL.InitByNodes(uk_prev_NL,0);
  u2_prev_NL.InitByNodes(uk_prev_NL,1);
  //p_prev_NL.InitByNodes(uk_prev_NL,2);
  
  uk = &val;
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  p.InitByNodes(uk,2);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  if ( compute_projections ){
    u1_coarse.InitByNodes(uk_coarse,0);
    u2_coarse.InitByNodes(uk_coarse,1);
    p_coarse.InitByNodes(uk_coarse,2);

    u1_fine.InitByNodes(uk_fine,0);
    u2_fine.InitByNodes(uk_fine,1);
    p_fine.InitByNodes(uk_fine,2);
  }

  // get master-slave info if periodic bc
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, &(*GRID)(compGrid), perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, &(*GRID)(compGrid), perList );

  // assemble matrix if not assembled
  if ( !A_assembled ){
    display->Status(2,"Assemble matrix A");
    // check if exact integration / quadrature is used
    if ( exact_integration ){
      // check if periodic bc
      if ( periodic_bc_x || periodic_bc_y ){
	discrete->createSeqMatPeriodic( &A, (*GRID)(compGrid), noeq, perList );
	discrete->discretiseLHSIntExactPeriodicInitial( *equation, (*GRID)(compGrid), A, noeq, perList );
      } else{
	discrete->createSeqMat( &A, (*GRID)(compGrid), noeq );
	discrete->discretiseLHSIntExact( *equation, (*GRID)(compGrid), A, noeq );
      }
    } else{
      discrete->createSeqMat( &A, (*GRID)(compGrid), noeq );
      discrete->discretiseLHS( *equation, *mapper, (*GRID)(compGrid), *quadrule, *element, A );
    }
    // if assembled, check if to reassemble
  } else if ( A_assembled && re_assemble_matrix ){
    display->Status(2,"Assemble matrix A");
    ierr = MatZeroEntries(A);
    // check if exact integration / quadrature is used
    if ( exact_integration ){
      // check if periodic bc
      if ( periodic_bc_x || periodic_bc_y ){
	discrete->discretiseLHSIntExactPeriodic( *equation, (*GRID)(compGrid), A, noeq, perList );
      } else{
	discrete->discretiseLHSIntExact( *equation, (*GRID)(compGrid), A, noeq );
      }
    } else{
      discrete->discretiseLHS( *equation, *mapper, (*GRID)(compGrid), *quadrule, *element, A );
    }
  }

  // FIXME: Temp sol.
  Vec b;
  discrete->createSeqRHS( &b, (*GRID)(compGrid), noeq );

  // set dirichlet bc
  SetBoundaryConditions(A,b);

  Matrix = A;
  
  // destroy solution variables 
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity2d::ComputeDiscreteResidual( real time, real deltat, real& norm_2, real& norm_inf )
{
  int ierr;
  
  Vec u;
  Vec b;

  // Initialize Fields to be attached to equation
  InitEquationData( time, deltat );

  u1_prev_TS.InitByNodes(uk_prev_TS,0);
  u2_prev_TS.InitByNodes(uk_prev_TS,1);
  //p_prev_TS.InitByNodes(uk_prev_TS,2);

  u1_prev_NL.InitByNodes(uk_prev_NL,0);
  u2_prev_NL.InitByNodes(uk_prev_NL,1);
  //p_prev_NL.InitByNodes(uk_prev_NL,2);
  
  u1.InitByNodes(uk,0);
  u2.InitByNodes(uk,1);
  p.InitByNodes(uk,2);
  
  dt.InitByConstant(&deltat);
  t.InitByConstant(&time);
  Re.InitByConstant(&ReNo);

  if ( compute_projections ){
    u1_coarse.InitByNodes(uk_coarse,0);
    u2_coarse.InitByNodes(uk_coarse,1);
    p_coarse.InitByNodes(uk_coarse,2);

    u1_fine.InitByNodes(uk_fine,0);
    u2_fine.InitByNodes(uk_fine,1);
    p_fine.InitByNodes(uk_fine,2);
  }

  // get master-slave info if periodic bc
  MV_Vector<int> perList;
  if ( periodic_bc_x ) getPerList( bndx0, bndx1, &(*GRID)(compGrid), perList );
  if ( periodic_bc_y ) getPerList( bndy0, bndy1, &(*GRID)(compGrid), perList );

  if ( re_assemble_matrix ){
    display->Status(2,"Assemble matrix A");
    ierr = MatZeroEntries(A);
    if ( exact_integration ){
      if ( periodic_bc_x || periodic_bc_y ){
	discrete->discretiseLHSIntExactPeriodic( *equation, (*GRID)(compGrid), A, noeq, perList );
      } else{
	discrete->discretiseLHSIntExact( *equation, (*GRID)(compGrid), A, noeq );
      }
    } else{
      discrete->discretiseLHS( *equation, *mapper, (*GRID)(compGrid), *quadrule, *element, A );
    }
  }
  
  display->Status(2,"Assemble vector b");
  discrete->createSeqRHS( &u, &b, (*GRID)(compGrid), noeq );
  if ( exact_integration ){
    if ( periodic_bc_x || periodic_bc_y ){
      discrete->discretiseRHSIntExactPeriodic( *equation, (*GRID)(compGrid), b, noeq, perList );
    } else{
      discrete->discretiseRHSIntExact( *equation, (*GRID)(compGrid), b, noeq );
    }
  } else{
    discrete->discretiseRHS( *equation, *mapper, (*GRID)(compGrid), *quadrule, *element, b );	
  }
  
  // set dirichlet bc
  SetBoundaryConditions(A,b);
  
  MV_Vector<real> uprev( noeq * noNodes );
  int i;
  for ( i=0; i < noNodes; i++ ){
    uprev(i*noeq)     = u1.GetNodalValue(i);
    uprev(i*noeq + 1) = u2.GetNodalValue(i);
  }
  MV_Vector<int> idx( noeq * noNodes );
  for ( i=0; i < noeq * noNodes; i++ ) idx(i) = i;
  ierr = VecSetValues(u,noeq*noNodes,&idx(0),&uprev(0),INSERT_VALUES); CHKERRQ( ierr ); 
  
  if ( periodic_bc_x || periodic_bc_y ){
    MV_Vector<int> DirichletIdx(0);  // Should be of size 0
    MV_Vector<real> DirichletVal(0); // Should be of size 0
    ierr = dirichletBC( u, bndx1, 0.0, &(*GRID)(compGrid), 3, 0, DirichletIdx, DirichletVal );
    ierr = dirichletBC( u, bndx1, 0.0, &(*GRID)(compGrid), 3, 1, DirichletIdx, DirichletVal );
    ierr = dirichletBC( u, bndx1, 0.0, &(*GRID)(compGrid), 3, 2, DirichletIdx, DirichletVal );		
  }
  
  Vec res;
  VecDuplicate( b, &res);
  Scalar minus = - 1.0;
  VecScale( &minus, b ); 
  MatMultAdd( A, u, b, res );
  ierr = VecNorm( res, NORM_2, &norm_2 );
  ierr = VecNorm( res, NORM_INFINITY, &norm_inf );
  
  ierr = VecDestroy(res); CHKERRQ(ierr);
  ierr = VecDestroy(u); CHKERRQ(ierr);
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity2d::InitEquationData( real time, real deltat )
{      
  data.newsize(12);
  if ( compute_projections ) data.newsize(18);


  data(0)  = &u1_prev_TS;
  data(1)  = &u2_prev_TS;
  data(2)  = &p_prev_TS;
  
  data(3)  = &u1_prev_NL;
  data(4)  = &u2_prev_NL;
  data(5)  = &p_prev_NL;
  
  data(6)  = &u1;
  data(7)  = &u2;
  data(8) = &p;
  
  data(9) = &dt;
  data(10) = &t;
  data(11) = &Re;
  
  if ( compute_projections ){
    data(12) = &u1_coarse;
    data(13) = &u2_coarse;
    data(14) = &p_coarse;

    data(15) = &u1_fine;
    data(16) = &u2_fine;
    data(17) = &p_fine;
  }

  equation->attachField(data);
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity2d::SetBoundaryConditions( Mat& A, Vec& b )
{
  // set dirichlet bc
  MV_Vector<int> DirichletIdx(0); // Should be of size 0
  MV_Vector<real> DirichletVal(0); // Should be of size 0
  
  int ierr;
  
  if ( !re_assemble_matrix && A_assembled ){
    
    if ( backward_facing_step ){
      ierr = dirichletBC( b, bfs_bndx1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndx1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, bfs_bndy1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }	  
    
    if ( jet ){
      ierr = dirichletBC_GRID( b, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( b, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    
    if ( bluff_body ){
      ierr = dirichletBC( b, ob_bndx0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndx1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }  
    
    if ( couette_bc_y ){
      ierr = dirichletBC( b, bndy1, 1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, bndy0, -1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0,  0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    
    if ( inflow_bc_x0 ){
      ierr = dirichletBCInflow( b, bndx0, &(*GRID)(compGrid), noeq, 0, 1.0, &turbulence, DirichletIdx, DirichletVal );
      ierr = dirichletBCInflow( b, bndx0, &(*GRID)(compGrid), noeq, 1, 0.0, &turbulence, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 1.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    
    if ( slip_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_x1 ){
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x1 ){ 
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y0 ){
      ierr = dirichletBC( b, bndy0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y0 ){
      ierr = dirichletBC( b, bndy0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y1 ){
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y1 ){
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
        
    if ( periodic_bc_x ){
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    
    if ( periodic_bc_y ){
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    
    if ( dual_drag_bc ){
      ierr = dirichletBC( b, ob_bndx0, 1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndx1, 1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy0, 1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy1, 1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }  
    
  } else {
    
    
    if ( backward_facing_step ){
      ierr = dirichletBC( A, b, bfs_bndx1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bfs_bndx1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, bfs_bndy1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bfs_bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }	  
    
    if ( jet ){
      ierr = dirichletBC_GRID( A, b, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( A, b, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }

    if ( bluff_body ){
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }  
    
    if ( couette_bc_y ){
      ierr = dirichletBC( A, b, bndy1, 1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, bndy0, -1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy0,  0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    
    if ( inflow_bc_x0 ){
      ierr = dirichletBCInflow( b, bndx0, &(*GRID)(compGrid), noeq, 0, 1.0, &turbulence, DirichletIdx, DirichletVal );
      ierr = dirichletBCInflow( b, bndx0, &(*GRID)(compGrid), noeq, 1, 0.0, &turbulence, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_x0 ){
      ierr = dirichletBC( A, b, bndx0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx0, 1.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    
    if ( slip_bc_x0 ){
      ierr = dirichletBC( A, b, bndx0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x0 ){
      ierr = dirichletBC( A, b, bndx0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_x1 ){
      ierr = dirichletBC( A, b, bndx1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x1 ){ 
      ierr = dirichletBC( A, b, bndx1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y0 ){
      ierr = dirichletBC( A, b, bndy0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y0 ){
      ierr = dirichletBC( A, b, bndy0, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y1 ){
      ierr = dirichletBC( A, b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y1 ){
      ierr = dirichletBC( A, b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
        
    if ( periodic_bc_x ){
      ierr = dirichletBC( A, b, bndx1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndx1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
   
    if ( periodic_bc_y ){
      ierr = dirichletBC( A, b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
    }
   
    if ( dual_drag_bc ){
      ierr = dirichletBC( A, b, ob_bndx0, 1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndx1, 1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndx1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy0, 1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy0, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( A, b, ob_bndy1, 1.0, &(*GRID)(compGrid), noeq, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( A, b, ob_bndy1, 0.0, &(*GRID)(compGrid), noeq, 1, DirichletIdx, DirichletVal );
      
    }  
    
    A_assembled = true;
    
  }	
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity2d::SetDomain()
{
  
  bndx0.newsize(nsd,2);
  bndx1.newsize(nsd,2);
  bndy0.newsize(nsd,2);
  bndy1.newsize(nsd,2);
  
  ob_bndx0.newsize(nsd,2);
  ob_bndx1.newsize(nsd,2);
  ob_bndy0.newsize(nsd,2);
  ob_bndy1.newsize(nsd,2);
  
  bfs_bndx1.newsize(nsd,2);
  bfs_bndy1.newsize(nsd,2);
  
  bndx0(0,0) = bnd_x0;
  bndx0(0,1) = bnd_x0;
  bndx0(1,0) = bnd_y0;
  bndx0(1,1) = bnd_y1;
  bndx0(2,0) = bnd_z0;
  bndx0(2,1) = bnd_z1;
  
  bndx1(0,0) = bnd_x1;
  bndx1(0,1) = bnd_x1;
  bndx1(1,0) = bnd_y0;
  bndx1(1,1) = bnd_y1;
  bndx1(2,0) = bnd_z0;
  bndx1(2,1) = bnd_z1;
  
  bndy0(0,0) = bnd_x0;
  bndy0(0,1) = bnd_x1;
  bndy0(1,0) = bnd_y0;
  bndy0(1,1) = bnd_y0;
  bndy0(2,0) = bnd_z0;
  bndy0(2,1) = bnd_z1;
  
  bndy1(0,0) = bnd_x0;
  bndy1(0,1) = bnd_x1;
  bndy1(1,0) = bnd_y1;
  bndy1(1,1) = bnd_y1;
  bndy1(2,0) = bnd_z0;
  bndy1(2,1) = bnd_z1;
  
  ob_bndx0(0,0) = 0.375;
  ob_bndx0(0,1) = 0.375;
  ob_bndx0(1,0) = 0.375;
  ob_bndx0(1,1) = 0.625;
  
  ob_bndx1(0,0) = 0.625;
  ob_bndx1(0,1) = 0.625;
  ob_bndx1(1,0) = 0.375;
  ob_bndx1(1,1) = 0.625;
  
  ob_bndy0(0,0) = 0.375;
  ob_bndy0(0,1) = 0.625;
  ob_bndy0(1,0) = 0.375;
  ob_bndy0(1,1) = 0.375;
  
  ob_bndy1(0,0) = 0.375;
  ob_bndy1(0,1) = 0.625;
  ob_bndy1(1,0) = 0.625;
  ob_bndy1(1,1) = 0.625;
  
  bfs_bndx1(0,0) = 0.5;
  bfs_bndx1(0,1) = 0.5;
  bfs_bndx1(1,0) = 0.0;
  bfs_bndx1(1,1) = 0.5;
  
  bfs_bndy1(0,0) = 0.0;
  bfs_bndy1(0,1) = 0.5;
  bfs_bndy1(1,0) = 0.5;
  bfs_bndy1(1,1) = 0.5;
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity2d::ReAssembleMatrix( bool am )
{
  re_assemble_matrix = am;
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int StatProblemVelocity2d::SetTurbulentInflow( real trb )
{
  int ierr;
  int i,j;
  int *bndNodes; 
  MV_Vector<int> bNodes(noNodes);
  int noBndNodes;

  ierr = getBndNodes(bNodes, &noBndNodes, bndx0, &(*GRID)(compGrid));

  turbulence.newsize(noBndNodes,noeq);
  for ( i=0; i < noBndNodes; i++ ){
    for ( j=0; j < noeq; j++ ){
      turbulence(i,j) = trb * 2.0*(drand48() - 0.5);
    }
  }
  
  return 0;
}
//----------------------------------------------------------------------------------------------------


