// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "TranProblemNS2d.hh"

//----------------------------------------------------------------------------------------------------
TranProblemNS2d::TranProblemNS2d( MV_Vector<Grid>* G, MV_Vector<GridRelation>* P ) : TranProblem(G,P) 
{
  // Open the OpenDX file
  char output_file[KW_LINELENGTH];
  settings->Get("output file primal",output_file);
  opendx = new OpenDX(output_file,2,2,1);
  opendx->SetLabel(0,"Velocity");
  opendx->SetLabel(1,"Pressure");
  
  // Get inner variables
  settings->Get("reynolds number",&reynolds_number);
  settings->Get("compute projections",&compute_projections);

  // Initialize output data matlab files
  no_data_entries = 0;
  no_residual_entries = 0;
  if ( write_data )     no_data_entries        = 39;
  if ( write_residuals) no_residual_entries    = 12;
  matlab_data = NULL;
  if ( write_data || write_residuals ){
    matlab_data = new Matlab(no_data_entries + no_residual_entries);
  }
  
  if ( write_data ){
    matlab_data->SetLabel(0,"|| u ||_1");
    matlab_data->SetLabel(1,"|| u ||_2");
    matlab_data->SetLabel(2,"|| u ||_{max}");
    matlab_data->SetLabel(3,"| u |_{1,1}");
    matlab_data->SetLabel(4,"| u |_{1,2}");
    matlab_data->SetLabel(5,"| u |_{1,max}");
    matlab_data->SetLabel(6,"|| u ||_{1,1}");
    matlab_data->SetLabel(7,"|| u ||_{1,2}");
    matlab_data->SetLabel(8,"|| u ||_{1,max}");
    matlab_data->SetLabel(9,"| u |_{2,1}");
    matlab_data->SetLabel(10,"| u |_{2,2}");
    matlab_data->SetLabel(11,"| u |_{2,max}");
    matlab_data->SetLabel(12,"|| u ||_{2,1}");
    matlab_data->SetLabel(13,"|| u ||_{2,2}");
    matlab_data->SetLabel(14,"|| u ||_{2,max}");
    matlab_data->SetLabel(15,"|| p ||_1");
    matlab_data->SetLabel(16,"|| p ||_2");
    matlab_data->SetLabel(17,"|| p ||_{max}");
    matlab_data->SetLabel(18,"| p |_{1,1}");
    matlab_data->SetLabel(19,"| p |_{1,2}");
    matlab_data->SetLabel(20,"| p |_{1,max}");
    matlab_data->SetLabel(21,"|| p ||_{1,1}");
    matlab_data->SetLabel(22,"|| p ||_{1,2}");
    matlab_data->SetLabel(23,"|| p ||_{1,max}");
    matlab_data->SetLabel(24,"| p |_{2,1}");
    matlab_data->SetLabel(25,"| p |_{2,2}");
    matlab_data->SetLabel(26,"| p |_{2,max}");
    matlab_data->SetLabel(27,"|| p ||_{2,1}");
    matlab_data->SetLabel(28,"|| p ||_{2,2}");
    matlab_data->SetLabel(29,"|| p ||_{2,max}");
    matlab_data->SetLabel(30,"|| div u ||_1");
    matlab_data->SetLabel(31,"|| div u ||_2");
    matlab_data->SetLabel(32,"|| div u ||_{max}");
    matlab_data->SetLabel(33,"|| du/dt ||_1");
    matlab_data->SetLabel(34,"|| du/dt ||_2");
    matlab_data->SetLabel(35,"|| du/dt ||_{max}");
    matlab_data->SetLabel(36,"|| dp/dt ||_1");
    matlab_data->SetLabel(37,"|| dp/dt ||_2");
    matlab_data->SetLabel(38,"|| dp/dt ||_{max}");
  }

  if ( write_residuals ){
    matlab_data->SetLabel(no_data_entries+0,"|| R1 ||_1");
    matlab_data->SetLabel(no_data_entries+1,"|| R1 ||_2");
    matlab_data->SetLabel(no_data_entries+2,"|| R1 ||_{max}");
    matlab_data->SetLabel(no_data_entries+3,"|| R2 ||_1");
    matlab_data->SetLabel(no_data_entries+4,"|| R2 ||_2");
    matlab_data->SetLabel(no_data_entries+5,"|| R2 ||_{max}");
    matlab_data->SetLabel(no_data_entries+6,"|| R3 ||_1");
    matlab_data->SetLabel(no_data_entries+7,"|| R3 ||_2");
    matlab_data->SetLabel(no_data_entries+8,"|| R3 ||_{max}");
    matlab_data->SetLabel(no_data_entries+9,"|| R4 ||_1");
    matlab_data->SetLabel(no_data_entries+10,"|| R4 ||_2");
    matlab_data->SetLabel(no_data_entries+11,"|| R4 ||_{max}");
  }

}
//----------------------------------------------------------------------------------------------------
TranProblemNS2d::~TranProblemNS2d()
{
  // Delete the OpenDX file
  delete opendx;
}      
//----------------------------------------------------------------------------------------------------
int TranProblemNS2d::Solve()
{
  char tmp[256];
  int i,j,k;
  
  display->Status(0,"Solving Navier-Stokes primal problem.");
  
  real progress = 0.0;
  
  // Update internal variables
  Update();
  
  // Reset OpenDX files
  opendx->Reset();
  opendx_residuals->Reset();
  
  // Reset Matlab file
  if ( matlab_data )
    matlab_data->Reset();

  // Set time t to starting time T0
  t = T0;

  // Set tolerances for solution method
  SetTolerances();

  // Check output levels
  CheckOutputLevel();

  // Reset output files
  opendx->Reset();
  opendx_residuals->Reset();
  
  // Set FEM discretisation 
  Discretiser          discrete;
  Mapper               mapper(nsd);
  ElementTriLin        element;
  QuadratureRules      quadrule(TET,1);
  EquationVel2d_cG1cG1 eq_vel;
  EquationPre2d_cG1cG1 eq_pre;

  // Set number of equations 
  noeq_vel = eq_vel.noEq();
  noeq_pre = eq_pre.noEq();
  noeq     = noeq_vel + noeq_pre;
  
  // Initialize solution vectors
  uk.newsize(no_nodes,noeq);
  uk_prev_TS.newsize(no_nodes,noeq); 
  uk_prev_NL.newsize(no_nodes,noeq); 
  print_uk.newsize(printNoNodes,noeq); 

  // Set initial data
  initial_data = zero_initial_data;
  SetInitialData();

  // Initialize stationary problems
  sp_vel = new StatProblemVelocity2d( GRID ); 
  sp_pre = new StatProblemPressure2d( GRID ); 

  // Attach FEM discretisation to stationary problem
  sp_pre->AttachFEM( &discrete, &mapper, &quadrule, &element, &eq_pre );
  sp_vel->AttachFEM( &discrete, &mapper, &quadrule, &element, &eq_vel );

  // Initialize projections (for subgrid modeling)
  if ( compute_projections ){
    uk_fine.newsize(no_nodes,noeq); 
    uk_coarse.newsize(no_nodes,noeq); 
  }

  // Initialize subgrid modeling utilities
  SetSubgridModelingUtilities();

  // Reassemble the momentum matrix or not
  sp_vel->ReAssembleMatrix(true);
  
  // Write data to file
  display->Status(2,"Write data to file");
  if ( write_data )      WriteData( &uk, &uk_prev_TS, t, dt );
  if ( write_residuals ) CompContResidual( &uk, &uk_prev_TS, t, dt, contResidual );
    
  // Start time marching 
  display->Status(2,"Start computation at t = %.1f",t);
  while ( t <= T ){
    
    // Save solution if we should
    if ( SaveSample() ){
      
      // Write solution to file
      for ( i=0; i < printNoNodes; i++ ){
	for ( j=0; j < noeq; j++ ) print_uk(i,j) = uk(i,j);
      }
      opendx->AddFrame(&(*GRID)(printGrid),&print_uk,t);
      
      // Write Reynolds stresses to file
      if ( write_reynolds_stresses ){
	for ( i=0; i < N; i++ ){
	  (*GRIDREL)(1).projNestedElData(tau,tau_print);
	  sprintf(tmp,"tau%i_%.6f.inp",i+1,t); writeInp(tmp, &(*GRID)(compGrid-1), tau_print, 1, i);
	  sprintf(tmp,"tau%i_%.6f.m",i+1,t); writeMfile(tmp, &(*GRID)(compGrid-1), tau_print, 1, i);
	}	
      }
    
    }
    
    // Set time step
    GetMeshSize(grid_h);
    dt = grid_h;
    t += dt;
    
    // Write some output for gdisp
    progress = (t-T0) / (T-T0);
    display->Progress(0,progress,"Primal solution: %.1f %%. Currently at t = %f",progress*100.0,t);
	 
    // Set solution from previous time step
    uk_prev_TS = uk;
    sp_pre->Set_uk_prev_TS(&uk_prev_TS);
    sp_vel->Set_uk_prev_TS(&uk_prev_TS);
    
    // Set solution from previous outer non linear iteration
    uk_prev_NL = uk;
    sp_pre->Set_uk_prev_NL(&uk_prev_NL);
    sp_vel->Set_uk_prev_NL(&uk_prev_NL);
    
    // Set coarse-fine projections for subgrid model
    if ( compute_projections ) sp_vel->SetMultiScale(&uk_coarse,&uk_fine);
    
    // Solve non linear problem
    SolveNonLinearProblemUsingFixPoint();
	 
    // Write solution to file
    for ( i=0; i < printNoNodes; i++ ){
      for ( j=0; j < noeq; j++ ) print_uk(i,j) = uk(i,j);
    }
    //sprintf(tmp,"unow.inp"); writeInp(tmp, &(*GRID)(compGrid-1), print_uk, 3, 0);
    //sprintf(tmp,"pnow.inp"); writeInp(tmp, &(*GRID)(compGrid-1), print_uk, 1, 3);
    //sprintf(tmp,"u_restart.inp"); writeInp(tmp, &(*GRID)(compGrid), uk, 3, 0);
    
    // Write data to file
    display->Status(2,"Write data to file");
    if ( write_data )      WriteData( &uk, &uk_prev_TS, t, dt );
    if ( write_residuals ) CompContResidual( &uk, &uk_prev_TS, t, dt, contResidual );
    
    // Check if the solution is stationary
    if ( CheckIfStationary() ) break;
  }

  // Solution computed to end time
  display->Status(0,"Solution has been computed to endtime T = %f",t);

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS2d::SolveNonLinearProblemUsingFixPoint()
{
  // Start non linear iteration
  for ( NL_it=0; NL_it < max_no_NL_iter; NL_it++ ){ 
      
    // Display information
    display->Status(1,"Starting non-linear iteration %d.",NL_it+1);
    
    // Compute Reynolds stresses - for subgrid model
    if ( compute_reynolds_stresses ){
      display->Status(5,"Estimate Reynolds stresses (Update subgrid model)");
      //scale_expol.compRS( uk, tau, (*GRIDREL)(0), (*GRIDREL)(1), (*GRIDREL)(2), h, H, HH );
      //scale_expol.compRSCov( uk, tau, (*GRIDREL)(0), (*GRIDREL)(1), (*GRIDREL)(2) );
      scale_expol.ComputeReynoldsStresses( &uk, &tau, &(*GRIDREL)(0), &(*GRIDREL)(1), 
					   &(*GRIDREL)(2), &(*GRIDREL)(3), &(*GRIDREL)(4) );
    }
    
    // Compute projections - for subgrid model
    int i,j;
    if ( compute_projections ){
      display->Status(5,"Extract coarse and fine scale field");
      (*GRIDREL)(0).projNodalData(uk,uk_coarse);
      (*GRIDREL)(0).projNodalData(uk_coarse);
      for ( i=0; i < no_nodes; i++ ){
	for ( j=0; j < noeq; j++ ) uk_fine(i,j) = uk(i,j) - uk_coarse(i,j);
      }
    }
    
    // Start inner iteration
    for ( UP_it=0; UP_it < max_no_UP_iter; UP_it++ ){
      
      // Solve continuity equation
      display->Status(2,"Solving for the pressure.");
      sp_pre->Timestep( uk, t, dt );
      
      // Solve momentum equation
      display->Status(2,"Solving for the velocity.");
      sp_vel->Timestep( uk, t, dt );
      
      // Re assemble the momentum matrix or not
      sp_vel->ReAssembleMatrix(false);
      
      // Compute residuals for inner iteration
      sp_pre->Set_uk(&uk);
      sp_vel->Set_uk(&uk);
      sp_pre->ComputeDiscreteResidual( t, dt,pre_res_norm_2, pre_res_norm_inf );
      sp_vel->ComputeDiscreteResidual( t, dt,vel_res_norm_2, vel_res_norm_inf );
      
      // Check if inner loop has converged
      if ( CheckUPConvergence() ) break;
    }
    
    // Re assemble the momentum matrix or not
    sp_vel->ReAssembleMatrix(true);
    
    // Compute residuals for non linear outer iteration 
    uk_prev_NL = uk;
    sp_pre->Set_uk_prev_NL(&uk_prev_NL);
    sp_vel->Set_uk_prev_NL(&uk_prev_NL);
    sp_pre->ComputeDiscreteResidual( t, dt,pre_res_norm_2, pre_res_norm_inf );
    sp_vel->ComputeDiscreteResidual( t, dt,vel_res_norm_2, vel_res_norm_inf );
    
    // Re assemble the momentum matrix or not
    sp_vel->ReAssembleMatrix(false);
    
    // Check if outer loop has converged
    if ( CheckNonLinearConvergence() ) break;
  }
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
bool TranProblemNS2d::CheckNonLinearConvergence()
{
  bool non_linear_convergence = false;

  NL_res_norm_2 = sqrt( sqr(pre_res_norm_2) + sqr(vel_res_norm_2) );
  if ( pre_res_norm_inf > vel_res_norm_inf ) NL_res_norm_inf = pre_res_norm_inf;
  else NL_res_norm_inf = vel_res_norm_inf;

  display->Status(1,"VAL6=R(NL) %.6f",NL_res_norm_2);

  display->Status(4,"l2-norm of NL-residuals: ");
  display->Status(4,"%.6f (momentum equation)",vel_res_norm_2);
  display->Status(4,"%.6f (continuity equation)",pre_res_norm_2);
  display->Status(4,"%.6f (total l2-residual)",NL_res_norm_2);
  display->Status(4,"%.6f (total max-residual)",NL_res_norm_inf);
    
  if (  NL_res_norm_2 < NL_tol ) non_linear_convergence = true;
  if ( (NL_res_norm_2 > NL_max_tol) && (NL_it == (max_no_NL_iter-1)) ){
    display->Error("NL residual too large.");
    exit(1);
  }

  return non_linear_convergence;
}
//----------------------------------------------------------------------------------------------------
bool TranProblemNS2d::CheckUPConvergence()
{
  bool UP_convergence = false;
  
  UP_res_norm_2 = sqrt( sqr(pre_res_norm_2) + sqr(vel_res_norm_2) );
  if ( pre_res_norm_inf > vel_res_norm_inf ) UP_res_norm_inf = pre_res_norm_inf;
  else UP_res_norm_inf = vel_res_norm_inf;	  

  display->Status(1,"VAL5=R(UP) %.6f",UP_res_norm_2);
	
  display->Status(4,"l2-norm of UP-residuals: ");
  display->Status(4,"%.6f (momentum equation)",vel_res_norm_2);
  display->Status(4,"%.6f (continuity equation)",pre_res_norm_2);
  display->Status(4,"%.6f (total l2-residual)",UP_res_norm_2);
  display->Status(4,"%.6f (total max-residual)",UP_res_norm_inf);

  if ( UP_res_norm_2 < UP_tol ) UP_convergence = true;
  if ( (UP_res_norm_2 > UP_max_tol) && (UP_it == (max_no_UP_iter-1)) ){
    display->Error("UP residual too large.");
    exit(1);
  }
  
  return UP_convergence;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS2d::SetSubgridModelingUtilities()
{
  settings->Get("compute reynolds stresses",&compute_reynolds_stresses);
  settings->Get("write reynolds stresses",&write_reynolds_stresses);

  subgrid_model_on_divergence_form = true;

  // N = (sqr(nsd)+nsd)/2;
  if ( subgrid_model_on_divergence_form ) N = 6; 
  // N =  sqr(nsd);
  else N = 9;             
  // 3 levels 
  N *= 3;

  if ( compute_reynolds_stresses ) tau.newsize(no_elms,N); 
  if ( write_reynolds_stresses )   tau_print.newsize(printNoElms,N);

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS2d::SetInitialData()
{
  display->Status(1,"Set initial conditions");

  uk = 0.0;
  
  MV_ColMat<real> read_uk;

  char tmp[256];
  real x,y,pert,noRolls;
  real pi = KW_PI;
  int i,j;

  switch ( initial_data ){ 
  case zero_initial_data: 
    break;
  case white_noise: 
    for ( i=0; i < no_nodes; i++ ){
      for ( j=0; j < noeq_vel; j++ ){
	uk(i,j) = 0.1 * 2.0*(drand48() - 0.5);
	if ( j == 0 ) uk(i,j) += 1.0;
      }
    }
    break;
  case read_u0_inp:
    read_uk.newsize(printNoNodes,4);
    read_uk = 0.0;
    sprintf(tmp,"u_0.inp");
    readInp(tmp, &(*GRID)(printGrid), read_uk, noeq_vel, 0);
    if ( (*GRIDREL)(0).GetNodalProlongationSize() == 0 ) (*GRIDREL)(0).initNodalProlongation();
    (*GRIDREL)(0).projNodalData(read_uk,uk);
    break;
  case restart_inp:
    sprintf(tmp,"u_restart.inp");
    readInp(tmp, &(*GRID)(compGrid), uk, noeq_vel, 0);
    break;
  default:
    display->Error("The initial condition is not implemented.");
    exit(1);
  }

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS2d::WriteData( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt )
{
  int el,nod,i;

  // Initializing Fields
  Field Ufld_1(&(*GRID)(compGrid));
  Field Ufld_2(&(*GRID)(compGrid));
  Field Pfld(&(*GRID)(compGrid));
  
  Ufld_1.InitByNodes(U,0);
  Ufld_2.InitByNodes(U,1);
  Pfld.InitByNodes(U,3);

  // Compute norms of solution
  real L1_norm_U   = Ufld_1.Lp_norm(1) + Ufld_2.Lp_norm(1);
  real L2_norm_U   = sqrt( sqr(Ufld_1.Lp_norm(2)) + sqr(Ufld_2.Lp_norm(2)) );
  real Lmax_norm_U = max( Ufld_1.Lmax_norm(), Ufld_2.Lmax_norm() );

  real W11_seminorm_U   = Ufld_1.Wkp_seminorm(1,1) + Ufld_2.Wkp_seminorm(1,1);
  real W12_seminorm_U   = sqrt( sqr(Ufld_1.Wkp_seminorm(1,2)) + sqr(Ufld_2.Wkp_seminorm(1,2)) );
  real W1max_seminorm_U = max( Ufld_1.Wkmax_seminorm(1), Ufld_2.Wkmax_seminorm(1) );

  real W11_norm_U   = L1_norm_U + W11_seminorm_U;
  real W12_norm_U   = sqrt( sqr(L2_norm_U) + sqr(W12_seminorm_U) );
  real W1max_norm_U = max( Lmax_norm_U, W1max_seminorm_U );

  real W21_seminorm_U,W22_seminorm_U,W2max_seminorm_U,W21_norm_U,W22_norm_U,W2max_norm_U;
  W21_seminorm_U = W22_seminorm_U = W2max_seminorm_U = W21_norm_U = W22_norm_U = W2max_norm_U = 0.0;
  /*
  W21_seminorm_U   = Ufld_1.Wkp_seminorm(2,1) + Ufld_2.Wkp_seminorm(2,1);
  W22_seminorm_U   = sqrt( sqr(Ufld_1.Wkp_seminorm(2,2)) + sqr(Ufld_2.Wkp_seminorm(2,2)) );
  W2max_seminorm_U = max( Ufld_1.Wkmax_seminorm(2), Ufld_2.Wkmax_seminorm(2) );

  W21_norm_U   = W11_norm_U + W21_seminorm_U;
  W22_norm_U   = sqrt( sqr(W12_norm_U) + sqr(W22_seminorm_U) );
  W2max_norm_U = max( W1max_norm_U, W2max_seminorm_U );
  */

  real L1_norm_P   = Pfld.Lp_norm(1);
  real L2_norm_P   = Pfld.Lp_norm(2);
  real Lmax_norm_P = Pfld.Lmax_norm();

  real W11_seminorm_P   = Pfld.Wkp_seminorm(1,1);
  real W12_seminorm_P   = Pfld.Wkp_seminorm(1,2);
  real W1max_seminorm_P = Pfld.Wkmax_seminorm(1);

  real W11_norm_P   = L1_norm_P + W11_seminorm_P;
  real W12_norm_P   = sqrt( sqr(L2_norm_P) + sqr(W12_seminorm_P) );
  real W1max_norm_P = max( Lmax_norm_P, W1max_seminorm_P );

  real W21_seminorm_P,W22_seminorm_P,W2max_seminorm_P,W21_norm_P,W22_norm_P,W2max_norm_P;
  W21_seminorm_P = W22_seminorm_P = W2max_seminorm_P = W21_norm_P = W22_norm_P = W2max_norm_P = 0.0;
  /*
  W21_seminorm_P   = Pfld.Wkp_seminorm(2,1);
  W22_seminorm_P   = Pfld.Wkp_seminorm(2,2);
  W2max_seminorm_P = Pfld.Wkmax_seminorm(2);

  W21_norm_P   = W11_norm_P + W21_seminorm_P;
  W22_norm_P   = sqrt( sqr(W12_norm_P) + sqr(W22_seminorm_P) );
  W2max_norm_P = max( W1max_norm_P, W2max_seminorm_P );
  */

  // Compute the divergence of the velocity field 
  MV_Vector<real> divU(no_elms);
  for ( el=0; el < no_elms; el++ ) divU(el) = Ufld_1.EvalGradient(el,0) + Ufld_2.EvalGradient(el,1);
    
  Field Ufld_div(&(*GRID)(compGrid));
  Ufld_div.InitByElms(&divU);
  
  real L1_norm_divU   = Ufld_div.Lp_norm(1);
  real L2_norm_divU   = Ufld_div.Lp_norm(2);
  real Lmax_norm_divU = Ufld_div.Lmax_norm();

  // Compute time derivative   
  MV_ColMat<real> Udot(no_nodes,(*U).size(1));
  for ( nod=0; nod < no_nodes; nod++ ){
    for ( i=0; i < (*U).size(1); i++ ) Udot(nod,i) = ( (*U)(nod,i) - (*Uprev)(nod,i) ) / dt; 
  }
  Field Udotfld_1(&(*GRID)(compGrid));
  Field Udotfld_2(&(*GRID)(compGrid));
  Field Pdotfld(&(*GRID)(compGrid));
  Udotfld_1.InitByNodes(&Udot,0);
  Udotfld_2.InitByNodes(&Udot,1);
  Pdotfld.InitByNodes(&Udot,3);
  
  real L1_norm_Udot   = Udotfld_1.Lp_norm(1) + Udotfld_2.Lp_norm(1);
  real L2_norm_Udot   = sqrt( sqr(Udotfld_1.Lp_norm(2)) + sqr(Udotfld_2.Lp_norm(2)) );
  real Lmax_norm_Udot = max( Udotfld_1.Lmax_norm(), Udotfld_2.Lmax_norm() );

  real L1_norm_Pdot   = Pdotfld.Lp_norm(1);
  real L2_norm_Pdot   = Pdotfld.Lp_norm(2);
  real Lmax_norm_Pdot = Pdotfld.Lmax_norm();

  // Write norms of solution to matlab file
  matlab_data->SetTime(t);

  matlab_data->Set(0,L1_norm_U);
  matlab_data->Set(1,L2_norm_U);
  matlab_data->Set(2,Lmax_norm_U);
  matlab_data->Set(3,W11_seminorm_U);
  matlab_data->Set(4,W12_seminorm_U);
  matlab_data->Set(5,W1max_seminorm_U);
  matlab_data->Set(6,W11_norm_U);
  matlab_data->Set(7,W12_norm_U);
  matlab_data->Set(8,W1max_norm_U);
  matlab_data->Set(9,W21_seminorm_U);
  matlab_data->Set(10,W22_seminorm_U);
  matlab_data->Set(11,W2max_seminorm_U);
  matlab_data->Set(12,W21_norm_U);
  matlab_data->Set(13,W22_norm_U);
  matlab_data->Set(14,W2max_norm_U);
  matlab_data->Set(15,L1_norm_P);
  matlab_data->Set(16,L2_norm_P);
  matlab_data->Set(17,Lmax_norm_P);
  matlab_data->Set(18,W11_seminorm_P);
  matlab_data->Set(19,W12_seminorm_P);
  matlab_data->Set(20,W1max_seminorm_P);
  matlab_data->Set(21,W11_norm_P);
  matlab_data->Set(22,W12_norm_P);
  matlab_data->Set(23,W1max_norm_P);
  matlab_data->Set(24,W21_seminorm_P);
  matlab_data->Set(25,W22_seminorm_P);
  matlab_data->Set(26,W2max_seminorm_P);
  matlab_data->Set(27,W21_norm_P);
  matlab_data->Set(28,W22_norm_P);
  matlab_data->Set(29,W2max_norm_P);
  matlab_data->Set(30,L1_norm_divU);
  matlab_data->Set(31,L2_norm_divU);
  matlab_data->Set(32,Lmax_norm_divU);
  matlab_data->Set(33,L1_norm_Udot);
  matlab_data->Set(34,L2_norm_Udot);
  matlab_data->Set(35,Lmax_norm_Udot);
  matlab_data->Set(36,L1_norm_Pdot);
  matlab_data->Set(37,L2_norm_Pdot);
  matlab_data->Set(38,Lmax_norm_Pdot);
  matlab_data->Save();

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS2d::CompContResidual( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt, MV_Vector<real>& contResidual )
{
  int nod,el,sd;

  // R1 = | u_t + u*Du + Dp - f |  
  // R2 = max jump in gradient over element boundaries divided by h
  // R3 = time discontinuity (=0 for continuous time integration)
  // R4 = | div u |

  // Computing the R2 residual is expensive due to the approximation of second derivatives
  bool compute_R2 = false;
  neighbor = &((*GRID)(compGrid).getNeighbor());
  if ( compute_R2 ){
    if ( neighbor->elementSize() == 0 ) neighbor->init( (*GRID)(compGrid), false, false, true );
  }
  
  Field Ufld_1(&(*GRID)(compGrid));
  Field Ufld_2(&(*GRID)(compGrid));
  Field Pfld(&(*GRID)(compGrid));
  
  Ufld_1.InitByNodes(U,0);
  Ufld_2.InitByNodes(U,1);
  Pfld.InitByNodes(U,3);

  MV_ColMat<real> Udot(no_nodes,nsd);
  for ( nod=0; nod < no_nodes; nod++ ){
    for ( sd=0; sd < nsd; sd++ ) Udot(nod,sd) = ( (*U)(nod,sd) - (*Uprev)(nod,sd) ) / dt; 
  }
  Field Udotfld_1(&(*GRID)(compGrid));
  Field Udotfld_2(&(*GRID)(compGrid));
  Udotfld_1.InitByNodes(&Udot,0);
  Udotfld_2.InitByNodes(&Udot,1);
  
  // R3=0 for continuous time integration
  MV_ColMat<real> R1(no_elms,nsd);
  MV_Vector<real> R2(no_elms);
  //MV_ColMat<real> R3(no_elms,nsd);
  MV_Vector<real> R4(no_elms);

  Field Ffld_1(&(*GRID)(compGrid));
  Field Ffld_2(&(*GRID)(compGrid));
  real zero = 0.0;
  Ffld_1.InitByConstant(&zero);
  Ffld_2.InitByConstant(&zero);
    
  for ( el=0; el < no_elms; el++ ){
    R1(el,0) = fabs( Udotfld_1.Eval(el) + Ufld_1.Eval(el)*Ufld_1.EvalGradient(el,0) + Ufld_2.Eval(el)*Ufld_1.EvalGradient(el,1) 
		     + Pfld.EvalGradient(el,0) - Ffld_1.Eval(el) );   
    R1(el,1) = fabs( Udotfld_2.Eval(el) + Ufld_1.Eval(el)*Ufld_2.EvalGradient(el,0) + Ufld_2.Eval(el)*Ufld_2.EvalGradient(el,1) 
		     + Pfld.EvalGradient(el,1) - Ffld_2.Eval(el) );   
  }
  Field R1fld_1(&(*GRID)(compGrid));
  Field R1fld_2(&(*GRID)(compGrid));
  R1fld_1.InitByElms(&R1,0);
  R1fld_2.InitByElms(&R1,1);

  // Compute norms of R1 residual
  real L1_norm_R1   = R1fld_1.Lp_norm(1) + R1fld_2.Lp_norm(1);
  real L2_norm_R1   = sqrt( sqr(R1fld_1.Lp_norm(2)) + sqr(R1fld_2.Lp_norm(2)) );
  real Lmax_norm_R1 = max( R1fld_1.Lmax_norm(), R1fld_2.Lmax_norm() );

  // Compute norms of R2 residual
  real L1_norm_R2,L2_norm_R2,Lmax_norm_R2;
  if ( compute_R2 ){
    MV_Vector<real> locGrad(nsd);
    MV_Vector<real> neighborGrad(nsd);
    int i,el_start,el_stop,noNeighbors;
    for ( el=0; el < no_elms; el++ ){
      R2(el) = 0.0;
      Ufld_1.EvalGradient(el,locGrad);
      el_start = neighbor->elementIrow(el);
      el_stop = neighbor->elementIrow(el+1);
      noNeighbors = el_stop - el_start;
      for ( i=el_start; i < el_stop; i++ ){
	Ufld_1.EvalGradient(neighbor->elementJcol(i),neighborGrad);
	for ( sd=0; sd < nsd; sd++ ){
	  if ( fabs(locGrad(sd) - neighborGrad(sd)) > R2(el) ) R2(el) = fabs(locGrad(sd) - neighborGrad(sd));
	}
      }
      Ufld_2.EvalGradient(el,locGrad);
      el_start = neighbor->elementIrow(el);
      el_stop = neighbor->elementIrow(el+1);
      noNeighbors = el_stop - el_start;
      for ( i=el_start; i < el_stop; i++ ){
	Ufld_2.EvalGradient(neighbor->elementJcol(i),neighborGrad);
	for ( sd=0; sd < nsd; sd++ ){
	  if ( fabs(locGrad(sd) - neighborGrad(sd)) > R2(el) ) R2(el) = fabs(locGrad(sd) - neighborGrad(sd));
	}
      }
      R2(el) /= ( reynolds_number * LocalMeshSize(el) );
    }
    Field R2fld(&(*GRID)(compGrid));
    R2fld.InitByElms(&R2);

    L1_norm_R2   = R2fld.Lp_norm(1);
    L2_norm_R2   = R2fld.Lp_norm(2);
    Lmax_norm_R2 = R2fld.Lmax_norm();
  }

  // Compute norms of R4 residual
  for ( el=0; el < no_elms; el++ ) R4(el) = Ufld_1.EvalGradient(el,0) + Ufld_2.EvalGradient(el,1);
  Field R4fld(&(*GRID)(compGrid));
  R4fld.InitByElms(&R4);
  
  real L1_norm_R4   = R4fld.Lp_norm(1);
  real L2_norm_R4   = R4fld.Lp_norm(2);
  real Lmax_norm_R4 = R4fld.Lmax_norm();
  
  if ( !compute_R2 ){
    L1_norm_R2   = 0.0;
    L2_norm_R2   = 0.0;
    Lmax_norm_R2 = 0.0;
  }
  real L1_norm_R3   = 0.0;
  real L2_norm_R3   = 0.0;
  real Lmax_norm_R3 = 0.0;

  // Write norms to matlab file
  matlab_data->SetTime(t);

  matlab_data->Set(no_data_entries+0,L1_norm_R1);
  matlab_data->Set(no_data_entries+1,L2_norm_R1);
  matlab_data->Set(no_data_entries+2,Lmax_norm_R1);
  matlab_data->Set(no_data_entries+3,L1_norm_R2);
  matlab_data->Set(no_data_entries+4,L2_norm_R2);
  matlab_data->Set(no_data_entries+5,Lmax_norm_R2);
  matlab_data->Set(no_data_entries+6,L1_norm_R3);
  matlab_data->Set(no_data_entries+7,L2_norm_R3);
  matlab_data->Set(no_data_entries+8,Lmax_norm_R3);
  matlab_data->Set(no_data_entries+9,L1_norm_R4);
  matlab_data->Set(no_data_entries+10,L2_norm_R4);
  matlab_data->Set(no_data_entries+11,Lmax_norm_R4);
  matlab_data->Save();

  // Write element representation of residual
  contResidual.newsize(no_elms);
  for ( el=0; el < no_elms; el++ ) contResidual(el) = R1(el,0) + R1(el,1) + R4(el);
  if ( compute_R2 ){
    for ( el=0; el < no_elms; el++ ) contResidual(el) += R2(el);
  }    

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS2d::SetTolerances()
{
  // Set tolerances for iterative methods
  NL_tol   = 1.0e-2;
  UP_tol   = 1.0e-2;
  stat_tol = 1.0e-10;

  NL_max_tol = NL_tol;
  UP_max_tol = 100.0;

  max_no_UP_iter = 1;
  max_no_NL_iter = 10;  

  UP_conv = true;
  NL_conv = true;

  return 0;
}
//----------------------------------------------------------------------------------------------------


