#include "kraftwerk.h"
#include "TranProblemNS2d.hh"
#include "Adaptivity.h" 
#include "CreateSurfaces.h"
#include "Globals.hh"
#include "Settings.hh"

extern Globals *globals;
extern Settings *settings;

int GenerateMesh( MV_Vector<Grid>* GRID, int no_levels );

int GenerateGridRelations( MV_Vector<Grid>* GRID, MV_Vector<GridRelation>* GRIDREL );

int SetProblemData();

int SetKWParameters();


int main(int argc, char **argv)
{

  bool solve_primal = true;
  bool solve_dual   = false;
  
  // Set kraftwerk parameters
  SetKWParameters();

  // Set problems parameters
  SetProblemData();

  // Initialize kraftwerk
  kw_init(argc,argv);

  // Save parameters
  kw_save_parameters();
  
  // Read grid, and define Grid hierarchy
  int no_levels = 3; // Should be >2 for extrapolation to work.

  MV_Vector<Grid> GRID;
  GenerateMesh( &GRID, no_levels );
  MV_Vector<GridRelation> GRIDREL;
  GenerateGridRelations( &GRID, &GRIDREL ); 
  int fine_grid = GRID.size()-1;

  globals->Set("current computational level", fine_grid);
  globals->Set("current output level", fine_grid);
  globals->Set("current input level primal", fine_grid);

  settings->Set("write data", false);
  settings->Set("write residuals", false);

  // Define time dependent problem
  TranProblemNS2d tp(&GRID,&GRIDREL);

  kw_test();
  
  MV_ColMat<real> uk;
  
  // Solve the Navier-Stokes equations
  if ( solve_primal ){
    tp.Solve();
  }
  
  // Finished
  kw_end();
  
  return 0;
}


int SetKWParameters(){

  // Set problem
  kw_set_problem("navier-stokes");
  
  // Set parameters
  kw_set_parameter("problem description",   "Time-dependent 2d Navier-Stokes, driven cavity, Re = 1000");
  kw_set_parameter("output file prefix",    "driven_cavity_2d");
  kw_set_parameter("output samples",        5);
  //kw_set_parameter("write data",            false);
  //kw_set_parameter("write residuals",       true);
  kw_set_parameter("display",               "terminal");
  kw_set_parameter("start time",            0.0 );
  kw_set_parameter("final time",            1.0 );

  return 0;
}

int SetProblemData(){

  globals->Set("write couette pertubations",    false);
  globals->Set("write poiseuille pertubations", false);

  globals->Set("compute projections",  false);

  globals->Set("reynolds number",      1000.0);

  globals->Set("backward facing step", false);
  globals->Set("bluff body",           false);
  globals->Set("jet",                  false);
  globals->Set("driven cavity",        true);

  globals->Set("slip bc x0",           false);
  globals->Set("slip bc x1",           false);
  globals->Set("slip bc y0",           false);
  globals->Set("slip bc y1",           false);
  globals->Set("slip bc z0",           false);
  globals->Set("slip bc z1",           false);

  globals->Set("no slip bc x0",        false);
  globals->Set("no slip bc x1",        true);
  globals->Set("no slip bc y0",        true);
  globals->Set("no slip bc y1",        true);
  globals->Set("no slip bc z0",        true);
  globals->Set("no slip bc z1",        true);

  globals->Set("periodic bc x",        false);
  globals->Set("periodic bc y",        false);
  globals->Set("periodic bc z",        false);

  globals->Set("inflow bc x0",         false);
  globals->Set("outflow bc x0",        false);
  globals->Set("outflow bc x1",        false);

  globals->Set("couette bc y",         false);

  globals->Set("driven cavity bc x0",  true);
  globals->Set("driven cavity bc y1",  false);

  globals->Set("dual drag bc",         false);

  globals->Set("bnd x0",               0.0);
  globals->Set("bnd x1",               1.0);
  globals->Set("bnd y0",               0.0);
  globals->Set("bnd y1",               1.0);
  globals->Set("bnd z0",               0.0);
  globals->Set("bnd z1",               0.0);

  return 0;
}

int GenerateMesh( MV_Vector<Grid>* GRID, int no_levels ){
  
  int max_edges = 3; // Should be equal to 6 for uniform refinements.
  bool uniform_refinement = true;

  (*GRID).newsize(no_levels);

  (*GRID)(0).scan("trigrid_4_4.inp");
  //createSurfaces((*GRID)(0),0.5);
  //(*GRID)(0).presentSurfGridAndGeometry ();
  (*GRID)(0).presentGridHier();
  //int max_no_surfaces = (*GRID)(0).getSurfaceTriangulation().getNoSurfaces();
  //GridSurfaceTriang& gsurf = (*GRID)(0).getSurfaceTriangulation();

  int i,j;
  MV_Vector<int> elMarker;
  cout << "Create refined grids" << endl;
  if ( uniform_refinement ){
    for ( i=0; i < no_levels-1; i++ ){
      (*GRID)(i).refineUniformly(&(*GRID)(i+1),max_edges,true);
      //(*GRID)(i+1).getSurfaceTriangulation().initRefinedGrid((*GRID)(i+1),false);
      (*GRID)(i+1).presentGridHier();
      //(*GRID)(i+1).presentSurfGridAndGeometry();
    }
  } else{
    for ( i=0; i < no_levels-1; i++ ){
      if ( i < 2 ){
	cout << "refine elements uniformly, level " << i+1 << endl;
	(*GRID)(i).refineUniformly(&(*GRID)(i+1),max_edges,true);
	//(*GRID)(i+1).getSurfaceTriangulation().initRefinedGrid((*GRID)(i+1),false);
	(*GRID)(i+1).presentGridHier();
	//(*GRID)(i+1).presentSurfGridAndGeometry();
      }	else{  
	cout << "refine elements non uniformly, level " << i+1 << endl;
	elMarker.newsize((*GRID)(i).getNoElms());
	elMarker = 0;
	for ( j=0; j < (*GRID)(i).getNoElms(); j++ ){
	  if (    ( fabs((*GRID)(i).getCoor((*GRID)(i).loc2glob(j,0),0)-0.625) <= 0.5 ) 
	       || ( fabs((*GRID)(i).getCoor((*GRID)(i).loc2glob(j,0),1)-0.625) <= 0.5 ) 
	       || ( fabs((*GRID)(i).getCoor((*GRID)(i).loc2glob(j,0),2)-0.625) <= 0.5 ) ){
	    elMarker(j) = 1;
	  }
	}
	(*GRID)(i).refine(&(*GRID)(i+1),elMarker);
	//(*GRID)(i+1).getSurfaceTriangulation().initRefinedGrid((*GRID)(i+1),false);
	(*GRID)(i+1).presentGridHier();
	//(*GRID)(i+1).presentSurfGridAndGeometry();
      }
    }
  }

  return 0;
}



int GenerateGridRelations( MV_Vector<Grid>* GRID, MV_Vector<GridRelation>* GRIDREL ){  
  
  int fine_grid = (*GRID).size()-1;
  
  /*
    GridRelation GR_h_2h(&(*GRID)(fine_grid),&(*GRID)(fine_grid-1));
    GridRelation GR_h_4h(&(*GRID)(fine_grid),&(*GRID)(fine_grid-2));
    GridRelation GR_2h_4h(&(*GRID)(fine_grid-1),&(*GRID)(fine_grid-2));
  */
  //GridRelation GR_h_8h(&(*GRID)(fine_grid),&(*GRID)(fine_grid-3));
  //GridRelation GR_2h_8h(&(*GRID)(fine_grid-1),&(*GRID)(fine_grid-3));
  //GridRelation GR_4h_8h(&(*GRID)(fine_grid-2),&(*GRID)(fine_grid-3));

  MV_Vector<GridRelation> GR;

 /*
    GR(0) = GR_h_2h;
    GR(1) = GR_h_4h;
    GR(2) = GR_2h_4h;
  */
  //GR(3) = GR_h_8h;
  //GR(4) = GR_2h_8h;
  //GR(5) = GR_4h_8h;

  //GR(0).initNodalMap();
  
  //GR(0).initNestElMap();
  //GR(1).initNestElMap();
  //GR(2).initNestElMap();
  //GR(3).initNestElMap();
  //GR(4).initNestElMap();
  //GR(5).initNestElMap();

  (*GRIDREL) = GR;
    
  return 0;  
}







/*

//int init_type = KW_INIT_RESTART_DX;
int init_type = KW_INIT_NONE;
  
void set_initial_condition(MV_ColMat<real> &u0, int noNodes, int fine_grid, int init_type,
									MV_Vector<Grid> &GRID, MV_Vector<GridRelation> &GRIDREL)
{
  // FIXME: This should be done somewhere else

  if ( init_type != KW_INIT_RESTART_DX )
	 return;
  
  real x,y,z;
  int i,j;
  OpenDX *opendx = new OpenDX("driven_cavity_re1000_t30_h32.dx",2,3,1);
  
  u0 = 0.0;
  
  // Read data
  opendx = new OpenDX("driven_cavity_re1000_t30_h32.dx",2,3,1);
  opendx->GetData(&GRID(fine_grid),&u0,10.0);
  delete opendx;
  
  // Add white noise
  double noise_level = 0.0;
  if ( noise_level ){
	 cout << "Adding white noise." << endl;
 	 for (i=0;i<noNodes;i++)
		for (j=0;j<3;j++){
		  // The multigrid solver doesn't seem to like us messing with the boundary conditions
		  if ( fabs(u0(i,j)) > 0.001 )
			 u0(i,j) += noise_level * 2.0*(drand48() - 0.5);
		}
  }

}

void set_initial_condition_dual(MV_ColMat<real> &u0, MV_Vector<Grid> &grid)
{
  real x,y,z;

  int i;
  int fine_grid = grid.size()-1;
  int noNodes   = u0.size(0);

  real a  = 0.05;
  real c  = -1.0;
  real x0 = 0.7;
  real y0 = 0.5;
  real z0 = 0.5;
  
  // Set initial data for the dual
  u0 = 0.0;
  for (i=0;i<noNodes;i++){
	 x = grid(fine_grid).getCoor(i,0);
	 y = grid(fine_grid).getCoor(i,1);
	 z = grid(fine_grid).getCoor(i,2);


	 //	 if ( (x>=x1) && (x<=x2) && (y>=y1) && (y<=y2) && (z>=z1) && (z<=z2) )
	 //		u0(i,2) = 1.0;

	 // The divergence must be zero!

	 u0(i,0) = divzero_field(1,0,a,c,x0,y0,z0,x,y,z);
	 u0(i,1) = divzero_field(1,1,a,c,x0,y0,z0,x,y,z);
	 u0(i,2) = divzero_field(1,2,a,c,x0,y0,z0,x,y,z);
	 
  }
   
}

// FIXME: This should not be placed here

double divzero_field(int direction, int component, double a, double c,
							double x0, double y0, double z0,
							double x,  double y,  double z)
{
  // Returns component <component> of a field with zero divergence
  // with main direction <direction> (0,1 or 2 for x,y or z)

  x -= x0;
  y -= y0;
  z -= z0;

  double r2   = x*x+y*y+z*z;
  double a2   = a*a;
  double expv = exp(-0.5*r2/a2);
  
  double field[3];

  switch (direction){
  case 0:
	 field[0] = (2.0-(y*y+z*z)/a2);
	 field[1] = y*x/a2;
	 field[2] = z*x/a2;
	 break;
  case 1:
	 field[0] = x*y/a2;
	 field[1] = (2.0-(x*x+z*z)/a2);
	 field[2] = z*y/a2;
	 break;
  case 2:
	 field[0] = x*z/a2;
	 field[1] = y*z/a2;
	 field[2] = (2.0-(x*x+y*y)/a2);
	 break;
  default:
	 cout << "Illegal direction for field!" << endl;
	 field[0] = 0.0;
	 field[1] = 0.0;
	 field[2] = 0.0;
  }
  
  return (c*field[component]*expv);
}

*/
